//////////////////////////////////////////////////////////////////////
//
// LBIND.cpp:  binding related routines
//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
// Copyright (c) 2000  Gesytec GmbH
//
// Date:     6. 1.00
//
// Author: mb, mboehlke@gesytec.de
//
//////////////////////////////////////////////////////////////////////

#include "windows_compat.h"
#include "ni_privat.h"
#include "ni_host.h"
#include "ni_mgmt.h"
#include "nm_codes.h"
#include "g2Lon.h"
#include "LHA.h"
#include "LBIND.h"
#include "XIF.h"
#include "pers_ram.h"
#include "xDbg.h"
#include "Utils.h"

#include <string.h>
#include <stdio.h>

#define BINDSIG1	0x2D
#define BINDSIG2	0x75

typedef struct {
    BYTE	sflag1;
    BYTE	sflag2;
} STARTFLAGS;

extern	bool ldv_LocalOut(UCHAR _code, UCHAR* _data, short len);
extern	bool ldv_LocalReq(UCHAR _code, UCHAR* _data, short len, void* indata);

//***************************************************************************

CBind::CBind(CHANeuron *host)
{
//	DBGERR("SET DGB_BIND");
//	DBGSET(DBG_BIND);
	mHost = host;
}

//---------------------------------------------------------------------------

CBind::~CBind(void)
{
}

//---------------------------------------------------------------------------

bool CBind::ChkBinding(void)
{
DBG(DBG_BIND, "+CBind::ChkBinding");
	CXIF *xif = new CXIF;
	bool ok = xif->OpenXIF();
	if (!ok) {
		BYTE f[2];
		f[0] = 0;
		f[1] = 0;
		PersistWrite(PRAM_STARTFLAGS, &f, sizeof(f)); // clear signatur
	}

	if (ok) ok = ReadBinding(xif);					// read interface
if (!ok) DBGERR(" ReadBinding failed");
	if (ok) ok = InitBinding(xif);					// initialize structures
if (!ok) DBGERR(" InitBinding failed");

	delete xif;										// delete utility

if (ok) DBG(DBG_BIND, "-CBind::ChkBinding OK");
else	DBGERR("-CBind::ChkBinding failed");
	return (ok);
}

//---------------------------------------------------------------------------

bool CBind::ReadBinding(CXIF *xif)
{
DBG(DBG_XIF, "+ReadBinding");
	short			n;
	nv_fixed_struct *pfix;
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	DBG(DBG_XIF, "ReadHead(1)");
	if (!xif->ReadHead(1)) return (DBGERR("ReadHead(1) failed"));

	//if (DBG(DBG_BIND)) {
		char dbuf[64];
		sprintf(dbuf, "NVcount = %d", xif->mNvs);
		DBG(DBG_BIND, dbuf);
	//}
	if (xif->mNvs <= 0) return (false);

	DBG(DBG_XIF, "ReadNVs(1)");
	for (n = 0; n < xif->mNvs; n++) {
		if (!xif->ReadNV(1, n)) return (DBGERR("ReadNV failed"));
	}
	if (DBG(DBG_BIND)) {
		char dbuf[64];
		sprintf(dbuf, "mSNVTLen = %d  %x", xif->mSNVTLen, xif->mSNVTLen);
		DBG(DBG_BIND, dbuf);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	mHost->mNVanz = xif->mNvs;							// count of NVs
	memcpy(mProgid, xif->mPid, sizeof(mProgid));		// PID

// create structures ...
	mHost->mNvf = new nv_fixed_struct[mHost->mNVanz];	// fixed structure
	memset(mHost->mNvf, 0, sizeof(nv_fixed_struct) * mHost->mNVanz);

	mHost->mSnvtlen = xif->mSNVTLen +16; // (16 is just some more)
	mHost->mpSnvt   = new BYTE[mHost->mSnvtlen];		// snvt structure
	xif->SetSnvtP(mHost->mpSnvt, mHost->mSnvtlen);
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	DBG(DBG_XIF, "ReadHead(1.1)");
	if (!xif->ReadHead(1))								// skip header
		return (DBGERR("ReadHead(1.1) failed"));
	DBG(DBG_XIF, "ReadNVs(2)");
	for (n = 0; n < xif->mNvs; n++) {
		// fill snvt description ...
		if (!xif->ReadNV(2, n)) return (DBGERR("ReadNV2 failed"));

		pfix = mHost->GetNvFixed(n);					// pointer into fix table
		if (!pfix) return (false);

		pfix->nv_length = xif->mNVLength;				// nv_length
		pfix->nv_dir    = xif->mNVDir;					// nv_direction
		pfix->nv_sync	= xif->mNVDesc.nv_sync;			// nv_sync

		CNV *nv = mHost->AddNV(xif->mNVName, xif->mNVDoc, pfix, &xif->mNVDesc);

	// preset configs to not bound ...
		WORD	  sel = 0x3FFF - n;
		nv_struct nvc;

		nvc.nv_selector_hi = (BYTE)(sel >> 8);
		nvc.nv_direction   = xif->mNVDir ? 1 : 0;
		nvc.nv_priority	   = 0;
		nvc.nv_selector_lo = (BYTE)sel;
		nvc.nv_addr_index  = 0x0F;
		nvc.nv_auth 	   = 0;	
		nvc.nv_service	   = 0;	
		nvc.nv_turnaround  = 0;	

		nv->UpdNVStruct(&nvc);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	DBG(DBG_XIF, "ReadHead(3)");		
	if (!xif->ReadHead(3)) 
		return (DBGERR("ReadHead(3) failed"));			// add nodes selfdoc

	for (n = 0; n < xif->mNvs; n++) {
														// fill snvt selfdoc ...
		if (!xif->ReadNV(3, n)) return (DBGERR("ReadNV3 failed"));
	}
	DBG(DBG_BIND, "snvt: ", mHost->mpSnvt, 32);
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	return (true);
}

//---------------------------------------------------------------------------

bool CBind::InitBinding(CXIF *xif)
{
	
	STARTFLAGS	sflags;
	bool		init = false;

	mAdrtbl_anz = 15;			// ???
	mNodeState = GetNodeState();			// aktuellen Status holen
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// get Startflags ...
	DBG(DBG_BIND, "get Startflags");
    if (!PersistRead(PRAM_STARTFLAGS, &sflags, sizeof(STARTFLAGS))) {
		return (DBGERR("Read Startflags failed"));
	}

	DBG(DBG_BIND, "Flags read: ", (BYTE *)&sflags, sizeof(sflags));
	if ((xif->mSig[0] == 0) && (xif->mSig[1] == 0)) {
		DBG(DBG_BIND,"RESET xif->mSig");
		xif->mSig[0] = BINDSIG1;
		xif->mSig[1] = BINDSIG2;
	}
	if ((sflags.sflag1 != xif->mSig[0]) || (sflags.sflag2 != xif->mSig[1])) {
		init = true;

		DBG(DBG_BIND, "+ClrBinding");
		WORD dbg = DBGSET(0);					// disable debug
		ClrBinding(xif);						// remove all binding info
		DBGSET(dbg);
		DBG(DBG_BIND, "-ClrBinding");

		sflags.sflag1 = xif->mSig[0];
		sflags.sflag2 = xif->mSig[1];
		if (!PersistWrite(PRAM_STARTFLAGS, &sflags, sizeof(STARTFLAGS)))
			return (DBGERR("Write Startflags failed"));
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if (!bnd_readonly()) {
		return (DBGERR("bnd_readonly failed"));
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if (!init && !ReadNVCfgs()) {
		return (DBGERR("ReadNVCfgs failed"));
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (!mHost->SetNodeMode(CHANGE_STATE, CNFG_ONLINE))
		return (DBGERR("SetNodeMode CNFG_ONLINE failed"));
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	mHost->mMode = APPL_ONLINE;
    return (true);
}        

//---------------------------------------------------------------------------

bool CBind::bnd_readonly(void)
// read read_only structure 
{
DBG(DBG_BIND,"+bnd_readonly");
	read_only_data	rod;
	if (!L_ReadMem(NULL, 0, sizeof(read_only_data), READ_ONLY_RELATIVE, (BYTE *)&rod))
		return (DBGERR("Read ROD failed"));

    if (memcmp(&rod.id_string, mProgid, ID_STR_LEN)) {	// check if equal
		mHost->UpdNODE(mProgid);
		memcpy(&rod.id_string, mProgid, ID_STR_LEN);	// write Program-ID 
		if (!L_WriteMem(NULL, 0x0D, 8, READ_ONLY_RELATIVE, mProgid)) {
			return (DBGERR("Write PID failed"));
		}
	}
DBG(DBG_BIND,"-bnd_readonly");
    return (true);
}

//---------------------------------------------------------------------------

bool CBind::ReadNVCfgs(void)
// read config data from persistant memory (here file NVCfg.dat) 
{
	DBG(DBG_BIND, "+ReadNVCfgs");
	CNV			*nv;
    short		i;
	nv_struct	nvc;

	WORD dbg = DBGSET(0);					// disable debug
    for (i = 0; i < mHost->mNVanz; i++) {
		nv = mHost->NVbyInd(i);
		if (!nv) return (DBGERR("ReadNVCfgs nv?"));

		if (!PersistRead(PRAM_NVCBASE + i, &nvc, sizeof(nv_struct))) {
			DBGSET(dbg);
			return (DBGERR("ReadNVCfgs failed"));
		}

		if (true) { //DBG(0)) {
			char dbuf[64];
			sprintf(dbuf, "NVCfg: %d ", i);
			hexdump(&dbuf[strlen(dbuf)], (BYTE *)&nvc, sizeof(nv_struct), 0);
			DBGERR(dbuf);
		}

		nv->UpdNVStruct(&nvc);					// update CNV members
    }
	DBGSET(dbg);
	DBG(DBG_BIND, "-ReadNVCfgs");
	return (true);
}

//---------------------------------------------------------------------------

bool CBind::WriteNVCfg(short idx, nv_struct *pnvc)
{
	if (!PersistWrite(PRAM_NVCBASE + idx, pnvc, sizeof(nv_struct)))
		return (DBGERR("WriteNVCfg failed"));
	return (true);

}

//---------------------------------------------------------------------------

static addr_struct Unused_adrtbl = {
    0,0, 0,0, 0,0,0,0, 
} ;

static UCHAR Blank_domain[] = {
    0,0,0,0,0,0, 0, 0x80, 0,    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
} ;
static UCHAR Unused_domain[] = {
    0,0,0,0,0,0, 0, 0x80, 0xFF, 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
} ;

bool CBind::ClrBinding(CXIF *xif) 
{
	short	i;
	BYTE	outbuf[512];
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// clear adrtable ...
    memcpy(&outbuf[1], &Unused_adrtbl, sizeof(addr_struct));
    for (i = 0; i < mAdrtbl_anz; i++) {
        *outbuf = (BYTE)i;        
        if (!ldv_LocalOut(NM_update_adr, outbuf, sizeof(addr_struct))) {
            return (DBGERR("NM_update_adr failed"));
        }
    }
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// clear nv config ...
	CNV		  *nv;
	nv_struct nvc;
	for (i = 0; i < mHost->mNVanz; i++) {
		nv = mHost->NVbyInd(i);
		if (!nv) return (DBGERR(" (Clear NVConfig no nv"));
		nv->GetNVStruct(&nvc);
		if (!WriteNVCfg(i, &nvc)) return (false);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	return (true);
}

//---------------------------------------------------------------------------

typedef struct {
    unsigned int  xmit_errors;		// !!! Daten sind big endian !!!
    unsigned int  transaction_timeouts; // !!! Daten sind big endian !!!
    unsigned int  rcv_transaction_full; // !!! Daten sind big endian !!!
    unsigned int  lost_msgs;		// !!! Daten sind big endian !!!
    unsigned int  missed_msgs;		// !!! Daten sind big endian !!!
    unsigned char reset_cause;
    unsigned char node_state;	    // Interpret with 'nm_node_state'
    unsigned char version_number;
    unsigned char error_log;
    unsigned char model_number;
} node_status;

BYTE CBind::GetNodeState(void)
{
	DBG(DBG_BIND, "GetNodeState");
    node_status ns;

    if (ldv_LocalReq(NM_query_status, NULL, 0, &ns)) return (ns.node_state);

	DBGERR("GetNodeState failed");
    return (0);
}

//---------------------------------------------------------------------------

