//////////////////////////////////////////////////////////////////////
//
// LHA.cpp:  implementation of "LON Host Application"
//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
// Copyright (c) 1999  Gesytec GmbH
//
// Date:     6. 1.00
//
// Author: mb, mboehlke@gesytec.de
//
//
//////////////////////////////////////////////////////////////////////

#include "windows_compat.h"
#include "ni_privat.h"
#include "ni_host.h"
#include "nm_codes.h"
#include "g2Lon.h"
#include "watcher.h"
#include "LHA.h"
#include "pers_ram.h"
#include "xDbg.h"
#include "fifo.h"
#include "Utils.h"

#include <unistd.h>		// fuer usleep(), fork(), setsid() und chdir()

#define NV_IDX_ESC	0xFF

// some globals ...
extern  HANDLE		gInEvt;					// incoming event
extern	CFifo		*gInMsgs;				// messages from ldv_read
extern	CFifo		*gNVUpds;				// list of NVs which are updated
extern	CFifo		*gMsgs;					// messages to application

static	short		gRun = 0;

static bool ldv_RespOut(ExpAppBuf* reqbuf, UCHAR _code, UCHAR* _data, short len);

//---------------------------------------------------------------------------

static DWORD IThreadStub(DWORD arg)
{
	CHANeuron *pp = (CHANeuron *)arg;
	return (pp->Receive());
}

//***************************************************************************
// CHANeuron extends the CNeuron for LHA processing ...
CHANeuron::CHANeuron(NODE *node)
{
DBGERR("CHANeuron::CHANeuron");
//if (!node) DBGERR("CHANeuron mNODE == 0");
//	CNeuron::CNeuron(node);
//if (!mNODE) DBGERR("CHANeuron mNODE == 0");
	mNODE	 = node;

	mBind	 = NULL;							// ref to bind object
	mNVanz	 = 0;
	mSnvtlen = 0;
	mNvf	 = NULL;
	mpSnvt	 = NULL;
	mMode	 = APPL_OFFLINE;
	mState	 = 0;
	mIsLHA	 = false;

	gMsgs = new CFifo(5000);					// create fifo for msgs
}

//---------------------------------------------------------------------------

CHANeuron::~CHANeuron(void)
{
	if (gRun) gRun = 1;	// force thread to terminate
	//while (gRun) Sleep(20);
	while (gRun) usleep(20*1000);

	if (mBind)   delete mBind;		mBind   = NULL;
	if (gNVUpds) delete gNVUpds;	gNVUpds = NULL;
	if (gMsgs)   delete gMsgs;		gMsgs   = NULL;
}

//---------------------------------------------------------------------------

bool CHANeuron::ChkBinding(void)
{
	DBG(DBG_LHA, "+ChkBinding");
	if (!mBind) mBind = new CBind(this);	// create binding utils object

	if (mBind->ChkBinding()) {
		mIsLHA	= true;						// we are LHA
		gNVUpds = new CFifo(500);			// create fifo for nv updates
		DBG(DBG_LHA, " ChkBinding OK");
	}
	else {
		DBG(DBG_LHA, " ChkBinding failed");
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	delete mBind;							// dont need this anymore
	mBind = NULL;

	gRun = 2;								// allow thread to run
	CreateThread((LPVOID) &IThreadStub, this);		// create thread incoming
	return (true);
}

//---------------------------------------------------------------------------

bool CHANeuron::IsLHA(void)
{
	return (mIsLHA);
}

//---------------------------------------------------------------------------

CNV	*CHANeuron::AddNV(char *name, char *selfdoc,
					  nv_fixed_struct  *pfix,
					  snvt_desc_struct *desc)
{
	short idx = NVCount();

	CNV	*nv = new CNV(this, idx);
	m_nvs.AddTail(nv);

	my_strncpy(nv->mNVName, name, sizeof(nv->mNVName));		// name

	if (nv->mpNVDoc) delete nv->mpNVDoc;
	nv->mpNVDoc = new char[strlen(selfdoc) +1];				// selfdoc
	strcpy(nv->mpNVDoc, selfdoc);

	nv->mSnvt    = desc->snvt_type_index;
	nv->mPri_len = pfix->nv_length;

	return (nv);
}

//---------------------------------------------------------------------------

DWORD CHANeuron::Receive(void)
{
	DWORD		dwWaitResult;
	short		len;
    ExpAppBuf	buf;

	while (gRun == 2)
	{
		dwWaitResult = WaitForSingleObject(gInEvt, 100);

		while (true) {
			len = gInMsgs->Get((BYTE *)&buf);
			if (!len) break;						// no msg received

			if (buf.msg_hdr.pnv.msg_type) {			// incoming NV-Message
				DBG(DBG_LHA, "+IncomingNV ", &buf, len);
				IncomingNV(&buf, len);
			}
			else {									// incoming exp-Message
//				DBG(DBG_LHA, "+IncomingMsg ", &buf, len);
				IncomingMsg(&buf, len);
			}
		}
	}
	gRun = 0;
	return (0);
}

//---------------------------------------------------------------------------

bool CHANeuron::IncomingNV(ExpAppBuf* buf, short mlen) // ????
{
	if (!IsLHA()) return (false);

    UCHAR inx, len;
	nv_fixed_struct	*pfix;

    inx  = buf->_data.pnv.index;
	pfix = GetNvFixed(inx);
	if (!pfix) return (DBGERR("IncomingNV !pfix"));

    if ((buf->msg_hdr.pnv.cmpl_code == MSG_NOT_COMPL) &&
		(pfix->nv_dir == 0)) {

		len = pfix->nv_length;
		if (!len) return (DBGERR("IncomingNV length"));

		CNV	*nv = NVbyInd(inx);						// get NV
		if (!nv) return (DBGERR("IncomingNV nv?"));

		nv->SetValue(buf->_data.pnv._data);
		DBG(DBG_LHA, "NV update -> fifo");
		gNVUpds->Put((BYTE *)&nv, sizeof(nv));		// put ref into fifo

		return (true);
    }

	return (DBGERR("IncomingNV MSG_NOT_COMPL | nv_dir"));
}

//---------------------------------------------------------------------------

bool CHANeuron::IncomingHostNV(ExpAppBuf* buf)
{
	DBG(DBG_LHA, "+Incoming HostNV");
	if (mMode != APPL_ONLINE) return (false);

    WORD			sel;
    nv_struct		nvc;
	nv_fixed_struct	*pfix;
    bool			auth_ok;
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    sel = (buf->_data.unv.NV_selector_hi << 8) + buf->_data.unv.NV_selector_lo;

	CNV *nv = W_GetNVbySel(this, sel);
	if (!nv) return (DBGERR("NV not found"));

	pfix = GetNvFixed(nv->Index());
	nv->GetNVStruct(&nvc);
// Message is authentic if it was authenticated by the Neuron or
// authentication is not configured for this NV
	auth_ok = buf->msg_hdr.exp.auth || !nvc.nv_auth;
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
/*	if (buf->msg_hdr.exp.st == REQUEST) {		// it's an NV poll ????
		UnprocessedNV _data;

        _data.must_be_one	 = 1;
        _data.direction      = ~buf->_data.unv.direction;
        _data.NV_selector_hi = buf->_data.unv.NV_selector_hi;
        _data.NV_selector_lo = buf->_data.unv.NV_selector_lo;
        len = 1;
        if (auth_ok && mOnline) {
            memcpy(GetNvFixed( inx )->nv_address, _data._data,
                    ldv_GetNvFixed( inx )->nv_length ) ;
            len += ldv_GetNvFixed( inx )->nv_length ;
        }
        ldv_RespOut( buf, *(UCHAR*)&_data, ((UCHAR*)&_data) + 1, len ) ;
        return( 0 ) ;
    }
*/
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Come here if it's not a poll
	if (!auth_ok) return(DBGERR("auth failed"));	// failed authentication

    if ((buf->msg_hdr.exp.cmpl_code == MSG_NOT_COMPL) &&
		(pfix->nv_dir == 0)) {

		nv->SetValue(buf->_data.unv._data);
		DBG(DBG_LHA, "NV update -> fifo");
		gNVUpds->Put((BYTE *)&nv, sizeof(nv));		// put ref into fifo
		return (true);
    }
	return (DBGERR("LHA_WRITE2NVOUT"));
}

//---------------------------------------------------------------------------

bool CHANeuron::IncomingMsg(ExpAppBuf* buf, short mlen)
{
	bool lha = IsLHA();

    if((buf->msg_hdr.exp.st == REQUEST) && (buf->msg_hdr.exp.response)){
		DBG(DBG_LHA, "IncomingResp3");
//????		return( ldv_IncomingResp( yinx, buf ) ) ;
		return (true);
    }
	else if (buf->_data.unv.must_be_one) {
		return (lha && IncomingHostNV(buf));
	}
    else {
// 		DBG(DBG_LHA, "IncomingMsg ", (BYTE *)&buf->_data.exp._code, 1);

		switch (buf->_data.exp._code) {
			case NM_nv_fetch		: return (lha && FetchNV(buf));

			case NM_query_snvt		: return (lha && QuerySNVT(buf));

			case NM_set_node_mode	: return (lha && SetNodeModeH(buf)) ;

			case NM_query_nv_config : return (lha && QueryNVConfig(buf));

			case NM_update_nv_config: return (lha && UpdNVConfig(buf));

			default	:
				if (gMsgs) return (gMsgs->Put((BYTE *)buf, mlen));
        }
    }
	return (false);
}

//---------------------------------------------------------------------------

bool CHANeuron::FetchNV(ExpAppBuf* buf)
{
	DBG(DBG_LHA, "+FetchNV");
    UCHAR	mydata[32], _code ;
    short	nvlen, mylen;
    UCHAR	*dp = mydata ;

    if (buf->msg_hdr.exp.st != REQUEST) return (DBGERR("FetchNV no REQUEST"));

    short inx = GetNvIndex(buf);
	CNV	  *nv = NVbyInd(inx);							// get NV object

    if (!nv) {
		DBGERR("FetchNV nv?");
        mylen = 0 ;
        _code = NM_nv_fetch_fail ;
    }
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    else {
		nvlen = nv->Length();

        if (inx < NV_IDX_ESC) {
            *dp++ = (UCHAR)inx;
            mylen = nvlen + 1;
        }
        else{
            *dp++ = NV_IDX_ESC;
			*dp++ = (BYTE)(inx >> 8);
			*dp++ = (BYTE)inx;
            mylen = nvlen + 3;
        }

		nv->GetValue(dp);
        _code = NM_nv_fetch_succ ;
    }
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	ldv_RespOut(buf, _code, mydata, mylen);
	DBG(DBG_LHA, "-FetchNV");
	return (true);
}

//---------------------------------------------------------------------------

bool CHANeuron::QuerySNVT(ExpAppBuf* buf)
{
//	DBG(DBG_LHA, "+QuerySNVT");
	if (!mpSnvt) return (false);

    WORD	mylen, off;
    UCHAR	mydata[64];

    mylen = buf->_data.exp._data[2];
	if (mylen > sizeof(mydata)) return (DBGERR(" QuerySNVT length"));

    off	= (buf->_data.exp._data[0] << 8) + buf->_data.exp._data[1];
	if ((off + mylen) > mSnvtlen) return (DBGERR(" QuerySNVT offset"));

    memcpy(mydata, mpSnvt + off, mylen);
//	DBG(DBG_LHA, " QuerySNVT ", mydata, mylen);

	ldv_RespOut(buf, NM_query_snvt_succ, mydata, mylen);
//	DBG(DBG_LHA, "-QuerySNVT");
	return (true);
}

//---------------------------------------------------------------------------

bool CHANeuron::QueryNVConfig(ExpAppBuf* buf)
{
//	DBG(DBG_LHA, "+QueryNVConfig");
    UCHAR	mydata[sizeof(nv_struct)], _code;
	short	inx, mylen;

    if (buf->msg_hdr.exp.st != REQUEST) return (DBGERR(" QueryNVConfig no REQUEST"));

    if ((inx = GetNvIndex(buf)) < 0) {
        mylen = 0 ;
        _code = NM_query_nv_config_fail;
    }
    else {
		CNV *nv = NVbyInd(inx);
		if (!nv) return (DBGERR(" QueryNVConfig no nv"));
		nv->GetNVStruct((nv_struct *)mydata);
        mylen = sizeof(nv_struct);
        _code = NM_query_nv_config_succ;
    }

//DBG(0, "Config ", mydata, mylen);
	ldv_RespOut(buf, _code, mydata, mylen);
//	DBG(DBG_LHA, "-QueryNVConfig");
	return (true);
}

//---------------------------------------------------------------------------

bool CHANeuron::UpdNVConfig(ExpAppBuf* buf)
{
	DBG(DBG_LHA, "+UpdNVConfig");
    UCHAR	_code, *dp;

//???? if (buf->msg_hdr.exp.st != REQUEST) return (DBGERR(" UpdNVConfig no REQUEST"));

	_code = NM_update_nv_config_fail;				// assume failure
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (buf->_data.exp._data[0] == NV_IDX_ESC)
        dp = &buf->_data.exp._data[3];
    else
        dp = &buf->_data.exp._data[1];
	DBG(DBG_LHA, " UpdNVConfig ********* ", dp, sizeof(nv_struct));
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    short	inx	= GetNvIndex(buf);
	CNV		*nv	= NVbyInd(inx);
	if (nv) {
		nv->UpdNVStruct((nv_struct *)dp);			// update CNV members
		PersistWrite(PRAM_NVCBASE + inx, dp, sizeof(nv_struct));
        _code = NM_update_nv_config_succ ;
	}
	else {
		DBG(DBG_ALL, "UpdNVConfig inx?");
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	ldv_RespOut(buf, _code, NULL, 0);
	DBG(DBG_LHA, "-UpdNVConfig");
	return (true);
}

//---------------------------------------------------------------------------

bool CHANeuron::SetNodeModeH(ExpAppBuf *buf)
{
	DBG(DBG_LHA, "+SetNodeModeH");

// When a set_node_mode (on/off) line is received, the corresponding
// command is sent down to the network interface
	mMode  = buf->_data.exp._data[0];
	mState = buf->_data.exp._data[1];

	if ((mMode == CHANGE_STATE) && (mState == CNFG_ONLINE)) mMode = APPL_ONLINE;

	BYTE mybuf[2];
	mybuf[0] = (mMode == APPL_ONLINE) ? niONLINE : niOFFLINE;
	mybuf[1] = 0;
	L_LdvWrite((ExpAppBuf *)mybuf, 2);

	DBG(DBG_LHA, "-SetNodeModeH");
	return (true);
}

//---------------------------------------------------------------------------

short CHANeuron::GetNvIndex(ExpAppBuf* buf)
{
    short inx = buf->_data.exp._data[0] ;

    if (buf->_data.exp._data[0] == NV_IDX_ESC) {
		inx = buf->_data.exp._data[1];
		inx = (inx << 8) + buf->_data.exp._data[2];
    }

    if (inx < mNVanz) return (inx);
    return (-1);
}

//---------------------------------------------------------------------------

nv_fixed_struct	*CHANeuron::GetNvFixed(short idx)
{
	if (idx >= mNVanz) return (NULL);
	return (&mNvf[idx]);
}

//---------------------------------------------------------------------------

void CHANeuron::UpdNODE(BYTE *pid)
{
if (!mNODE) DBGERR("UpdNODE mNODE == 0");
	if (!mNODE) return; // ProgID fuer Host Node
	memcpy(mNODE->pid, pid, sizeof(mNODE->pid));
}

//---------------------------------------------------------------------------

static UCHAR gLdv_hostnvhdr[] = {
/*-----------------------NiHdr----------------------------------------------*/
 	0x10 | 0x00              , /*  1: niCOMM + Queue ni(N)TQ(_P)   variabel */
    16	                     , /*  2: Preset folgende Laenge       variabel */
/*-----------------------MsgHdr---------------------------------------------*/
	0x00 | 0x00              , /*  3: UnProcNV + Tag(AdrTbl-Index) variabel */
    0x00                     , /*  4: expl. Adressierung                    */
    2	                     , /*  5: preset Nettodatenlaenge      variabel */
/*-----------------------Addr-----------------------------------------------*/
    IMPLICIT                 , /*  6: impl. Addr-Struktur                   */
    0                        , /*  7: addr table entry number      unused   */
    0x00                     , /*  8:                                       */
    0x00                     , /*  9:                                       */
    0x00                     , /* 10:                                       */
    0x00                     , /* 11: NID_0                                 */
    0x00                     , /* 12: NID_1                                 */
	0x00                     , /* 13: NID_2                                 */
    0x00                     , /* 14: NID_3                                 */
    0x00                     , /* 15: NID_4                                 */
    0x00                     , /* 16: NID_5                                 */
/*-----------------------Code/Data------------------------------------------*/
    0x80                     , /* 17: must_be_one + dir + selhigh  variabel */
    0x00                     , /* 18: selector low                 variabel */
} ;

static NI_Queue gNi_queue[ 2 ][ 2 ] = {        // define MIP queue
    { niTQ, niTQ_P }, { niNTQ, niNTQ_P }
} ;

bool CHANeuron::SendNV(CNV *nv, BYTE *data)
{
	DBG(DBG_LHA, "+SendNV");
	if (!nv->isOutput()) return (false);

	short len = nv->Length();
	if (!len) return (DBGERR("Send NV Length = 0?"));

	if (mMode != APPL_ONLINE) return (DBGERR("Send offline?"));

	nv_fixed_struct	*pfix = GetNvFixed(nv->Index());
	if (!pfix) return (DBGERR("Send pfix?"));

	nv_struct nvc;
	nv_struct *pnvc = &nvc;
	nv->GetNVStruct(&nvc);

/*????  if (nvc.nv_addr_index != 15) { // gebunden ?
		DBG(DBG_LHA, " Send unbound NV?");
		return (false);
	}
*/
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	ExpAppBuf hdr;
    memcpy(&hdr, gLdv_hostnvhdr, sizeof(gLdv_hostnvhdr));

    // Transaction queue
    hdr.ni_hdr.q.queue  = gNi_queue[pnvc->nv_service == UNACKD ? 1 : 0]
                                   [pnvc->nv_priority];

    hdr.ni_hdr.q.length += len;

    hdr.msg_hdr.exp.st   = pnvc->nv_service ;
    hdr.msg_hdr.exp.auth = pnvc->nv_auth ;
    hdr.msg_hdr.exp.tag  = pnvc->nv_addr_index ;
    hdr.msg_hdr.exp.length += len ;

    hdr._data.unv.direction      = ~pnvc->nv_direction ;
    hdr._data.unv.NV_selector_hi = pnvc->nv_selector_hi ;
    hdr._data.unv.NV_selector_lo = pnvc->nv_selector_lo ;
// copy data into member variable and into message
	nv->SetValue(data);
	memcpy(&hdr.data.data[1], data, len);
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
////ComplType cpl = L_LdvWrite(&hdr, HEAD_CODE + len);
	ComplType cpl = L_LdvSndRcv(&hdr, HEAD_CODE + len +1, NULL, NULL);
	if (cpl != MSG_SUCCEEDS) {
		return (DBGERR("Send NV failed"));
	}
	return (true);
}

//***************************************************************************

static UCHAR gLdv_resphdr[] = {
/*-----------------------NiHdr----------------------------------------------*/
 	0x10 | 0x04              , /*  1: niCOMM + niNTQ                        */
    15	                     , /*  2: preset Laenge ohne Daten     variabel */
/*-----------------------MsgHdr---------------------------------------------*/
	0x00 | 0x60 | 0x00       , /*  3: ExpMsg + Request + Tag       variabel */
    0x00 | 0x01              , /*  4: impl. Addr + Response                 */
    1	                     , /*  5: preset Nettolaenge (Codebyte) variabel */
/*-----------------------Addr-----------------------------------------------*/
    SUBNET_NODE              , /*  6: Addr-Struktur                         */
    0x00 | 0x00              , /*  7: Domain + Node                variabel */
    0x40 | 0x03              , /*  8: rpt-timer + retry                     */
    0x00 | 0x04              , /*  9: rsvd2 + tx-Timer                      */
    0x00                     , /* 10: Subnet                       variabel */
    0x00                     , /* 11: NID_0                                 */
    0x00                     , /* 12: NID_1                                 */
	0x00                     , /* 13: NID_2                                 */
    0x00                     , /* 14: NID_3                                 */
    0x00                     , /* 15: NID_4                                 */
    0x00                     , /* 16: NID_5                                 */
/*-----------------------Code/Data------------------------------------------*/
    0x00                     , /* 17: Code                         variabel */
    0x00                     , /* 18: data[0]                      variabel */
} ;

static void ldv_SetRespHdr(ExpAppBuf* hdr, ExpAppBuf* reqbuf, UCHAR _code,
                           short len)
{
    memcpy(hdr, gLdv_resphdr, sizeof(gLdv_resphdr));

    hdr->msg_hdr.exp.tag    = reqbuf->msg_hdr.exp.tag;		// use same tag

    hdr->addr.snd.sn.domain = reqbuf->addr.rcv.domain;
    hdr->addr.snd.sn.node   = reqbuf->addr.rcv.source.node;
    hdr->addr.snd.sn.subnet = reqbuf->addr.rcv.source.subnet;

    hdr->_data.exp._code    = _code;

    hdr->ni_hdr.q.length    += len;
    hdr->msg_hdr.exp.length += len;
}

//---------------------------------------------------------------------------

static bool ldv_RespOut(ExpAppBuf* reqbuf, UCHAR _code, UCHAR* _data, short len)
{
    ExpAppBuf	mybuf;

    ldv_SetRespHdr(&mybuf, reqbuf, _code, len);

	if (len > 0) memcpy(&mybuf.data.data[0], _data, len);

	ComplType cpl = L_LdvWrite(&mybuf, HEAD_CODE + len);
	if (cpl == MSG_SUCCEEDS) return (true);

	return (false);
}

//---------------------------------------------------------------------------

static UCHAR gLdv_lochdr[] = {
/*-----------------------NiHdr----------------------------------------------*/
 	0x20 | 0x02              , /*  1: niNETMGMT + niTQ                      */
    15	                     , /*  2: preset Laenge ohne Daten     variabel */
/*-----------------------MsgHdr---------------------------------------------*/
	0x00 | 0x60 | 0x00       , /*  3: ExpMsg + Request + Tag       variabel */
    0x00                     , /*  4: impl. Adressierung                    */
    1	                     , /*  5: preset Nettolaenge (Codebyte) variabel */
/*-----------------------Addr-----------------------------------------------*/
    LOCAL                    , /*  6: Addr-Struktur                         */
    0x00                     , /*  7:                                       */
    0x00                     , /*  8:                                       */
    0x00                     , /*  9:                                       */
    0x00                     , /* 10:                                       */
    0x00                     , /* 11:                                       */
    0x00                     , /* 12:                                       */
	0x00                     , /* 13:                                       */
    0x00                     , /* 14:                                       */
    0x00                     , /* 15:                                       */
    0x00                     , /* 16:                                       */
/*-----------------------Code/Data------------------------------------------*/
    0x00                     , /* 17: Code                         variabel */
    0x00                     , /* 18: data[0]                      variabel */
} ;

static void ldv_SetLocHdr(ExpAppBuf* hdr, UCHAR _code, short len)
{
    memcpy(hdr, gLdv_lochdr, sizeof(gLdv_lochdr));

    hdr->_data.exp._code	 = _code;
    hdr->ni_hdr.q.length    += len;
    hdr->msg_hdr.exp.length += len;
}

//---------------------------------------------------------------------------

bool ldv_LocalOut(UCHAR _code, UCHAR* _data, short len)
{
	short		rcvLen;
    ExpAppBuf	mybuf, rcvBuf;

    ldv_SetLocHdr(&mybuf, _code, len);

	if (len > 0) memcpy(&mybuf.data.data[0], _data, len +1);

	ComplType cpl = L_LdvSndRcv(&mybuf, len +17, &rcvBuf, &rcvLen);
	if (cpl == MSG_SUCCEEDS) {
		return (true);
	}

	return (false);
}

//---------------------------------------------------------------------------

bool ldv_LocalReq(UCHAR _code, UCHAR* _data, short len, void* indata)
{
	DBG(DBG_LHA, "ldv_LocalReq");
	short		rcvLen;
    ExpAppBuf	mybuf, rcvBuf;

    ldv_SetLocHdr(&mybuf, _code, len);

	if (len > 0) memcpy(&mybuf.data.data[0], _data, len);

//	DBG(0, "+L_LdvSndRcv");
	ComplType cpl = L_LdvSndRcv(&mybuf, len +17, &rcvBuf, &rcvLen);
//	DBG(0, "-L_LdvSndRcv");
	if (cpl == MSG_SUCCEEDS) {
		memcpy(indata, rcvBuf._data.exp._data, rcvBuf.msg_hdr.exp.length - 1);
		return (true);
	}

	return (DBGERR("ldv_LocalReq failed"));
}

//***************************************************************************
