//////////////////////////////////////////////////////////////////////
//
// XIF.cpp:  XIF file decoding
//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
// Copyright (c) 2000  Gesytec GmbH
//
// Date:    12. 1.00
//
// Author: mb, mboehlke@gesytec.de
//
//
//////////////////////////////////////////////////////////////////////

#include "windows_compat.h"
#include "XIF.h"
#include "DefFile.h"
#include "xDbg.h"
#include "Utils.h"

// Workaround (sonst Compilerfehler wegen Macro-Redefine)
#undef __NFDBITS
#undef __FDMASK

#include <stdlib.h>
#include <stdio.h>
#include <string.h>


static char *aSS(char *str, char *sub);	// after SubString

//***************************************************************************

CXIF::CXIF()
{
	mSig[0]	   = 0;
	mSig[1]	   = 0;
	mDef	   = NULL;
	mSnvtStart = NULL;
	mSnvtPut   = NULL;
	mSnvtMax   = NULL;
}

//---------------------------------------------------------------------------

CXIF::~CXIF(void)
{
	if (mDef) delete mDef;
}

//---------------------------------------------------------------------------

bool CXIF::OpenXIF(void)
{
	DBG(DBG_XIF, "+OpenXIF");
	mDef = new CDefFile;
	bool ok = mDef->Open("Lolp072f.xif");	// try to open
	if (!ok)
	{
		delete mDef, mDef = NULL;
		DBG(DBG_XIF, " OpenXIF failed");
		return (false);
	}

	mStart = mDef->mScan;						// save scan pointer
	DBG(DBG_XIF, "-OpenXIF");
	return (ok);
}

//---------------------------------------------------------------------------

bool CXIF::ReadHead(short pass)
{
	DBG(DBG_XIF, "+ReadHead");
	short version	= 0;
	mSNVTLen		= 6;
	mDef->mScan		= mStart;					// reset scan pointer
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Version of XIF file ...
	if (!ReadLine()) return (false);			// line 1 of header
DBG(DBG_XIF, mLine);
	char *s = aSS(mLine, "Version ");
	if (!s) return (false);
	if (!strnieq(s, "3.1", 3)) return (false);
	version = 3;
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (!ReadLine()) return (false);			// line 2 of header
	if (!ReadLine()) return (false);			// line 3 of header
	if (!ReadLine()) return (false);			// line 4 of header
	if (*mLine) {
		sscanf(mLine, "%2x%2x", &mSig[0], &mSig[1]);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// ProgramID ...
	if (!ReadLine()) return (false);			// line 5 of header
	sscanf(mLine, "%2x:%2x:%2x:%2x:%2x:%2x:%2x:%2x",
		&mPid[0], &mPid[1], &mPid[2], &mPid[3],
		&mPid[4], &mPid[5], &mPid[6], &mPid[7]);
	DBG(DBG_XIF, "PID ", mPid, 8);
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// domains, adr entries, msgs, nvs, tags, ...
	if (!ReadLine()) return (false);			// line 6 of header
//DBG(DBG_XIF, mLine);
	int ndoms, nadrs, msgs;
	sscanf(mLine, "%i %i %i %i %i",
		&ndoms, &nadrs, &msgs, &mNvs, &mTags);
// following fields not decoded (mb)	//??? XIF
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (!ReadLine()) return (false);			// line 7 of header
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	//if (version = 3) {	// hier ist doch wohl (version==3) gemeint ?!?
	if (version == 3) {
		if (!ReadLine()) return (false);		// line  8 of header
		if (!ReadLine()) return (false);		// line  9 of header
		if (!ReadLine()) return (false);		// line 10 of header
	}
	else return (false);
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (!ReadLine()) return (false);			// line 11 of header
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (!ReadLine()) return (false);			// line 12 self identification
DBG(DBG_XIF, mLine);
	if (mLine[0] != '"') {
		mSNVTLen++;
		if (pass == 3) snvtAddByte(0);			// no node self doc
	}
	else {
		mSNVTLen += strlen(mLine); // we count the " and not the zero
		if (pass == 3) snvtAddString(&mLine[1]);
		while (true) {
			if (!ReadLine()) return (false);
DBG(DBG_XIF, mLine);
			if (mLine[0] != '"') break;			// end of self identification
			mSNVTLen += strlen(&mLine[1]);
			if (pass == 3) {
				snvtRemByte();					// remove trailing zero
				snvtAddString(&mLine[1]);
			}
		}
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	DBG(DBG_XIF, "-ReadHead");
	return (true);
}

//---------------------------------------------------------------------------

bool CXIF::ReadNV(short pass, short nvind)
{
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// read first line of netvar ...
	if (!ReadLine()) return (false);
//if (pass == 1) DBG(DBG_XIF, mLine);
	if (!strnieq(mLine, "VAR ", 4)) return (false);		// must start with VAR

	mNVLength		= 0;
	mNVDesc.ext_rec = 1;								// we always add ext's
	cpytoken(mNVName, sizeof(mNVName), mLine, 1, ' ');	// name of nv
	mNVDoc[0]		= 0;
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// read connection parms ...
	if (!ReadLine()) return (false);

	mNVDesc.nv_offline	= snvtDescBit(1);				// offline
	mNVDir				= snvtDescBit(4);				// direction
	mNVService			= getShort(5);					// service type
// 0 1 63 0 0 1 0 1 0 1 0 0 1
	mNVDesc.nv_service_type_config	= snvtDescBit(6);	// service config
	mNVDesc.nv_auth_config			= snvtDescBit(8);	// auth config
	mNVDesc.nv_priority_config		= snvtDescBit(10);	// priority config
	mNVDesc.nv_polled				= snvtDescBit(11);	// polled
	mNVDesc.nv_sync					= snvtDescBit(12);	// sync
	mNVDesc.nv_config_class			= snvtDescBit(13);	// config class
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// read self doc ...
	if (!ReadLine()) return (false);

	STER	ster;						// extension flags
	ster.free = 0;
	ster.nc   = 0;
	ster.sd	  = 0;						// selfdoc
	ster.nm	  = 1;						// nv name
	ster.re	  = 0;
	ster.mre  = 0;
	BYTE *p = (BYTE *)&ster;

	if (mLine[0] != '"') {				// no selfdoc
		mSNVTLen += strlen(mNVName) +2;
		if (pass == 3) {
			snvtAddByte(*p);
			snvtAddString(mNVName);
		}
	}
	else {	// we have selfdoc
		cpytoken(mNVDoc, sizeof(mNVDoc), &mLine[1], 0, 0);	// selfdoc (line 1 only !)
		ster.sd	= 1;					// selfdoc
		mSNVTLen += strlen(mNVName) +2;
		if (pass == 3) {
			snvtAddByte(*p);
			snvtAddString(mNVName);
		}
		mSNVTLen += strlen(mLine);
		if (pass == 3) snvtAddString(&mLine[1]);
		while (true) {
			char *last = mDef->mScan;
			if (!ReadLine()) return (false);
			if (mLine[0] != '"') {					// end of self identification
				mDef->mScan = last;
				break;
			}
			mSNVTLen += strlen(&mLine[1]);
			if (pass == 3) {
				snvtRemByte();						// remove trailing zero
				snvtAddString(&mLine[1]);
			}
		}
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (!ReadLine()) return (false);					// nv characteristic
	mNVDesc.snvt_type_index = (BYTE)getShort(1);		// snvt type
	mSNVTLen += 2;
	if (pass == 2) {
		BYTE *p = (BYTE *)&mNVDesc;						// write descriptor
		snvtAddByte(*p++);
		snvtAddByte(*p++);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	while (true) {
		char *last = mDef->mScan;
		if (!ReadLine()) break;
		if (strnieq(mLine, "VAR ", 4)) { mDef->mScan = last; break; }

		short t = getShort(1);
		if		(t == 0) mNVLength++;		// char
		else if (t == 1) mNVLength++;		// int (Neuron)
		else if (t == 2) mNVLength += 2;	// long (Neuron)
		else if (t == 3) {					// bitfield
			t = getShort(2);
			if (t == 0) mNVLength++;	// new one
		}
		else if (t == 4) {
			DBGERR("XIF ReadNV found data type 4");
			return (false);
		}
		else if (t == 5) {
			DBGERR("XIF ReadNV found data type 5");
			return (false);
		}
		else {
			DBGERR("XIF ReadNV found data type ?");
			return (false);
		}
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	return (true);
}

//---------------------------------------------------------------------------

bool CXIF::ReadLine(void)
{
	if (mDef->EoF()) return (false);
	bool ok = mDef->getLine(mLine, sizeof(mLine));
	if (ok) mDef->skipLine();
// 	if (ok) DBG(DBG_XIFLINE, mLine);
	return (ok);
}

//---------------------------------------------------------------------------

void CXIF::SetSnvtP(BYTE *psnvt, WORD max)
{
	mSnvtStart = psnvt;
	mSnvtPut   = psnvt;
	mSnvtMax   = &psnvt[max];

	*mSnvtPut++ = 0;						// Laenge gesamt high byte
	*mSnvtPut++ = 2;						// Laenge gesamt low  byte

    snvtAddByte((BYTE)mNvs);				// Anzahl NV's gesamt low byte
    snvtAddByte(1);							// Version 1
    snvtAddByte((BYTE)(mNvs >> 8));			// Anzahl NV's gesamt high byte
    snvtAddByte((BYTE)mTags);				// bindable message tags
}

//---------------------------------------------------------------------------

bool CXIF::snvtAddByte(BYTE b)
{
	BYTE *psnvt = mSnvtStart;
	if (!psnvt) return (false);

	psnvt[1]++;
	if (!psnvt[1]) psnvt[0]++;

	*mSnvtPut++ = b;
	return (true);
}

//---------------------------------------------------------------------------

void CXIF::snvtRemByte(void)
{
	BYTE *psnvt = mSnvtStart;
	if (!psnvt) return;

	psnvt[1]--;
	if (psnvt[1] == 0xFF) psnvt[0]--;
}

//---------------------------------------------------------------------------

static char *NonPrintable(char *s, char *dp)
{
	BYTE b = 0;

	s++;	if (!*s)	   return (s);
			if (*s != 'x') return (s);

	s++;	if (!*s)	   return (s);
	b = nib2bin(*s) << 4;

	s++;	if (!*s)	   return (s);
	b += nib2bin(*s);

	*dp = b;
	return (s);
}

static void NonPrintables(char *s)
{	// XIF not printable \xHH
	char *dp = s;
	while (*s) {
		if (*s != '\\') *dp++ = *s++;
		else			s = NonPrintable(s, dp++);
	}
	*dp = 0;
}

bool CXIF::snvtAddString(char *s)
{
	NonPrintables(s);
	while (*s) snvtAddByte((BYTE)*s++);
	return (snvtAddByte(0));
}

//---------------------------------------------------------------------------

int CXIF::snvtDescBit(short nr)
{
	char buf[8];
	cpytoken(buf, sizeof(buf), mLine, nr -1, ' ');
	return ((buf[0] == '1') ? 1 : 0);
}

//---------------------------------------------------------------------------

short CXIF::getShort(short nr)
{
	char buf[16];
	cpytoken(buf, sizeof(buf), mLine, nr -1, ' ');
	return (atoi(buf));
}

//***************************************************************************

static char *aSS(char *str, char *sub)	// after SubString
{
	short len = strlen(sub);
	while (*str) {
		if (strnieq(str, sub, len)) return (&str[len]);
		str++;
	}
	return (NULL);
}

//***************************************************************************

