#include "windows_compat.h"
#include "fifo.h"
#include <string.h>

//static	bool				gInitialized	= false;	// lock	 inited
//static	CRITICAL_SECTION	gLock;						// lock access

//***************************************************************************

class CFifoCrit
{
public:
			 	CFifoCrit	(CRITICAL_SECTION *hdl);
	virtual 	~CFifoCrit	();

protected:
	CRITICAL_SECTION	*mHdl;
};

//---------------------------------------------------------------------------

CFifoCrit::CFifoCrit(CRITICAL_SECTION *hdl)
{
	mHdl = hdl;
	EnterCriticalSection(mHdl);
}

//---------------------------------------------------------------------------

CFifoCrit::~CFifoCrit()
{
	LeaveCriticalSection(mHdl);
}

//***************************************************************************

CFifo::CFifo(WORD size)
{
	InitializeCriticalSection(&mLock);
	mFifo	= (BYTE *)new BYTE[size];
	mPut	= mFifo;
	mGet	= mFifo;
	mEnd	= mFifo + size;
}

//---------------------------------------------------------------------------

CFifo::~CFifo(void)
{
	if (mFifo) delete mFifo;
	DeleteCriticalSection(&mLock);
}

//---------------------------------------------------------------------------

bool CFifo::Put(BYTE *data, WORD len)
{
	CFifoCrit lock(&mLock);
	if (len == 0) return (false);
	BYTE *put = mPut;
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// check if we can copy continously ...
	if (mGet > put) {					// max free is get pointer
		if ((WORD)(mGet - put) <= (len + sizeof(len))) return (false);

		*(WORD *)put = len;			put += sizeof(WORD);
		memcpy(put, data, len);		put += len;

		mPut = put;
		return (true);
	}
// (put >= get) max free is end of fifo
	else if ((WORD)(mEnd - put) > (len + sizeof(len))) {	// enough space ?

		*(WORD *)put  = len;		put += sizeof(WORD);
		memcpy(put, data, len);		put += len;

		mPut = put;
		return (true);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// we have to copy beyond end of fifo ...
	put = addFifo(mPut, (BYTE *)&len, sizeof(len));
	if (!put) return (false);

	put = addFifo(put, data, len);
	if (!put) return (false);

	mPut = put;
	return (true);
}

//---------------------------------------------------------------------------

WORD CFifo::GetLen(void)
{
	CFifoCrit lock(&mLock);
	BYTE	*get;
	WORD	len;

	get = fromFifo(mGet, (BYTE *)&len, sizeof(len));
	if (!get) return (0);						// fifo is empty
	return (len);
}

//---------------------------------------------------------------------------

WORD CFifo::Get(BYTE *data)
{
	CFifoCrit lock(&mLock);
	BYTE	*get;
	WORD	len;

	get = fromFifo(mGet, (BYTE *)&len, sizeof(len));
	if (!get) return (0);						// fifo is empty

	mGet = fromFifo(get, data, len);
	return (len);
}

//---------------------------------------------------------------------------
// copy data to Fifo ...
BYTE *CFifo::addFifo(BYTE *put, BYTE *data, WORD len)
{
	while (len--) {
		*put++ = *data++;
		if (put >= mEnd) put = mFifo;
		if (put == mGet) return (NULL);		// no space
	}
	return (put);
}

//---------------------------------------------------------------------------
// copy from Fifo to data ...
BYTE *CFifo::fromFifo(BYTE *get, BYTE *data, WORD len)
{
	if (get == mPut) return (NULL);				// fifo is empty

	while (len--) {
	 	*data++ = *get++;
		if (get >= mEnd) get = mFifo;
	}
	return (get);
}

//***************************************************************************
