//////////////////////////////////////////////////////////////////////
//
// g2LonC.cpp:  implementation of a simple LonWorks API
//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
// Copyright (c) 1999  Gesytec GmbH
//
// Date:    2. 2.99	
//
// Author: mb, mboehlke@gesytec.de 
//
//
//////////////////////////////////////////////////////////////////////

#include "windows_compat.h"
#include "ni_privat.h"
#include "g2Lon.h"
#include "LHA.h"
#include "xDbg.h"
#include "fifo.h"
#include "Utils.h"
#include "DebugMsg.h"

// Workaround (sonst Compilerfehler wegen Macro-Redefine)
#undef __NFDBITS
#undef __FDMASK

#include <fcntl.h>		// fr O_WRONLY, open()
#include <stdio.h>		// fr stderr, fprintf(), fopen(), fget(), popen(), perror(), ...
#include <unistd.h>		// fr usleep(), fork(), setsid() und chdir()
#include <string.h>

// fr OpenRemoteNet()
#include <iostream>
#include <string>
#include <stdlib.h>
#include <unistd.h>

#include <stdio.h>
#include <netdb.h>

#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <errno.h>


		char		*VERSION	= "g2LON V2.12t3";

#define NODENV(x) ((NVSTRUCT *)(node->panvs + (x * sizeof(NVSTRUCT))))

extern	NIDSTRUCT	*gNIDs;
extern	short		gNIDcount;
extern	BYTE		gRPTRETRY;
extern	BYTE		gTXTIMER;

static	CRITICAL_SECTION	gLock;
static	BYTE				gNull[8] = { 0, 0, 0, 0,  0, 0, 0, 0 };
		CFifo		*gNVUpds = NULL;
		CFifo		*gMsgs	 = NULL;

//***************************************************************************

CNet::CNet(void)
{
	InitializeCriticalSection(&gLock);
	mLP4 = INVALID_HANDLE_VALUE;
}

//---------------------------------------------------------------------------

CNet::~CNet(void)
{
	DeleteAll();								// delete CNeuron objects
	L_CloseLoad();								// remove intermediate data
	DeleteCriticalSection(&gLock);
}

//---------------------------------------------------------------------------

char *CNet::Version(void)						// get Version of g2LON
{
	return (VERSION);
}

//---------------------------------------------------------------------------

bool CNet::Open(DWORD dwIndex)
{
// open LON driver ...

	char szDevName[10];
	sprintf(szDevName, "/dev/lon%i", dwIndex);
    return Open(szDevName);
}
	
bool CNet::Open(const char *DevName)
{
  	mLP4 = open(DevName, O_RDWR);
  	if (mLP4 < 0)
    {
		char errmsg[30];
		sprintf(errmsg, "opening %s failed", DevName);
        perror(errmsg);
        return false;
    }

/*	TCHAR szDevName[6];
	
	wsprintf(szDevName, TEXT("LON%d:"), dwIndex);
	mLP4 = CreateFile(szDevName, GENERIC_READ|GENERIC_WRITE, 0, NULL, 
					  OPEN_EXISTING, 0, NULL);
*/


	//Sleep(1000);								// wait for Neuron Reset
	usleep(1000*1000);							// wait for Neuron Reset


/*	if (mLP4 == INVALID_HANDLE_VALUE) {
//		RETAILMSG(1, (TEXT("Couldn't CreateFile(%s)\r\n"), szDevName));
		return (false);
	} 
//	RETAILMSG(1,(TEXT("*****Open Driver ok\r\n")));
*/
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	L_Open(mLP4, FALSE);							// initialize
//	RETAILMSG(1,(TEXT("******Initialize ok\r\n")));
	if (!L_CheckLocal())
	{
//		RETAILMSG(1,(TEXT("******Error Local node\r\n")));
		return (false);	// check local node 
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//	RETAILMSG(1,(TEXT("*******UpldLocal\r\n")));
	CNeuron	*neuron = UpldLocal();			// get local data
	if (!neuron) return (false);			// upload local failed
//	RETAILMSG(1,(TEXT("*****upldLocal ok\r\n")));
	return (true);
}

//---------------------------------------------------------------------------

bool CNet::ReOpenRemoteNet()
{
// connect to LON server ...

  	struct sockaddr_in servername;

 	int sock = socket (PF_INET, SOCK_STREAM, 0);
  	if (sock < 0)
    {
      perror ("socket (client)");
      return false;
    }

	struct hostent *hostinfo;
	servername.sin_family = AF_INET;
	servername.sin_port = htons (49152);
	hostinfo = gethostbyname (RemoteAdr);
	if (hostinfo == NULL) 
	{
	    fprintf (stderr, "Unknown host %s.\n", RemoteAdr);
	    return false;
	}
	servername.sin_addr = *(struct in_addr *) hostinfo->h_addr;
	
	if (0 > connect (sock,
	                 (struct sockaddr *) &servername,
	                 sizeof (servername)))
	{
	  perror ("connect (client)");
	  return false;
	}

    fprintf (stderr, "connect(%s)=OK\n", RemoteAdr);

	mLP4 = sock;
	L_ReOpen(mLP4, TRUE);							// initialize
	return (true);
}


bool CNet::OpenRemoteNet(const char* adr)
{
// connect to LON server ...

  	struct sockaddr_in servername;

 	int sock = socket (PF_INET, SOCK_STREAM, 0);
  	if (sock < 0)
    {
      perror ("socket (client)");
      return false;
    }

	struct hostent *hostinfo;
	servername.sin_family = AF_INET;
	servername.sin_port = htons (49152);
	hostinfo = gethostbyname (adr);
	if (hostinfo == NULL) 
	{
	    fprintf (stderr, "Unknown host %s.\n", adr);
	    return false;
	}
	servername.sin_addr = *(struct in_addr *) hostinfo->h_addr;
	
	if (0 > connect (sock,
	                 (struct sockaddr *) &servername,
	                 sizeof (servername)))
	{
	  perror ("connect (client)");
	  return false;
	}

    fprintf (stderr, "connect(%s)=OK\n", adr);

	mLP4 = sock;
    RemoteAdr = strdup(adr);
/*
  	mLP4 = open(szDevName, O_RDWR);
  	if (mLP4 < 0)
    {
		char errmsg[30];
		sprintf(errmsg, "opening /dev/lon%i failed", dwIndex);
        perror(errmsg);
        return false;
    }
*/

//	usleep(1000*1000);							// wait for Neuron Reset

	L_Open(mLP4, TRUE);							// initialize

	if (!L_CheckLocal())
	{
//		RETAILMSG(1,(TEXT("******Error Local node\r\n")));
		return (false);	// check local node 
	}

	CNeuron	*neuron = UpldLocal();			// get local data
	if (!neuron) return (false);			// upload local failed
	return (true);
}

//---------------------------------------------------------------------------

void CNet::Close(void)
{
	//if (mLP4 != INVALID_HANDLE_VALUE) CloseHandle(mLP4);
	if (mLP4 != INVALID_HANDLE_VALUE) close(mLP4);
	mLP4 = INVALID_HANDLE_VALUE;
}

//---------------------------------------------------------------------------

bool CNet::SendSvp(void)
{
	CNeuron *neuron = NeuronByInd(0);
	if (!neuron) return (false);

	return (L_SendSvp(neuron->NID(), neuron->PID()));
}

//---------------------------------------------------------------------------

BYTE CNet::SetRcvTrx(BYTE rcvtrx)
{
	BYTE prev = gTXTIMER;
	gTXTIMER  = rcvtrx;
	return (prev);
}

BYTE CNet::SetRptTry(BYTE rpttry)
{
	BYTE prev = gRPTRETRY;
	gRPTRETRY = rpttry;
	return (prev);
}

//---------------------------------------------------------------------------

short CNet::Scan(BYTE dom_idx)
{
	CLockCrit lock(&gLock);

	//if ((dom_idx < 0) || (dom_idx > 1)) return (0);

	// BYTE = unsigned 8bit --> wie soll das kleiner 0 werden ?

	if (dom_idx != 0) return (0);	// vielleicht war das so gemeint.. (?)

	L_DoScan(dom_idx);
	return (gNIDcount);
}


//---------------------------------------------------------------------------

bool CNet::UploadAll(bool reload)				// ??? reload
{
	CLockCrit lock(&gLock, ALLOW_2ND_LOCK);
	NIDSTRUCT *nid = gNIDs;
	CNeuron	  *neuron;

// loads neurons from gNIDs list (not from own list) ???
	while (nid) {
		neuron = NeuronByNID(nid->nidpid);		// check if already in list
		if (!neuron) {
			neuron = Upload(nid->nidpid);
			if (neuron) AddNeuron(neuron);		// add neuron to list
		}
		nid = nid->next;
	}
	return (true);
}

//---------------------------------------------------------------------------

CNeuron	*CNet::UpldLocal(void)
{
	NODE *node = new NODE;
	memset(node, 0, sizeof(NODE));
	if (!L_UploadLocal(node)) { delete node; return (NULL); }

	CHANeuron *neuron = new CHANeuron(node);
	neuron->mIsLocal  = true;
	AddTail(neuron);						// local node is first in list

	neuron->ChkBinding();					// check if LONHOST.xif is there
	return (neuron);
}

//---------------------------------------------------------------------------

CNeuron *CNet::Upload(BYTE *nidpid)
{
	CLockCrit lock(&gLock);
	CNeuron   *neuron = NeuronByNID(nidpid);
	if (neuron) {							// neuron is already known
		if (neuron->IsHost()) return (neuron);	// dont reload the host

		if (neuron->Upload(nidpid)) return (neuron);	// reload 
		else						return (NULL);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	neuron = new CNeuron();					// create new neuron
	if (neuron->Upload(nidpid)) {			// if upload ok
		return (neuron);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	delete neuron;							//	otherwise delete it
	return (NULL);
}

//---------------------------------------------------------------------------

bool CNet::AddNeuron(CNeuron *neuron)
{
	CLockCrit lock(&gLock);
// search for double ???
	AddTail(neuron);
	return (true);
}
/*
	CNeuron *np;
	for (short npos = 0; npos < gNet->GetCount(); npos++) {
		np = (CNeuron *)gNet->GetAt(npos);

		if ((neuron->SubnetNr() != np->SubnetNr()) || 
			(neuron->NodeNr() != np->NodeNr())) continue;

		if (np->NODEPTR() == NULL) {
			gNet->RemoveAt(gNet->Find(np));
			delete np;
			break;
		}
	}
*/
//---------------------------------------------------------------------------

bool CNet::DelNeuron(CNeuron *neuron)
{
	CLockCrit lock(&gLock);
	RemoveAt(Find(neuron));
	delete neuron;
	return (true);
}

//---------------------------------------------------------------------------

CNeuron *CNet::NeuronByNID(BYTE *nid)
{
	CNeuron *neuron;
//	by IDel: NID == 0 --> return own Neuron  
	if (nid == 0) return NeuronByInd(0);
	for (short pos = 0; pos < GetCount(); pos++) {
		neuron = (CNeuron *)GetAt(pos);
		if (!neuron->mNODE) continue;
		if (memcmp(neuron->NID(), nid, 6) == 0) return (neuron);
	}
	return (NULL);
}

//---------------------------------------------------------------------------

CNeuron *CNet::NeuronByInd(WORD index)
{
	return ((CNeuron *)GetAt(index));
}

//---------------------------------------------------------------------------

CNeuron *CNet::NeuronByAdr(BYTE dlen, BYTE *dom, BYTE net, BYTE nodenr)
{
	CNeuron *neuron;
	for (short npos = 0; npos < GetCount(); npos++) {
		neuron = (CNeuron *)GetAt(npos);
		if (neuron->AdrEq(dlen, dom, net, nodenr) >= 0) return (neuron);
	}
	return (NULL);
}

//---------------------------------------------------------------------------

bool CNet::Status(BYTE *nid, BYTE *status)
{
	return (L_QueryStatus(nid, status));
}

//---------------------------------------------------------------------------

short CNet::ReadMem(BYTE didx, BYTE subnet, BYTE nodenr, 
					WORD adr, short size, BYTE mode, BYTE *buf)
{
    return (L_ReadMem(didx, subnet, nodenr, adr, size, mode, buf));
}

//---------------------------------------------------------------------------

bool CNet::WriteMem(BYTE didx, BYTE subnet, BYTE nodenr,
					 WORD adr, short size, BYTE mode, BYTE *buf)
{
	return (L_WriteMem(didx, subnet, nodenr, adr, size, mode, buf));
}

//---------------------------------------------------------------------------

bool CNet::SetNodeMode(BYTE didx, BYTE subnet, BYTE nodenr, 
					   BYTE mode, BYTE state)
{
	return (L_SetNodeMode(didx, subnet, nodenr, mode, state));
}

//---------------------------------------------------------------------------

bool CNet::Wink(BYTE didx, BYTE subnet, BYTE nodenr)
{
	return (L_Wink(didx, subnet, nodenr));
}

//---------------------------------------------------------------------------

bool CNet::QueryStatus(BYTE didx, BYTE subnet, BYTE nodenr, BYTE *status)
{
	return (L_QueryStatus(didx, subnet, nodenr, status));
}

//---------------------------------------------------------------------------

CNV *CNet::NVwasUpdated(void)
{
	if (!gNVUpds) return (NULL);

	CNV *nv;
	if (gNVUpds->Get((BYTE *)&nv)) return (nv);
	return (NULL);
}

//---------------------------------------------------------------------------

short CNet::MsgReceived(BYTE *msg)
{
	if (!gMsgs) return (0);
	return (gMsgs->Get(msg));
}

//---------------------------------------------------------------------------

WORD CNet::SetDbgMode(WORD mode)
{
	return (DBGSET(mode));
}

//***************************************************************************

CNeuron::CNeuron()
{
//DBGERR("CNeuron::CNeuron");
	mUserHdl  = 0;
	mNODE     = NULL;
	mDomIdx   = 0;
	mSubnetNr = 0;
	mNodeNr	  = 0;
	mIsLocal  = false;
}

//---------------------------------------------------------------------------

CNeuron::CNeuron(NODE *node)
{
//DBGERR("CNeuron::CNeuron(NODE *node)");
//if (!node) DBGERR("CNeuron node == 0");
	mUserHdl  = 0;
	mNODE     = node;
	mDomIdx   = 0;
	mSubnetNr = 0;
	mNodeNr	  = 0;
	mIsLocal  = false;

//	CreateNVs(); ??? fr normal mode ntig ?
}

//---------------------------------------------------------------------------

CNeuron::CNeuron(BYTE didx, BYTE subnet, BYTE node)
{
//DBGERR("CNeuron::CNeuron(BYTE didx, BYTE subnet, BYTE node)");
	mUserHdl  = 0;
	mNODE     = NULL;
	mIsLocal  = false;
	mDomIdx   = didx;
	mSubnetNr = subnet;
	mNodeNr   = node;
}

//---------------------------------------------------------------------------

CNeuron::~CNeuron(void)
{
	m_nvs.DeleteAll();
	DelNodePtr();
}

//---------------------------------------------------------------------------

short CNeuron::CreateNVs(void)
{
	if (!mNODE)		   return (0);
	if (!mNODE->panvs) return (0);

	CNV	*nv;
	for (short i = 0; i < mNODE->num_nvs; i++) {	// create and add CNVs
		nv = new CNV(this, i);
		m_nvs.AddTail(nv);
	}

	delete mNODE->panvs;
	mNODE->panvs = NULL;

	return (mNODE->num_nvs);
}

//---------------------------------------------------------------------------

bool CNeuron::Upload(BYTE *nidpid)
{
	if (IsHost()) return (false);				// is host neuron
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (mNODE) DelNodePtr();
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	mNODE = new NODE;
	memset(mNODE, 0, sizeof(NODE));				// upload relys on zeros
	memcpy(mNODE->nid, nidpid, 14);				// nid and pid !!!

	if (!L_UploadNode(mNODE)) {
		//RETAILMSG(1, (TEXT("CNeuron::Upload failed\r\n")));
		perror("CNeuron::Upload failed\r\n");
		DelNodePtr();
		return (false);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	CreateNVs();
	return (true);
}

//---------------------------------------------------------------------------

bool CNeuron::Upload(BYTE didx, BYTE subnet, BYTE node)
{
	if (IsHost()) return (false);					// is host neuron
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	BYTE nidpid[14];
	if (!L_ReadMem(didx, subnet, node, 0, 6, RM_RDONLY, nidpid)) {
		//RETAILMSG(1, (TEXT("CNeuron::read NID failed\r\n")));
		perror("CNeuron::read NID failed\r\n");
		return (false);
	}
	return (Upload(nidpid));
}

//---------------------------------------------------------------------------

void CNeuron::DelNodePtr(void)
{
	if (!mNODE) return;
	if (mNODE->panvs) delete mNODE->panvs;
	delete mNODE;
	mNODE = NULL;
}

//---------------------------------------------------------------------------

BYTE *CNeuron::NID(void)
{
	if (!mNODE) return (gNull);
	return (mNODE->nid);
}

//---------------------------------------------------------------------------

BYTE *CNeuron::PID(void)
{
	if (!mNODE) return (gNull);
	return (mNODE->pid);
}

//---------------------------------------------------------------------------

BYTE *CNeuron::Location(void)
{
	if (!mNODE) return (gNull);
	return (mNODE->location);
}

//---------------------------------------------------------------------------

BYTE CNeuron::SubnetNr(void)
{
	if (mNODE) return (mNODE->domain->subnet);
	return (mSubnetNr);
}

//---------------------------------------------------------------------------

BYTE CNeuron::NodeNr(void)
{
	if (mNODE) return (mNODE->domain->node & 0x7F);
	return (mNodeNr);
}

//---------------------------------------------------------------------------

bool CNeuron::IsHost(void)
{
	return (mIsLocal);
}

//---------------------------------------------------------------------------

//static bool AdrEq1(BYTE dlen, BYTE *dom, BYTE net, BYTE nodenr, DOMAIN *pdom)
static bool AdrEq1(BYTE dlen, BYTE *dom, BYTE net, BYTE nodenr, LON_DOMAIN *pdom)
{
	if ((pdom->node & 0x7F) != nodenr) return (false);
	if (pdom->subnet		!= net)	   return (false);
	if (pdom->len			!= dlen)   return (false);
	if (pdom->len			== 0)	   return (true);
	return (memcmp(pdom->id, dom, dlen) == 0);
}

short CNeuron::AdrEq(BYTE dlen, BYTE *dom, BYTE net, BYTE nodenr)
{
	short n = DomCount();
	if (n >= 1) {
		if (AdrEq1(dlen, dom, net, nodenr, Domain(0))) return (0);
	}
	if (n == 2) {
		if (AdrEq1(dlen, dom, net, nodenr, Domain(1))) return (1);
	}
	return (-1);
}

//---------------------------------------------------------------------------

bool CNeuron::SetNodeMode(BYTE mode, BYTE state)
{
	CLockCrit lock(&gLock);
	if (IsHost()) return (L_SetNodeMode(NULL, mode, state));
	if (!mNODE)   return (false);
	return (L_SetNodeMode(mNODE->nid, mode, state));
}

//---------------------------------------------------------------------------

bool CNeuron::ClearStatus(void)
{
	CLockCrit lock(&gLock);
	if (IsHost()) return (L_ClearStatus(NULL));
	if (!mNODE)   return (false);
	return (L_ClearStatus(mNODE->nid));
}

//---------------------------------------------------------------------------

bool CNeuron::Wink(void)
{
	CLockCrit lock(&gLock);
	if (IsHost()) return (false);
	if (!mNODE)   return (false);
	return (L_Wink(mNODE->nid));
}

//---------------------------------------------------------------------------

BYTE CNeuron::Status(void)
{
	CLockCrit lock(&gLock);
	BYTE *nid;
	BYTE status[12];

	if (mNODE || mIsLocal) {
		if (mIsLocal) nid = NULL;
		else	      nid = mNODE->nid;
		if (!L_QueryStatus(nid, status)) return (0);
	}
	else {
		if (!L_QueryStatus(0, mSubnetNr, mNodeNr, status)) return (0);
	}
	return (status[11]);
}

//---------------------------------------------------------------------------

bool CNeuron::LocSet(BYTE *location)
{
	CLockCrit lock(&gLock);
	if (!mNODE) return (false);

	BYTE *nid = NULL;
	if (!mIsLocal && mNODE) nid = mNODE->nid;
	if (!L_WriteMem(nid, 2, 6, RM_CFGREL, location)) return (false);
	if (mNODE) memcpy(mNODE->location, location, 6);
	return (true);
}

//---------------------------------------------------------------------------

bool CNeuron::UpdDomain(BYTE domidx, BYTE subnet, BYTE nodenr,
						BYTE domlen, BYTE *id)
{
	CLockCrit lock(&gLock);
	if (!mNODE) return (false);

	BYTE *nid = NULL;
	if (!mIsLocal) nid = mNODE->nid;
	bool ok = L_UpdDomain(nid, domidx, domlen, id, subnet, nodenr);

	if (ok) {							// update members
		LON_DOMAIN *pdom = Domain(domidx);
		if (!pdom) return (false);
		memcpy(pdom->id, id, sizeof(pdom->id));
		pdom->subnet = (BYTE)subnet;
		pdom->node   = (BYTE)nodenr;
		pdom->len	 = (BYTE)domlen;
	}
	return (ok);
}

//---------------------------------------------------------------------------

short CNeuron::NVFetch(BYTE* data, WORD index)
{
	CLockCrit lock(&gLock);
	BYTE net  = SubnetNr();	if (!net)  return (0);
	BYTE node = NodeNr();	if (!node) return (0);

	return (L_NVFetch(0, net, node, index, data));
}

//---------------------------------------------------------------------------

bool CNeuron::NVSet(WORD sel_dir, BYTE len, BYTE* data)
{
	CLockCrit lock(&gLock);
	BYTE net  = SubnetNr();	if (!net)  return (false);
	BYTE node = NodeNr();	if (!node) return (false);

	return (L_NVSet(0, net, node, sel_dir, len & 0x1F, data));
}

//---------------------------------------------------------------------------

short CNeuron::NVCount(void)
{
	return ((short)m_nvs.GetCount());
}

//---------------------------------------------------------------------------

CNV *CNeuron::NVbyInd(WORD index)
{
	CNV *nv = (CNV *)m_nvs.GetAt(index);
	if (nv && (nv->mIndex == index)) return (nv);

// the following is used if there are not all NVs in the list or
//	if the NVs are not in the right order ...
	for (short npos = 0; npos < NVCount(); npos++) {
		nv = (CNV *)m_nvs.GetAt(npos);
		if (nv->mIndex == index) return (nv);
	}
	return (NULL);
}

//---------------------------------------------------------------------------

short CNeuron::DomCount(void)						// count of domains 
{
	if (!mNODE) return (-1);
	return (mNODE->num_domains);
}

//---------------------------------------------------------------------------

LON_DOMAIN *CNeuron::Domain(BYTE domidx)
{
	if (!mNODE) return (NULL);

	if ((domidx != 0) && (domidx != 1)) return (NULL);
	return (&mNODE->domain[domidx]);
}

//***************************************************************************

CNV::CNV(CNeuron *neuron, WORD index)
{
	mNeuron		= neuron;
 	mIndex		= index;				// index of NV	
	mUserHdl	= 0;
	mSel_dir	= 0;					// direction[14] selector[13..0]
	mCfg		= 0x0F;					// turn[7] svc[6..5] auth[4] adr[3..0]
	mPri_len	= 0;					// prio[7] length[4..0]	
	mSnvt		= 0;					// SNVT type	
	mNVName[0]	= 0;
	mpNVDoc		= NULL;
	mpValue		= NULL;

	NODE *node  = mNeuron->mNODE;
	if (!node)		  return;			// HOST structure
	if (!node->panvs) return;			// no NVs uploaded

	NVSTRUCT *pnv = NODENV(index);
	mSel_dir	= pnv->sel_dir;			// direction[14] selector[13..0]
	mCfg		= pnv->cfg;				// turn[7] svc[6..5] auth[4] adr[3..0]
	mPri_len	= pnv->pri_len;			// prio[7] length[4..0]	
	mSnvt		= pnv->snvt;			// SNVT type	
	my_strncpy(mNVName, pnv->nm_name, sizeof(mNVName));

	mpNVDoc = new char[strlen(pnv->sd_name) +1];
	strcpy(mpNVDoc, pnv->sd_name);
}

//---------------------------------------------------------------------------

CNV::~CNV(void)
{
	if (mpNVDoc) delete mpNVDoc;
	if ((Length() > sizeof(mpValue)) && mpValue) delete mpValue;
}

//---------------------------------------------------------------------------

CNeuron	*CNV::Neuron(void)
{ 
	return (mNeuron);
}

//---------------------------------------------------------------------------

WORD CNV::Index(void)
{ 
	return (mIndex);
}

//---------------------------------------------------------------------------

bool CNV::isOutput(void)
{ 
	return ((mSel_dir & 0x4000) != 0);
}

//---------------------------------------------------------------------------

BYTE CNV::Length(void)
{
	return (mPri_len & 0x1F);
}

//---------------------------------------------------------------------------

BYTE CNV::Type(void)
{
	return (mSnvt);
}

//---------------------------------------------------------------------------

WORD CNV::Selector(void)
{
	return (mSel_dir & 0x3FFF);
}

//---------------------------------------------------------------------------

char *CNV::Name(void)
{
	return (mNVName);
}

//---------------------------------------------------------------------------

char *CNV::SelfDoc(void)					// selfdoc of NV
{
	if (!mpNVDoc) return ("");
	return (mpNVDoc);
}

//---------------------------------------------------------------------------

short CNV::Fetch(BYTE* data)
{
	CLockCrit lock(&gLock);
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (mNeuron->IsHost()) {
		short len = GetValue(data);
		return (len > 0);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	NODE *node = mNeuron->mNODE;
	if (!node) return (0);

	short len = L_NVFetch(node->nid, mIndex, data);
	if (len) SetValue(data);
	return (len);
}

//---------------------------------------------------------------------------

bool CNV::Set(BYTE* data) 
{
	CLockCrit lock(&gLock);
	bool ok;
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (mNeuron->IsHost()) {
		SetValue(data);
		ok = ((CHANeuron *)mNeuron)->SendNV(this, data);
		return (ok);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	NODE *node = mNeuron->mNODE;
	if (!node) return (false);

	ok = L_NVSet(node->nid, mSel_dir, Length(), data);
	if (ok) SetValue(data);
	return (ok);
}

//---------------------------------------------------------------------------

short CNV::GetValue(BYTE* data)
{
	CLockCrit lock(&gLock);
	BYTE len = Length();
	if (!len) return (0);

	if (len <= sizeof(mpValue)) memcpy(data, &mpValue, len);
	else {
		if (!mpValue) memset(data, 0, len);
		else		  memcpy(data, mpValue, len);
	}
	return (len);
}

//---------------------------------------------------------------------------

void CNV::SetValue(BYTE* data)
{
	//CLockCrit lock(&gLock);
	BYTE len = Length();
	if (!len) return;

	if (len <= sizeof(mpValue)) memcpy(&mpValue, data, len);
	else {
		if (!mpValue) mpValue = new BYTE[len];
		memcpy(mpValue, data, len);
	}
	return;
}

//---------------------------------------------------------------------------

void CNV::UpdNVStruct(void *ps)
{
	nv_struct *pnvc = (nv_struct *)ps;

	mSel_dir = pnvc->nv_selector_hi;
	mSel_dir = (mSel_dir << 8) + pnvc->nv_selector_lo;
	if (pnvc->nv_direction) mSel_dir |= 0x4000;
	if (pnvc->nv_priority)  mSel_dir |= 0x8000;

// 	BYTE mCfg;						//	turn[7] svc[6..5] auth[4] adr[3..0]
	mCfg = pnvc->nv_addr_index;
	if (pnvc->nv_auth)		 mCfg |= 0x10;
	if (pnvc->nv_turnaround) mCfg |= 0x80;
	BYTE h = pnvc->nv_service << 5;
	mCfg |= h;
}

//---------------------------------------------------------------------------

void CNV::GetNVStruct(void *ps)
{
	BYTE *dp = (BYTE *)ps;
	*dp++ = (BYTE)(mSel_dir >> 8);
	*dp++ = (BYTE)mSel_dir;
	*dp   = mCfg;
}

//***************************************************************************
/*
BOOL APIENTRY DllMain(HANDLE hModule,
                      DWORD  ul_reason_for_call, 
                      LPVOID lpReserved)
{
	switch (ul_reason_for_call) {
		case DLL_PROCESS_ATTACH:
 			//RETAILMSG(1, (TEXT("\r\nProgram: %s on %s %s %s\r\n"),
 			//			    TEXT(__FILE__),TEXT(__DATE__),TEXT(__TIME__)));
			break;
		case DLL_THREAD_ATTACH:
			break;
		case DLL_PROCESS_DETACH:
			break;
		case DLL_THREAD_DETACH:
			break;
    }
    return TRUE;
}
*/


//***************************************************************************
// am 21.5.2003 ergnzt von Ingo Grothues

bool CNeuron::UpdateAddress(BYTE adr_idx, BYTE format, BYTE domidx, BYTE subnet, BYTE node)
{
	CLockCrit lock(&gLock);
	BYTE *nid = NULL;
	
	if(mNODE) nid = mNODE->nid;

	ADTBL adtbl_entry;

	adtbl_entry.fmt_size = format;					// 0 = unbound or turnaround
													// 1 = subnet_node
													// 3 = broadcast
													
	adtbl_entry.dom_node = (domidx << 7) | node;	// domidx[7] node[6..0]
	adtbl_entry.net_grp = subnet;

	//adtbl_entry.rpt_retry   wird in L_UpdateAddress auf gRPTRETRY gesetzt
	//adtbl_entry.tx_timer	  wird in L_UpdateAddress auf gTXTIMER gesetzt

	return (L_UpdateAddress(nid, adr_idx, &adtbl_entry));
}


bool CNeuron::UpdateNVconfig(BYTE nv_idx)
{
	CLockCrit lock(&gLock);
	BYTE *nid = NULL;
	
	if(mNODE) nid = mNODE->nid;

	CNV* pNV = this->NVbyInd(nv_idx);

	return (L_UpdateNVconfig(nid, pNV));
}


bool CNeuron::QueryStatus2(BYTE* buf)
{
	BYTE* nid;
	
	if (mNODE || mIsLocal) 
	{
		if (mIsLocal) nid = NULL;
		else	      nid = mNODE->nid;
		
		return(L_QueryStatus2(nid, buf));
	}
	else 
	{
		return (L_QueryStatus2(0, this->SubnetNr(), this->NodeNr(), buf));
	}
}


void CNV::SetSelector(WORD sel)
{
	mSel_dir = mSel_dir & 0xC000;			// Bits[13..0] lschen
	mSel_dir = mSel_dir | (sel & 0x3FFF);	// Bits[13..0] setzen
}

void CNV::SetAddrIdx(BYTE idx)
{
	mCfg = mCfg & 0xF0;				// Bits[3..0] lschen
	mCfg = mCfg | (idx & 0x0F);		// Bits[3..0] setzen
}

void CNV::SetTurnaround(BYTE set)
{
	mCfg = mCfg & 0x7F;
	if(set) mCfg = mCfg | 0x80;
}

bool CNV::UpdateNVconfig()
{
	CLockCrit lock(&gLock);
	BYTE *nid = NULL;
	
	NODE* node = mNeuron->GetNodePtr();
	if(node) nid = node->nid;

	return (L_UpdateNVconfig(nid, this));
}
//***************************************************************************

