//////////////////////////////////////////////////////////////////////
//
// g2Lon.h:  interface to LON  (API)
//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
// Copyright (c) 1999  Gesytec GmbH
//
// Version: 2.10
//
// Author: mb, mboehlke@gesytec.de
//
//////////////////////////////////////////////////////////////////////

#ifndef G2LONC_H
#define G2LONC_H

#include "windows_compat.h"
#include "watcher.h"
#include "yaList.h"

#define LONC_API /* __declspec(dllexport) */

class CHANeuron;
class CNeuron;
class CNV;
class CBind;
typedef struct nodestruct NODE;		// internal node data

//---------------------------------------------------------------------------

//#pragma pack( push )
//#pragma pack( 1 )					// !!! one byte alignments !!!

typedef struct domstruct {			// Domain information
	BYTE	id[6];					//	domain id
	BYTE	subnet;					//	subnet number (1 - 127)
	BYTE	node;					//	node number   (1 - 255)
	BYTE	len;					//	length of domain id
	BYTE	key[6];					//	authentication key
} LON_DOMAIN;
//#pragma pack( pop )

//---------------------------------------------------------------------------

class LONC_API CNet : public CyaList			// a list of CNeuron objects
{
public:
	CNet					(void);					// create the net object
	~CNet					(void);					// destruct it

	char	*Version		(void);					// get Version of g2LON
	bool	Open			(DWORD dwIndex);		// open driver and init
	bool	Open			(const char *DevName);	// open driver and init
	void	Close			(void);					// close driver
	BYTE	SetRcvTrx		(BYTE rcvtrx);
	BYTE	SetRptTry		(BYTE rpttry);

	bool	AddNeuron		(CNeuron *neuron);		// add neuron into list
	bool	DelNeuron		(CNeuron *neuron);		// delete neuron
	CNeuron *NeuronByNID	(BYTE *nid);			// get by neuron id
	CNeuron *NeuronByInd	(WORD index);			// get by index
	CNeuron *NeuronByAdr	(BYTE dlen, BYTE *dom, BYTE net, BYTE nodenr);

	short	Scan			(BYTE domidx);			// scan the network
	bool	UploadAll		(bool reload);			// upload known neurons
	CNeuron	*Upload			(BYTE *nidpid);			// upload single neuron

	bool	Status			(BYTE *nid, BYTE *status);	// read status

	short	ReadMem			(BYTE didx, BYTE subnet, BYTE nodenr,
							 WORD adr, short size, BYTE mode, BYTE *buf);

	bool	WriteMem		(BYTE didx, BYTE subnet, BYTE nodenr,
							 WORD adr, short size, BYTE mode, BYTE *buf);

	bool	SetNodeMode		(BYTE didx, BYTE subnet, BYTE nodenr,
							 BYTE mode, BYTE state);

	bool	Wink			(BYTE didx, BYTE subnet, BYTE nodenr);
	bool	QueryStatus		(BYTE didx, BYTE subnet, BYTE nodenr, BYTE *status);


	bool	SendSvp			(void);
	CNV*	NVwasUpdated	(void);					// NV got update event
	short	MsgReceived		(BYTE *msg);			// Application Msg received
	WORD	SetDbgMode		(WORD mode);			// debug flags

	// am 21.5.2003 ergnzt von Ingo Grothues
	bool	OpenRemoteNet	(const char* adr);		// connect to lon server
	// am 29.1.2008 ergnzt von VS
	bool	ReOpenRemoteNet	();		                // reconnect to lon server

protected:
	HANDLE	 mLP4;									// handle to driver
	CNeuron	*UpldLocal		(void);					// upload local neuron
    char    *RemoteAdr;
};

//---------------------------------------------------------------------------

class LONC_API CNeuron : public CyaObj				// a neuron object
{
friend class CNet;
friend class CNV;
protected:
	CNeuron				();							// create a neuron
	CNeuron				(NODE *node);

public:
	~CNeuron			(void);

	CNeuron				(BYTE domidx, BYTE subnet, BYTE node);
// upload neurons data from net ...
	bool	Upload		(BYTE *nid);
	bool	Upload		(BYTE domidx, BYTE subnet, BYTE node);

	BYTE	*NID		(void);						// neuron id
	BYTE	*PID		(void);						// program id
	BYTE	*Location	(void);						// location

	short	DomCount	(void);						// count of domains
	LON_DOMAIN	*Domain		(BYTE domidx);				// domain info
	BYTE	SubnetNr	(void);						// subnet number of domain 0
	BYTE	NodeNr		(void);						// node   number of domain 0

	bool	IsHost		(void);						// local neuron ?
	short	AdrEq		(BYTE dlen, BYTE *dom, BYTE net, BYTE nodenr);
	short	NVCount		(void);						// count of NVs
	CNV		*NVbyInd	(WORD index);				// get NV

	BYTE	Status		(void);						// read status
	short	NVFetch		(BYTE* data, WORD index);	// fetch a NV
	bool	NVSet		(WORD sel_dir, BYTE len, BYTE* data);

	bool	SetNodeMode	(BYTE mode, BYTE state);
	bool	ClearStatus	(void);
	bool	Wink		(void);

	bool	LocSet		(BYTE *location);			// write location
	bool	UpdDomain	(BYTE domidx, BYTE subnet, BYTE nodenr,
						 BYTE domlen, BYTE *id);


	// am 21.5.2003 ergnzt von Ingo Grothues
	NODE*	GetNodePtr(void) { return mNODE; };
	bool	UpdateAddress(BYTE adr_idx, BYTE format, BYTE domidx, BYTE subnet, BYTE node);
	bool	UpdateNVconfig(BYTE nv_idx);
	bool	QueryStatus2(BYTE* buf);

protected:
	short	CreateNVs	(void);
	void	DelNodePtr	(void);

public:
	CyaList	m_nvs;									// list of NV objects
	DWORD	mUserHdl;								// free for the user

protected:
	NODE	*mNODE;									// ref to old interface
	bool	mIsLocal;								// is host adapter
// the following is used to access a neuron without uploading ...
	BYTE	mDomIdx;
	BYTE	mSubnetNr;
	BYTE	mNodeNr;
};

//---------------------------------------------------------------------------

class LONC_API CNV : public CyaObj		// a NV object
{
friend class CNet;
friend class CHANeuron;
friend class CNeuron;
friend class CBind;
protected:
	CNV		(CNeuron *neuron, WORD index);		// create a NV
	~CNV	(void);

public:
	CNeuron	*Neuron		(void);					// backref to neuron
	WORD	Index		(void);					// index of NV
	bool	isOutput	(void);					// NV is output NV
	BYTE	Length		(void);					// data size of NV
	BYTE	Type		(void);					// SNVT type of NV
	WORD	Selector	(void);					// selector of NV
	char	*Name		(void);					// name of NV
	char	*SelfDoc	(void);					// selfdoc of NV

	short	Fetch		(BYTE* data);			// fetch NV data
	bool	Set			(BYTE* data);			// send NV data

	short	GetValue	(BYTE* data);
	void	SetValue	(BYTE* data);

	// am 21.5.2003 ergnzt von Ingo Grothues
	WORD	GetSel_dir()	{ return mSel_dir;	};
	BYTE	GetCfg() 		{ return mCfg; 		};
	void	SetSelector(WORD sel);
	void	SetAddrIdx(BYTE idx);
	void	SetTurnaround(BYTE set);
	bool	UpdateNVconfig();

protected:
	void	UpdNVStruct	(void *ps);
	void	GetNVStruct	(void *ps);

public:
	DWORD		mUserHdl;						// free for the user

protected:
	CNeuron		*mNeuron;			// backref to neuron
 	WORD		mIndex;				// index of NV
	WORD		mSel_dir;			// prio[15] direction[14] selector[13..0]
	BYTE		mCfg;				// turn[7] svc[6..5] auth[4] adr[3..0]
	BYTE		mPri_len;			// prio[7] length[4..0]
	BYTE		mSnvt;				// SNVT type
	char		mNVName[17];		// Name of NV
	char		*mpNVDoc;			// Selfdoc of NV
	BYTE		*mpValue;			// Value of NV
};

//---------------------------------------------------------------------------
#endif // G2LONC_H
