/********************************************
* Dateiname  : main_functions.cpp           *
*                                           *
* Erstellt   : 05.03.2003                   *
* geaendert  : 15.05.2003                   *
*                                           *
* login      : ingo                         *
*                                           *
* Author     : Ingo Grothues                *
*                                           *
* Funktion   : Testprogramm fuer LON-API    *
*                                           *
********************************************/
//#include <string>
//#include <iostream>
using namespace std;

extern int listen_and_serve(const char *DevName, const char *service);	// in rawtcp.cpp

extern	NIDSTRUCT	*gNIDs;


void print_menu()
{
	fprintf(stdout, "\n\nLON-Tool V1.0  (known devices: %i)\n", known_devs);
	fprintf(stdout, "----------------------------------\n");
	fprintf(stdout, "1) Add devices by Network-Scan\n");
	fprintf(stdout, "2) Add devices by NID\n");
	fprintf(stdout, "3) List known devices\n\n");

	fprintf(stdout, "4) Send message\n");
	fprintf(stdout, "5) Set Domain/Subnet/Node\n");
	fprintf(stdout, "6) Query/Set Status\n\n");

	fprintf(stdout, "7) Edit NV's\n");
	fprintf(stdout, "8) Edit NVconfig\n");
	fprintf(stdout, "9) Edit Address-Table\n\n");

	fprintf(stdout, "0) Raw TCP/IP\n");
	fprintf(stdout, "x) Exit\n\n");
}


void set_input_mode()
{
	if(!isatty (STDIN_FILENO))
	{
  		fprintf (stderr, "Not a terminal.\n");
  		exit(1);
	}

	struct termios tattr;
	tcgetattr (STDIN_FILENO, &saved_attributes);

	tcgetattr (STDIN_FILENO, &tattr);
	//tattr.c_lflag &= ~(ICANON|ECHO);	 // clear ICANON + ECHO
	tattr.c_lflag &= ~(ICANON);	 // clear only ICANON
	tattr.c_cc[VMIN] = 1;
	tattr.c_cc[VTIME] = 0;
	tcsetattr (STDIN_FILENO, TCSAFLUSH, &tattr);
}


bool add_device(unsigned char* nid)
{
	CNeuron* neuron = NETWORK.NeuronByNID(nid);
	if(neuron)
	{
		fprintf(stdout, "Neuron already known.\n");
		return false;	// NID schon bekannt
	}

/*
	for(int i=1; i<=known_devs; i++)
	{
		CNeuron* dev = NETWORK.NeuronByInd(i);

		char nid1[20];
		char nid2[20];
		hexdump(nid1, nid, 6, ':');
		hexdump(nid2, dev->NID(), 6, ':');
		nid1[17] = 0;
		nid2[17] = 0;
		fprintf(stdout, "New Device      :   NID=%s\n", nid1);
		fprintf(stdout, "Known Device[%i] :   NID=%s\n",i, nid2);

		if(!memcmp(dev->NID(), nid, 6))
		{
			fprintf(stdout, "Neuron-ID already known.\n");
			return false;	// NID schon bekannt
		}
	}
*/
	neuron = new CNeuron(0,0,0);

	if(!neuron->Upload(nid))
	{
		fprintf(stdout, "Uploading new Neuron failed.\n");
		delete neuron;
		return false;
	}

	fprintf(stdout, "New Neuron found - adding device to network list.\n");

	if(!NETWORK.AddNeuron(neuron))
	{
		fprintf(stdout, "Adding new Neuron failed.\n");
		delete neuron;
		return false;
	}

	known_devs++;
	return true;
}



bool init_network(const char *DevName)
    {
    fprintf(stdout, "\nOpening Network-Driver %s ...", DevName);
    fflush(stdout);

    if (*DevName == '/')
        {
        if( NETWORK.Open(DevName) )	// versuche /dev/lon1 zu oeffnen
            {
            fprintf(stdout, "OK !\n");
            localDevice = NETWORK.NeuronByInd(0);
            return true;
            }
        fprintf(stdout, "failed !\n");
        }
    else
        {
        if(NETWORK.OpenRemoteNet(DevName))
            {
            fprintf(stdout, "Connection to %s successfully established.\n", DevName);
            return true;
            }
        else
            {
            fprintf(stdout, "Connection to %s failed !\n", DevName);
            }
        }
    return false;
    }


int scan_network()
{
	fprintf(stdout, "\nStarting LON-Scan...\n");
	int num = NETWORK.Scan(0);
	fprintf(stdout, "Found %i LON-Devices on domain 0. \n\n", num);
	fprintf(stdout, "(There may be devices which are already in CONFIGURED state.\n The Network-Scan finds only devices in UNCONFIGURED state.)\n");
	if(num>0)
	{
		fprintf(stdout, "\nChecking device data. Please wait a few seconds...\n");

		NIDSTRUCT *nid = gNIDs;
		//CNeuron	  *neuron;

		while(nid)
		{
			add_device(nid->nidpid);
			nid = nid->next;
		}

		//NETWORK.UploadAll(true);
	}
	return num;
}



char input_loop()
{
    // Initialisiere File-Descriptor-Sets
	fd_set active_fd_set, read_fd_set;		// Deklariere File-Descriptor-Sets
	FD_ZERO (&active_fd_set);				// setzt "active_fd_set" als empty set
	FD_SET (STDIN_FILENO, &active_fd_set);	// fuegt stdin dem "active_fd_set" hinzu

	struct timeval timeout;

	errno = 0;
   	read_fd_set = active_fd_set;

	while(true)
	{
		timeout.tv_sec = 0;
		timeout.tv_usec = 100*1000;		// alle 100ms Timeout

   		read_fd_set = active_fd_set;
		int rval = select(FD_SETSIZE, &read_fd_set, NULL, NULL, &timeout);
		if (rval < 0)
		{
		    if (errno == EINTR) continue;	// Alarm-Signal
			cerr << "error : select() [" << errno << "]\n";
		    exit (1);
		}
		else if (rval == 0)	// Timeout
		{
			char dump[512];
			unsigned char msg[256];

			CNV *pNV = NETWORK.NVwasUpdated();
			if(pNV)
			{
				fprintf(stdout, "\nNV was updated! \n");
				fprintf(stdout, "\nNVname = \"%s\", SELCT=%i\n", pNV->Name(), pNV->Selector());
			}

			int len = NETWORK.MsgReceived(msg);
			if(len)
			{
				hexdump(dump, msg, len, 0);
				fprintf(stdout, "\nMsg[%i] received : %s\n", len, dump);

				if(msg[16] == 0x7F)		// NM_Service_Pin
				{
					fprintf(stdout, "ServicePin-Message received !\n");
					add_device(&msg[17]);
					fprintf(stdout, "\n");
					print_menu();
				}
			}
		}
		else
		{
			for (int i=0; i < FD_SETSIZE; ++i)
		    {
		      	if (FD_ISSET(i, &read_fd_set))
		        {
					if(i==STDIN_FILENO) return fgetc(stdin);
				}
			}
		}
	}
}



bool add_nid()
{
	char nidpid[15];
	unsigned char bin[14];
	fprintf(stdout, "\n\nEnter NeuronID(+PID) : ");

	tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
	fgets(nidpid, 15, stdin);
	set_input_mode();

	for(int i=0; i<15; i++) if(nidpid[i]=='\n') nidpid[i]=0;
	htobin(nidpid, bin);

	return add_device((unsigned char*)bin);
}


int add_svpin()
{
	return 0;
}


void localDeviceInfo()
{
	char localNID[20];
	char localPID[25];
	hexdump(localNID, localDevice->NID(), 6, ':');
	hexdump(localPID, localDevice->PID(), 8, ':');
	localNID[17] = 0;
	localPID[23] = 0;
	fprintf(stdout, "Local Device :   NID=%s, PID=%s, STATUS=%i, NVcount=%i\n", localNID, localPID, localDevice->Status(), localDevice->NVCount());
}



void list_devices(bool menu)
{
	int start = 0;
	if(menu) start = 1;
/*
LON_DOMAIN	*Domain		(BYTE domidx);				// domain info

typedef struct domstruct {			// Domain information
	BYTE	id[6];					//	domain id
	BYTE	subnet;					//	subnet number (1 - 127)
	BYTE	node;					//	node number   (1 - 255)
	BYTE	len;					//	length of domain id
	BYTE	key[6];					//	authentication key
} LON_DOMAIN, DOMAIN;
*/
	for(int i=start; i<=known_devs; i++)
	{
		CNeuron* device = NETWORK.NeuronByInd(i);
		LON_DOMAIN*	domain = device->Domain(0);

		char DOM[15];
		//char SUBNET[3];
		//char NODE[3];
		char NID[20];
		char PID[25];

		hexdump(DOM, domain->id, domain->len, 0);
		//hexdump(SUBNET, &domain->subnet, 1, 0);
		//hexdump(NODE, &domain->node, 1, 0);
		hexdump(NID, device->NID(), 6, ':');
		hexdump(PID, device->PID(), 8, ':');

		DOM[domain->len*2] = 0;
		//SUBNET[2] = 0;
		//NODE[2] = 0;
		NID[17] = 0;
		PID[23] = 0;

		if(menu) fprintf(stdout, "%i) ", i);
		if(i==0) fprintf(stdout, "Local     :  LON_DOMAIN=%s, SUBNET=%i, NODE=%i, NID=%s, PID=%s, STATUS=%i, NVcount=%i\n", DOM, domain->subnet, domain->node, NID, PID, device->Status(), device->NVCount());
		else 	 fprintf(stdout, "Remote[%i]:  LON_DOMAIN=%s, SUBNET=%i, NODE=%i, NID=%s, PID=%s, STATUS=%i, NVcount=%i\n",i, DOM, domain->subnet, domain->node, NID, PID, device->Status(), device->NVCount());
	}
}


int choose_device(bool allitem)
{
/*	if(known_devs == 0)
	{
		fprintf(stdout, "\n\nNo devices found !\n");
		return -1;
	}
*/
	fprintf(stdout, "\n\nChoose device(s)\n");
	fprintf(stdout, "----------------\n");

	list_devices(true);
	if(allitem) fprintf(stdout, "%i) All devices\n\n", known_devs+1);
	else fprintf(stdout, "\n\n");

	return (input_loop() - '0');
}


void edit_ADTBLs()
{
	int dev = choose_device(false);

	if((dev >= 0)&&(dev <= known_devs))
	{
		CNeuron* device = NETWORK.NeuronByInd(dev);

		fprintf(stdout, "\n\nAction:\n");
		fprintf(stdout, "1) List AddressTable-Entrys\n");
		fprintf(stdout, "2) Set AddressTable-Entry\n");
		fprintf(stdout, "3) Back to Mainmenu\n\n");

		char choice = input_loop();
		choice -= '0';

		switch(choice)
		{
			case 1:
			{
				NODE* node = device->GetNodePtr();
				for(int i=0; i<node->num_adtbl; i++)
				{
					ADTBL* tbl = &node->adtbl[i];
					char entry[20];
					hexdump(entry, &tbl->fmt_size, 5, ':');
					fprintf(stdout, "AddressTable-Entry[%i] = %s\n", i, entry);
				}
				break;
			}

			case 2:
			{
				unsigned char adr_idx;
				unsigned char domidx;
				unsigned char subnet;
				unsigned char node;
				unsigned char backlog;
				short number;

				fprintf(stdout, "\n\nAddress-Type:\n");
				fprintf(stdout, "1) Unbound\n");
				fprintf(stdout, "2) Turnaround\n");
				fprintf(stdout, "3) Subnet/Node\n");
				fprintf(stdout, "4) Broadcast\n\n");

				char choice = input_loop();
				choice -= '0';

				fprintf(stdout, "\n\nEnter AddressTable-Index (0-14) : ");
				tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
				cin >> number;
				adr_idx = number;
				set_input_mode();

				switch(choice)
				{
					case 1:
					{
						if(!device->UpdateAddress(adr_idx, 0, 0, 0, 0))
						{
							fprintf(stdout, "Device[%i]->UpdateAddress(UNBOUND) failed !\n\n", dev);
						}
						break;
					}

					case 2:
					{
						if(!device->UpdateAddress(adr_idx, 0, 0, 0, 1))
						{
							fprintf(stdout, "Device[%i]->UpdateAddress(TURNAROUND) failed !\n\n", dev);
						}
						break;
					}

					case 3:
					{
						fprintf(stdout, "\n\nEnter Target-Domain (0/1) : ");
						tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
						cin >> number;
						domidx = number;
						set_input_mode();

						fprintf(stdout, "\n\nEnter Target-Subnet (1-127) : ");
						tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
						cin >> number;
						subnet = number;
						set_input_mode();

						if(choice == 3)
						{
							fprintf(stdout, "\n\nEnter Target-Node (1-255) : ");
							tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
							cin >> number;
							node = number;
							set_input_mode();

							if(!device->UpdateAddress(adr_idx, 1, domidx, subnet, node))
							{
								fprintf(stdout, "Device[%i]->UpdateAddress(SUBNET/NODE) failed !\n\n", dev);
							}
						}
						else
						{
							fprintf(stdout, "\n\nEnter backlog (0-63, 0=unknown) : ");
							tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
							cin >> number;
							backlog = number;
							set_input_mode();

							if(!device->UpdateAddress(adr_idx, 3, domidx, subnet, backlog))
							{
								fprintf(stdout, "Device[%i]->UpdateAddress(BROADCAST) failed !\n\n", dev);
							}
						}
						break;
					}
				}
			}
		}
	}
}

void upd_domain()
{
	int dev = choose_device(false);

	if((dev >= 0)&&(dev <= known_devs))
	{
		CNeuron* device = NETWORK.NeuronByInd(dev);

		char domain[3];
		unsigned char dombin[6];
		short number;
		unsigned char subnet;
		unsigned char node;
		memset(dombin, 0, 6);

		fprintf(stdout, "\n\nEnter Domain (hex, 2digits) : ");
		tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
		fgets(domain, 3, stdin);
		set_input_mode();

		for(int i=0; i<2; i++) if(domain[i]=='\n') domain[i]=0;
		htobin(domain, dombin);

		fprintf(stdout, "\n\nEnter Subnet (dec, 1-127) : ");
		tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
		cin >> number;
		subnet = number;
		set_input_mode();

		fprintf(stdout, "\n\nEnter Node (dec, 1-255) : ");
		tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
		cin >> number;
		node = number;
		set_input_mode();

		BYTE domlen = 1;
		if(domain[0]==0) domlen = 0;
		//bool CNeuron::UpdDomain(BYTE domidx, BYTE subnet, BYTE nodenr, BYTE domlen, BYTE *id)
		if(!device->UpdDomain(0, subnet, node, domlen, dombin))
		{
			fprintf(stdout, "Device[%i]->UpdDomain() failed !\n\n", dev);
		}
	}
}

void send_message()
{
	fprintf(stdout, "\n  Winking...\n");
	for(int i=0; i<=known_devs; i++)
	{
		CNeuron* device = NETWORK.NeuronByInd(i);
		device->Wink();
	}
}


void setget_status()
{
	int first, last, state;

	int dev = choose_device(true);

	if((dev >= 0)&&(dev <= known_devs+1))
	{
		fprintf(stdout, "\n\nAction:\n");
		fprintf(stdout, "1) Query Status\n");
		fprintf(stdout, "2) Set Status\n");
		fprintf(stdout, "3) Back to Mainmenu\n\n");

		char choice = input_loop();
		choice -= '0';

		switch(choice)
		{
			case 1:
			{
				if(dev == known_devs+1) { first = 1; 	last = known_devs;	}
				else					{ first = dev;	last  = dev;		}

				for(int i=first; i<=last; i++)
				{
					CNeuron* device = NETWORK.NeuronByInd(i);

					char dumpbuf[50];
					char NID[20];
					ND_query_status_response resp;

					if(!device->QueryStatus2(resp.buf))
					{
						fprintf(stdout, "\n\nDevice[%i]->QueryStatus() failed !\n\n", dev);
						return;
					}

					hexdump(NID, device->NID(), 6, ':');
					hexdump(dumpbuf, resp.buf, 15, ' ');
					NID[17] = 0;
					dumpbuf[49] = 0;

					fprintf(stdout, "\n\nDevice[NID=%s]->QueryStatus() = %s\n\n", NID, dumpbuf);

					// Big Endian -> Little Endian
					resp.expl.xmit_errors = 		 ((resp.expl.xmit_errors & 0x00FF) << 8) 		  | ((resp.expl.xmit_errors & 0xFF00) >> 8);
					resp.expl.transaction_timeouts = ((resp.expl.transaction_timeouts & 0x00FF) << 8) | ((resp.expl.transaction_timeouts & 0xFF00) >> 8);
					resp.expl.rcv_transaction_full = ((resp.expl.rcv_transaction_full & 0x00FF) << 8) | ((resp.expl.rcv_transaction_full & 0xFF00) >> 8);
					resp.expl.lost_msgs = 			 ((resp.expl.lost_msgs & 0x00FF) << 8) 			  | ((resp.expl.lost_msgs & 0xFF00) >> 8);
					resp.expl.missed_msgs = 		 ((resp.expl.missed_msgs & 0x00FF) << 8) 		  | ((resp.expl.missed_msgs & 0xFF00) >> 8);

					fprintf(stdout, "xmit_errors          : %i\n", resp.expl.xmit_errors);
					fprintf(stdout, "transaction_timeouts : %i\n", resp.expl.transaction_timeouts);
					fprintf(stdout, "rcv_transaction_full : %i\n", resp.expl.rcv_transaction_full);
					fprintf(stdout, "lost_msgs            : %i\n", resp.expl.lost_msgs);
					fprintf(stdout, "missed_msgs          : %i\n", resp.expl.missed_msgs);

					// resetcause-String suchen
					int n=0;
					//const char* resetstr;
					if(resp.expl.reset_cause == 0) n=0;
					if((resp.expl.reset_cause & 1)) n=1;
					if((resp.expl.reset_cause & 2)) n=2;
					if((resp.expl.reset_cause & 12)) n=3;
					if((resp.expl.reset_cause & 20)) n=4;

					fprintf(stdout, "reset_cause          : %i  (%s)\n", resp.expl.reset_cause, reset_cause_str[n]);
					fprintf(stdout, "node_state           : %i\n", resp.expl.node_state);
					fprintf(stdout, "version_number       : %i\n", resp.expl.version_number);

					// errlog-String suchen
					n=0;
					if(resp.expl.error_log != 0) n = resp.expl.error_log - 128;
					fprintf(stdout, "error_log            : %i  (%s)\n", resp.expl.error_log, err_log_str[n]);
					fprintf(stdout, "model_number         : %i\n\n", resp.expl.model_number);
				}
				break;
			}
			case 2:
			{
				fprintf(stdout, "\n\nModes:\n");
				fprintf(stdout, "1) APPL_OFFLINE\n");
				fprintf(stdout, "2) APPL_ONLINE\n");
				fprintf(stdout, "3) APPL_RESET\n\n");

				fprintf(stdout, "States:\n");
				fprintf(stdout, "4) APPL_UNCNFG\n");
				fprintf(stdout, "5) NO_APPL_UNCNFG\n");
				fprintf(stdout, "6) CNFG_ONLINE\n");
				fprintf(stdout, "7) CNFG_OFFLINE\n");
				fprintf(stdout, "8) SOFT_OFFLINE\n\n");

				fprintf(stdout, "9) Back to Mainmenu\n\n");

				char choice = input_loop();
				choice -= '0';

				/*
					#define		APPL_OFFLINE		0
					#define		APPL_ONLINE			1
					#define		APPL_RESET			2
					#define		CHANGE_STATE		3

					#define		APPL_UNCNFG			2
					#define		NO_APPL_UNCNFG		3
					#define		CNFG_ONLINE			4
					#define		CNFG_OFFLINE		6
					#define		SOFT_OFFLINE	   12
				*/

				switch(choice)
				{
					case 1:	state = APPL_ONLINE; break;
					case 2:	state = APPL_ONLINE; break;
					case 3:	state = APPL_RESET; break;
					case 4:	state = APPL_UNCNFG; break;
					case 5:	state = NO_APPL_UNCNFG; break;
					case 6:	state = CNFG_ONLINE; break;
					case 7:	state = CNFG_OFFLINE; break;
					case 8:	state = SOFT_OFFLINE; break;
				}

				if(dev == known_devs+1) { first = 1; 	last = known_devs;	}
				else					{ first = dev;	last  = dev;		}

				for(int i=first; i<=last; i++)
				{
					CNeuron* device = NETWORK.NeuronByInd(i);
					if(!device->SetNodeMode(CHANGE_STATE, state))
					{
						fprintf(stdout, "Device[%i]->SetNodeMode() failed !\n\n", dev);
					}
				}
				break;
			}
		}
	}
}


void read_NVs(int dev)
{
		CNeuron* device = NETWORK.NeuronByInd(dev);
		int nvs = device->NVCount();

		if(nvs > 0)
		{
			unsigned char buf[256];
			char dumpbuf[256];
			unsigned int n;

			fprintf(stdout, "  Reading NV's :\n");

			for(int i=0; i<nvs; i++)
			{
				CNV* pNV = device->NVbyInd(i);

				n = pNV->Fetch(buf);
				hexdump(dumpbuf, buf, n, 0);
				dumpbuf[n*2] = 0;
				fprintf(stdout, "NV[%i]->Fetch() = %i ..... DATA:%s\n", i, n, dumpbuf);
			}
		}
}



void write_NVs(int dev)
{
		CNeuron* device = NETWORK.NeuronByInd(dev);
		int nvs = device->NVCount();

		if(nvs > 0)
		{
			unsigned char buf[10];
			char dumpbuf[30];
			unsigned int num = 0;

			bool ok;
			unsigned char data[2] = { 0x01, 0x01};
			unsigned char data2[2] = { 0x00, 0x10};
			fprintf(stdout, "  Setting Input-NV's to 0x01 ...\n\n");

			for(int j=2; j<14; j++)
			{
				CNV* pNV = device->NVbyInd(j);
				CNV* pNV2 = device->NVbyInd(j+24);

				ok = pNV2->Set(data2);
				ok = pNV->Set(data);
				if(ok)	fprintf(stdout, "\nNV[%i](\"%s\")->Set() = true\n", j, pNV->Name());
				else	fprintf(stdout, "\nNV[%i]->Set() = false\n", j);

				num = pNV->Fetch(buf);
				hexdump(dumpbuf, buf, num, 0);
				dumpbuf[num*2] = 0;
				fprintf(stdout, "NV[%i]->Fetch() = %i ..... DATA:%s\n", j, num, dumpbuf);

				num = pNV2->Fetch(buf);
				hexdump(dumpbuf, buf, num, 0);
				dumpbuf[num*2] = 0;
				fprintf(stdout, "NV[%i]->Fetch() = %i ..... DATA:%s\n", j+24, num, dumpbuf);
			}
		}
}


void list_NVs(int dev)
{
		fprintf(stdout, "\n\n\n");
		CNeuron* device = NETWORK.NeuronByInd(dev);

		for(int i=0; i<device->NVCount(); i++)
		{
			CNV* pNV = device->NVbyInd(i);

			//	WORD	Index		(void);					// index of NV
			//	bool	isOutput	(void);					// NV is output NV
			//	BYTE	Length		(void);					// data size of NV
			//	BYTE	Type		(void);					// SNVT type of NV
			//	WORD	Selector	(void);					// selector of NV
			//	char	*Name		(void);					// name of NV
			//	char	*SelfDoc	(void);					// selfdoc of NV

			if(pNV->isOutput())
				fprintf(stdout, "NV[%i]:  NAME=\"%s\", DESCR=\"%s\", DIR=out, SIZE=%i, SNVT=%i, SELCT=%i \n", i, pNV->Name(), pNV->SelfDoc(), pNV->Length(), pNV->Type(), pNV->Selector());
			else
				fprintf(stdout, "NV[%i]:  NAME=\"%s\", DESCR=\"%s\", DIR=in, SIZE=%i, SNVT=%i, SELCT=%i \n", i, pNV->Name(), pNV->SelfDoc(), pNV->Length(), pNV->Type(), pNV->Selector());
		}

}


void edit_NVs()
{
	int dev = choose_device(false);

	if((dev >= 1)&&(dev <= known_devs))
	{

		fprintf(stdout, "\n\nAction:\n");
		fprintf(stdout, "1) Read NV's\n");
		fprintf(stdout, "2) Write NV's\n");
		fprintf(stdout, "3) List NV's\n");
		fprintf(stdout, "4) Back to Mainmenu\n\n");

		char choice = input_loop();
		choice -= '0';

		switch(choice)
		{
			case 1:	read_NVs(dev); break;
			case 2:	write_NVs(dev); break;
			case 3:	list_NVs(dev); break;
		}
	}
}

void edit_NVconfig()
{
	int dev = choose_device(false);

	if((dev >= 0)&&(dev <= known_devs))
	{
		CNeuron* device = NETWORK.NeuronByInd(dev);

		fprintf(stdout, "\n\nAction:\n");
		fprintf(stdout, "1) List NVconfigs\n");
		fprintf(stdout, "2) Set NVconfig\n");
		fprintf(stdout, "3) Back to Mainmenu\n\n");

		char choice = input_loop();
		choice -= '0';

		switch(choice)
		{
			case 1:
			{
				for(int i=0; i<device->NVCount(); i++)
				{
					CNV* pNV = device->NVbyInd(i);

					unsigned short	sel_dir = pNV->GetSel_dir();
					unsigned char	cfg 	= pNV->GetCfg();

					unsigned char prio = (sel_dir & 0x8000)? 1:0;
					unsigned char dir  = (sel_dir & 0x4000)? 1:0;
					unsigned short sel = sel_dir & 0x3FFF;

					unsigned char turn = (cfg & 0x80)? 1:0;
					unsigned char svc  = (cfg & 0x60) >> 5;
					unsigned char auth = (cfg & 0x10)? 1:0;
					unsigned short adr = cfg & 0x0F;

					fprintf(stdout, "NVcfg[%i]: prio=%i, dir=%i, selector=%i, turn=%i, service=%i, auth=%i, adr_idx=%i\n", i, prio, dir, sel, turn, svc, auth, adr);
				}
				break;
			}

			case 2:
			{
				unsigned char nv_idx;
				unsigned char adr_idx;
				unsigned char turn;
				unsigned short sel;
				unsigned short number;

				fprintf(stdout, "\n\nEnter NV-Index (0-%i) : ", device->NVCount());
				tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
				cin >> number;
				nv_idx = number;
				set_input_mode();

				fprintf(stdout, "\n\nEnter Selector (0-16383) : ");
				tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
				cin >> number;
				sel = number;
				set_input_mode();

				fprintf(stdout, "\n\nEnter AddressTable-Index (0-14) : ");
				tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
				cin >> number;
				adr_idx = number;
				set_input_mode();

				fprintf(stdout, "\n\nTurnaround ? (0=no, 1=yes) : ");
				tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
				cin >> number;
				turn = number;
				set_input_mode();

				CNV* pNV = device->NVbyInd(nv_idx);
				pNV->SetSelector(sel);
				pNV->SetAddrIdx(adr_idx);
				pNV->SetTurnaround(turn);

				if(!device->UpdateNVconfig(nv_idx))
				{
					fprintf(stdout, "Device[%i]->UpdateNVconfig() failed !\n\n", dev);
				}
				break;
			}
		}
	}
}



void raw_tcpip(const char *DevName, const char *service)
{
	fprintf(stdout, "\n\nAction:\n");
	fprintf(stdout, "1) Start service (listen for clients)\n");
	fprintf(stdout, "2) Connect to service (as client)\n");
	fprintf(stdout, "3) Back to Mainmenu\n\n");

	char choice;
	do
	{
		choice = input_loop();
	} while (choice == 0x0A);
	choice -= '0';

	switch(choice)
	{
		case 1:
		{
			NETWORK.Close();
			listen_and_serve(DevName, service);
			break;
		}

		case 2:
		{
			string adr;
			NETWORK.Close();

			fprintf(stdout, "\n\nEnter IP-Address : ");
			tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
			cin >> adr;
			set_input_mode();

			if(NETWORK.OpenRemoteNet(adr.c_str()))
			{
				fprintf(stdout, "Connection to %s successfully established.\n\n", adr.c_str());
			}
			else
			{
				fprintf(stdout, "Connection to %s failed !\n\n", adr.c_str());
			}
			break;
		}
	}
}
