#ifndef NI_PRIVATE_H
#define NI_PRIVATE_H

#include "ni_msg.h"
#include "g2Lon.h"

#define IOCTL_GETVERSION	0x43504C00
#define IOCTL_WATCHER		0x43504C01
#define IOCTL_READ			0x43504C02

#define	HEAD_CODE			17

//---------------------------------------------------------------------------
// LDV error codes ...
typedef enum {
    LDV_OK = 0,
    LDV_NOT_FOUND,
    LDV_ALREADY_OPEN,
    LDV_NOT_OPEN,
    LDV_DEVICE_ERR,
    LDV_INVALID_DEVICE_ID,
    LDV_NO_MSG_AVAIL,
    LDV_NO_BUFF_AVAIL,
    LDV_DEVICE_BUSY,
} LDVCode;

// LonTalk message services ...
#define SVC_ackd			0x00
#define SVC_unackd_rpt		0x20
#define SVC_unackd			0x40
#define SVC_request			0x60

// Memory relative definitions ..
#define RM_ABS				0
#define RM_RDONLY			1
#define RM_CFGREL			2

//---------------------------------------------------------------------------
//#pragma pack( push )
//#pragma pack( 1 )					// !!! one byte alignments !!!

typedef struct NIDSTRUCT {			// element of a list of known nids
	NIDSTRUCT	*next;				//	next in list
	BYTE		nidpid[14];			//	nid, pid
	BYTE		dom_idx;			//	domain index
} NIDSTRUCT;

//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

typedef struct {					// Adress table entry
	BYTE	fmt_size;				//	format or 0x80 | group size
	BYTE	dom_node;				//	domain[7] node/memb[6..0]
	BYTE	rpt_retry;				//	rpt_timer[7..4]      retry[3..0]
	BYTE	tx_timer;				//                       tx_timer[3..0]
	BYTE	net_grp;				//	destination subnet or group
} ADTBL;

//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

typedef struct {					// Network Variable
 	WORD	index;					//	index of NV
	WORD	sel_dir;				//	direction[14] selector[13..0]
	BYTE	cfg;					//	turn[7] svc[6..5] auth[4] adr[3..0]
	BYTE	pri_len;				//	prio[7] length[4..0]
	BYTE	snvt;					//	SNVT type
//	char	nvname[32];		// ???	//	Name or Selfdocu or both
	char	nm_name[17];			//  nv_name (nm extension)
	char	sd_name[32];			//  self_doc (sd extension)
} NVSTRUCT;

//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

typedef struct nodestruct {			// Neuron representation
	BYTE	*panvs;					//	points to an array of nv structures
	WORD	num_nvs;				//	<= 4096
	WORD	nodeflgs;				//	some flags

	BYTE	num_domains;			//	1 or 2
	BYTE	num_adtbl;				//	<= 15
	BYTE	nid[6];					//	Neuron  ID
	BYTE	pid[8];					//	Program ID
	BYTE	location[6];			//	Location string
	LON_DOMAIN	domain[2];				//	the two domain entries
	ADTBL	adtbl[15];				//	address table entries
	BYTE	bigbuf;					//	readmem in 32 byte blocks possible
} NODE;

//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

typedef struct {
	NI_Hdr	ni_hdr;					// cmd[7..4]            queue[3..0]
	union {
		struct {
 		 BYTE	svc_tag;			// 0[7] Service[6..5] auth[4]   tag[3..0]
 		 BYTE	flags;				// prio path cplcode[5..4] expl altp pool resp
 		 BYTE	data_len;
		} header;
		MsgHdr	msg_hdr;
	};

	union {
		struct {
			BYTE	format;				// rcv: domain[7] flex[6]
			union {
				struct {
					BYTE dom_node;      // domain[7] node/memb[6..0]
					BYTE rpt_retry;     // rpt_timer[7..4]      retry[3..0]
					BYTE tx_timer;      //                      tx_timer[3..0]
					BYTE dnet_grp;      // destination subnet or group
					BYTE nid[6];        // NEURON ID
				} send;
				struct {
					BYTE snet;          // source subnet
					BYTE snode;         // source node
					BYTE dnet_grp;      // destination subnet or group
					BYTE dnode_nid[7];  // destination node or NEURON ID
				} rcv;
				struct {
					BYTE snet;          // source subnet
					BYTE snode;         // source node
					BYTE dnet;          // destination subnet
					BYTE dnode;         // destination node
					BYTE group;
					BYTE member;
					BYTE reserved[4];
				} resp;
			} adr;
		} address;
		ExplicitAddr addr;				// Network address
	};


	union {
		struct {
			BYTE	code;				// message code or selector MSB
			BYTE	data[239];
		} data;
		MsgData	 _data; 				// Message data
	};
} ExpAppBuf;
//#pragma pack( pop )

//---------------------------------------------------------------------------
// the following interface is not available outside the DLL ...
bool	L_Open			(HANDLE devhdl, BOOL bRemote);
bool	L_ReOpen		(HANDLE devhdl, BOOL bRemote);
bool	L_CheckLocal	(void);
void	L_Close			(void);
void	L_CloseLoad		(void);				// remove intermediate data

bool	L_DoScan		(BYTE didx);
bool	L_UploadLocal	(NODE *node);
bool	L_UploadNode	(NODE *node);

bool	L_SetNodeMode	(BYTE *nid, BYTE mode, BYTE state);
bool	L_SetNodeMode	(BYTE didx, BYTE subnet, BYTE nodenr,
						 BYTE mode, BYTE state);

bool	L_QueryStatus	(BYTE *nid, BYTE *status);
bool	L_QueryStatus	(BYTE didx, BYTE subnet, BYTE nodenr, BYTE *status);

bool	L_ClearStatus	(BYTE *nid);

bool	L_Wink			(BYTE *nid);
bool	L_Wink			(BYTE didx, BYTE subnet, BYTE nodenr);

short	L_NVFetch		(BYTE *nid, WORD index, BYTE* buf);
short	L_NVFetch		(BYTE didx, BYTE subnet, BYTE nodenr,
						 WORD index, BYTE* buf);

bool	L_NVSet			(BYTE *nid, WORD sel_dir, WORD len, BYTE *data);
bool	L_NVSet			(BYTE didx, BYTE subnet, BYTE nodenr,
						 WORD sel_dir, WORD len, BYTE *data);

int		L_ReadMem		(BYTE *nid, WORD adr, short size, BYTE mode, BYTE *buf);
short	L_ReadMem		(BYTE didx, BYTE subnet, BYTE nodenr,
						 WORD adr, short size, BYTE mode, BYTE *buf);

bool	L_WriteMem		(BYTE *nid, WORD adr, short size, BYTE mode, BYTE *buf);
bool	L_WriteMem		(BYTE didx, BYTE subnet, BYTE nodenr,
						 WORD adr, short size, BYTE mode, BYTE *buf);

bool	L_SendMsg		(BYTE didx, BYTE subnet, BYTE nodenr,
						 BYTE svr, BYTE *data, BYTE len);

bool	L_UpdDomain		(BYTE *nid, BYTE didx, BYTE domlen, BYTE *id,
						 BYTE subnet, BYTE nodenr);

bool	L_SendSvp		(BYTE *nid, BYTE *pid);

ComplType L_LdvSndRcv	(ExpAppBuf *pout, short olen,
						 ExpAppBuf *pin,  short *ilen);
ComplType L_LdvWrite	(ExpAppBuf *pout, short olen);

// am 21.5.2003 ergnzt von Ingo Grothues
bool	L_UpdateAddress(BYTE *nid, BYTE addr_idx, ADTBL *entry);
bool	L_UpdateNVconfig(BYTE *nid, CNV *nv);
bool	L_QueryStatus2	(BYTE *nid, BYTE *status);
bool	L_QueryStatus2	(BYTE didx, BYTE subnet, BYTE nodenr, BYTE *status);

//---------------------------------------------------------------------------

#endif
