/****************************************************************************/
/*	  File:	pers_ram.c								   $Revision: 2 $		*/
/*									    									*/
/*	  Funktionen fuer persistentes RAM										*/
/*	  																		*/
/* 								    										*/
/*	  VSS-comments:									   Org: 9.9.99 Sam		*/
/* 	  $Log: /Viessmann_01/SW/LdvHA/PERS_RAM.C $								*/
/*																			*/
/*2     9.10.99 19:57 Samlowski												*/
/*Schreibfehler beseitigt													*/
/*																			*/
/*1     22.09.99 16:53 Samlowski											*/
/****************************************************************************/
/* (c) Copyright Gesytec GmbH 1993, 1997, 1999 All rights reserved.		    */
/****************************************************************************/

/****************************** Includes ************************************/

#include "windows_compat.h"
#include "ni_msg.h"
#include "ni_host.h"
#include "ni_mgmt.h"
#include "nm_codes.h"
#include "pers_ram.h"
#include "xDbg.h"

// Workaround (sonst Compilerfehler wegen Macro-Redefine)
#undef __NFDBITS
#undef __FDMASK

#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#define FLAG_ADDR_HIGH		(UCHAR)0xF0
#define FLAG_ADDR_LOW		(UCHAR)0xFE
#define NVCBASE_ADDR_HIGH	(UCHAR)0xF1
#define NVCBASE_ADDR_LOW	(UCHAR)0x00

#define NVC_LEN				sizeof(nv_struct)

extern	bool ldv_LocalOut(UCHAR _code, UCHAR* _data, short len);
extern	bool ldv_LocalReq(UCHAR _code, UCHAR* _data, short len, void* indata);

static	bool _PersistPut( UCHAR addrhigh, UCHAR addrlow, void* src,  int len );  
static	bool _PersistGet( UCHAR addrhigh, UCHAR addrlow, void* dest, int len );  
static  bool _NVCfgPut(short idx, void *pnvc, int len); 
static  bool _NVCfgGet(short idx, void *pnvc, int len); 
//---------------------------------------------------------------------------

bool PersistWrite( int pramid, void* src, int len ) 
{
//	DBG(0, "PersistWrite");
	if( pramid == PRAM_STARTFLAGS ){
		return( _PersistPut( FLAG_ADDR_HIGH, FLAG_ADDR_LOW, src, len ) ) ;
	}
	else if( pramid >= PRAM_NVCBASE ){
		return( _NVCfgPut(pramid-PRAM_NVCBASE, src, len) ) ;
	}
	return( FALSE ) ;
}

//---------------------------------------------------------------------------

bool PersistRead( int pramid, void* dest, int len ) 
{
//	DBG(0, "PersistRead");
	if( pramid == PRAM_STARTFLAGS ){
		return( _PersistGet( FLAG_ADDR_HIGH, FLAG_ADDR_LOW, dest, len ) ) ;
	}
	else if( pramid >= PRAM_NVCBASE ){
		return( _NVCfgGet(pramid-PRAM_NVCBASE, dest, len ) ) ;
	}
	return( FALSE ) ;
}	

//---------------------------------------------------------------------------

static bool _PersistPut(UCHAR addrhigh, UCHAR addrlow, void* src, int len)  
{
	DBG(0, "+PersistPut");
	UCHAR buf[25] ;
   
	if( len > 20 ) return( FALSE ) ;

    /* Write Neuron Memory */
    buf[ 0 ] = (UCHAR)NM_ABSOLUTE ;       /* mode Absolute                */
    buf[ 1 ] = addrhigh           ;       /* Offset highbyte              */
    buf[ 2 ] = addrlow       	  ;       /* Offset lowbyte               */
    buf[ 3 ] = (UCHAR)len         ;       /* Length                       */
    buf[ 4 ] = 1          		  ;       /* form Both_cs_Recalc          */
    memcpy( &buf[5], src, len )  ;
    if (!ldv_LocalOut(NM_write_memory, buf, 5 + len)) {
        return( FALSE ) ;         /* Fehler bei Write Memory				*/
    }
	DBG(0, "-PersistPut");
	return( TRUE ) ;
}

//---------------------------------------------------------------------------

static bool _PersistGet( UCHAR addrhigh, UCHAR addrlow, void* dest, int len )  
{
	DBG(0, "+PersistGet");
	UCHAR buf[4] ;
    
    if( len > 20 ) return( FALSE ) ;

    /* Read Neuron Memory */
    buf[0] = (UCHAR)NM_ABSOLUTE ;
    buf[1] = addrhigh ;                                     /* Offset high  */
    buf[2] = addrlow ;                                      /* Offset low   */
    buf[3] = (UCHAR)len ;         							/* Length       */

    if( !ldv_LocalReq(NM_read_memory, buf, 4, dest ) ){
        return( FALSE ) ;      /* keine/falsche Antwort erhalten            */
    }
	DBG(0, "-PersistGet");
	return (true);
}

//---------------------------------------------------------------------------

static bool _NVCfgPut(short idx, void *pnvc, int len) {
	HANDLE	file;

	file = open("GesyFlash/NVCfg.dat", O_RDWR | O_CREAT);
	if(file < 0)
	{
		fprintf(stderr, "Error opening file NVCfg.dat\r\n");
		perror("OpenNVCfg failed");
		return false;
	}
/*
	file = CreateFile(  TEXT("GesyFlash\\NVCfg.dat"),	// Open file
						GENERIC_WRITE,				// Open for write
						0,							// do not Share 
						NULL,						// No security
						OPEN_ALWAYS,				// Existing or not
						FILE_ATTRIBUTE_NORMAL,		// Normal file
						NULL);						// No template file 
	if (file == INVALID_HANDLE_VALUE) {
		RETAILMSG(1, (TEXT("Error opening file NVCfg.dat\r\n")));
		file = NULL;
		DBGERR("OpenNVCfg failed");
		return (false);
	}
*/


	DWORD dwBytesWritten;
	
	//SetFilePointer (file, idx * len, NULL, FILE_BEGIN);
	lseek(file, len, SEEK_SET);
	
	dwBytesWritten = write(file, pnvc, len);
	if(dwBytesWritten < 0)
	{
		perror("WriteNVCfg failed");
		close(file);
		return (DBGERR("WriteNVCfg failed"));
	}
/*    
    if (!WriteFile (file, pnvc, len, &dwBytesWritten, NULL)){
		CloseHandle(file);
		return (DBGERR("WriteNVCfg failed"));
	}
	CloseHandle(file);
//	delete file;
*/
	close(file);
	return (true);
}

//---------------------------------------------------------------------------

static bool _NVCfgGet(short idx, void *pnvc, int len) {
	DWORD	rlen;
	HANDLE	file;
	memset(pnvc,0,len);

	file = open("GesyFlash/NVCfg.dat", O_RDONLY);
	if(file < 0)
	{
		fprintf(stderr, "Error opening file NVCfg.dat\r\n");
		perror("OpenNVCfg failed");
		return true;					// return (true) ???
	}
/*
	file = CreateFile(  TEXT("GesyFlash\\NVCfg.dat"),	// Open file
						GENERIC_READ,				// Open for reading
						FILE_SHARE_READ,			// Share for reading
						NULL,						// No security
						OPEN_EXISTING,				// Existing file only
						FILE_ATTRIBUTE_NORMAL,		// Normal file
						NULL);						// No template file 
	if (file == INVALID_HANDLE_VALUE) {
//		RETAILMSG(1, (TEXT("Error opening file NVCfg.dat\r\n")));
		file = NULL;
		DBGERR("OpenNVCfg failed");
		return (true);
	}
*/

	if(lseek(file, len, SEEK_SET) < 0) 
		perror("Seek file NVCfg failed");
/*
	if (0xFFFFFFFF == SetFilePointer (file, idx * len, NULL, FILE_BEGIN)) {
		DBGERR("Seek file NVCfg failed");
	}
	if (!ReadFile(file, pnvc, len, &rlen, NULL)) {
		DBGERR("Read file NVCfg failed");
	}
*/
	
	rlen = read(file, pnvc, len);
	if(rlen <= 0)
		perror("Read file NVCfg failed");

//	CloseHandle(file);
	close(file);
	return (true);
}
