// raw-tcpip-lon-service

#include "windows_compat.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <signal.h>
#include <string.h>
#include <time.h>

#include "ethernet.h"


#define OPEN_LON_ON_START 0

/*
typedef struct
{
	unsigned short id;
	unsigned short len;
	char msg[256];
} MSG_BUF;
*/

int ListenSocket = -1, hSocket = -1;

void signalHandler(int signum)
{
    if(debugmode & 8) fprintf(stderr, "signalHandler(%d)\n", signum);

    if (hSocket >= 0)
    {
        struct linger so_linger;
        so_linger.l_onoff = TRUE;
        so_linger.l_linger = 0;
        if (setsockopt(hSocket, SOL_SOCKET, SO_LINGER, &so_linger, sizeof(so_linger)))
            perror("setsockopt(hSocket, SO_LINGER)");
        //shutdown(hSocket, SHUT_RDWR);

        //close(hSocket);
        //hSocket = -1;
    }

#if 01
    if (ListenSocket >= 0)
    {
        struct linger so_linger;
        so_linger.l_onoff = TRUE;
        so_linger.l_linger = 0;
        if (setsockopt(ListenSocket, SOL_SOCKET, SO_LINGER, &so_linger, sizeof(so_linger)))
            perror("setsockopt(ListenSocket, SO_LINGER)");
        //shutdown(ListenSocket, SHUT_RDWR);

        close(ListenSocket);
        ListenSocket = -1;
    }
#endif

    exit(EXIT_FAILURE);
}

int SetMySockOptions(int Socket)
    {
    int ret = TRUE;
    static int On = TRUE;

    if (setsockopt(Socket, SOL_SOCKET, SO_REUSEADDR,
        (char *)&On, sizeof(On) ) == SOCKET_ERROR)
        {
        perror("setsockopt(SO_REUSEADDR)");
        ret = FALSE;
        }

    /*
    If SO_KEEPALIVE is disabled (default), a TCP connection may remain
    idle until the connection is released at the protocol layer.  If
    SO_KEEPALIVE is enabled and the connection has been idle for   !!! two hours, !!!
    TCP sends a packet to the remote socket, expecting the remote
    TCP to acknowledge that the connection is still active.  If the remote
    TCP does not respond in a timely manner, TCP continues to send
    keepalive packets according to the normal retransmission algorithm. If
    the remote TCP does not respond within a particular time limit, TCP
    drops the connection. The next socket system call (for example,
    recv()) returns an error, and errno is set to ETIMEDOUT.
    */
    if (setsockopt(Socket, SOL_SOCKET, SO_KEEPALIVE,
        (char *)&On, sizeof(On) ) == SOCKET_ERROR)
        {
        perror("setsockopt(SO_KEEPALIVE)");
        ret = FALSE;
        }

    if (setsockopt(Socket, IPPROTO_TCP, TCP_NODELAY,
        (char *)&On, sizeof(On) ) == SOCKET_ERROR)
        {
        perror("setsockopt(TCP_NODELAY)");
        ret = FALSE;
        }

    return ret;
    }


int CreateListenSocket(const char *service, const char *proto)
    {
    int port, ListenSocket;
    size_t  addresslen;
    struct servent  *serv;
    struct sockaddr_in  address;
    //int       i;

    address.sin_family = AF_INET;
    address.sin_addr.s_addr = htonl(INADDR_ANY);
    port = atoi( service );
    if( port ) address.sin_port = htons(port);
    else
        {
        serv = getservbyname( service, proto );
        if( !serv ) return -2;
        address.sin_port = serv->s_port;
        }
    ListenSocket = socket(AF_INET, SOCK_STREAM, 0);
    if(ListenSocket < 0)
    {
        perror("socket()");
        return -4;
    }

    //i = 1;
    //setsockopt(ListenSocket, SOL_SOCKET, SO_REUSEADDR, (char *)&i, sizeof(i));

    addresslen = sizeof(struct sockaddr_in);
    if(bind(ListenSocket, (struct sockaddr *)&address, addresslen))
        {
        perror("bind()");
        close(ListenSocket);
        return -5;
        }
    SetMySockOptions(ListenSocket);
    if(listen(ListenSocket, 10))
        {
        perror("listen()");
        close(ListenSocket);
        return -6;
        }
    return ListenSocket;
    }


char *hexdump(char *dp, char *sp, short n, char sep, int pos)
    {
    char    c, h;
    //int pos = -4;
    while (n--)
        {
        pos++;
        c = *sp++;                      // aktuelles Zeichen

        h = (c >> 4) & 0xF;             // Bits 4..7
        if (h <= 9) *dp++ = h + 0x30;   // Ziffer als 0-9 ausgeben
        else        *dp++ = h + 0x37;   // Ziffer als A-F ausgeben

        h = c & 0xF;                    // Bits 0..3
        if (h <= 9) *dp++ = h + 0x30;   // Ziffer als 0-9 ausgeben
        else        *dp++ = h + 0x37;   // Ziffer als A-F ausgeben

        if (sep)    *dp++ = sep;        // ggfs. Trennzeichen einf�gen
        if ( ((pos & 0x0f) == 0) && sep)    *dp++ = sep;        // ggfs. extra Trennzeichen einf�gen
        }
    *dp = 0;        // terminieren
    return (dp);
    }


int listen_and_serve(const char *DevName, const char *service)
    {
    int hLdv = -1;
    struct sockaddr_in address;
    socklen_t addrlen = sizeof(struct sockaddr_in);
    struct pollfd fds[2];
    int result;

    IP_MESSAGE l2tbuf;
    int l2tcount, l2tpos;
    IP_MESSAGE t2lbuf;
    int t2lcount, t2lpos, t2lstate;


    signal(SIGPIPE, SIG_IGN);
    signal(SIGINT,  signalHandler);  // ^C
    signal(SIGTERM, signalHandler);
#if OPEN_LON_ON_START
    hLdv = open(DevName, O_RDWR);
    if(hLdv < 0)
        {
        fprintf(stderr, "Could not open %s", DevName);
        perror(": ");
        return -1;
        }
#endif
//    int LonClosed = FALSE;

    ListenSocket = CreateListenSocket(service, "tcp");
    if(ListenSocket < 0)
        {
        perror("Could not open ListenSocket");
        return -1;
        }
    SetMySockOptions(ListenSocket);

    if(debugmode & 8) fprintf(stderr, "hSocket opened.\n");

    hSocket = -1;
    while( 1 )
        {
		int Alive_Timeout = Alive_Timeout_Init;
		if (hSocket >= 0)
            {
			struct linger so_linger;
			so_linger.l_onoff = TRUE;
			so_linger.l_linger = 0;
			if (setsockopt(hSocket, SOL_SOCKET, SO_LINGER, &so_linger, sizeof(so_linger)))
				perror("setsockopt(hSocket, SO_LINGER)");
            close(hSocket);
			hSocket = -1;
			//sleep(1);
            if(debugmode & 8) fprintf(stderr, "Connection closed.\n");
            }
/*
        if (LonClosed)
            {
            if(debugmode & 8) fprintf(stderr, "Lon closed.\n");
            break;
            }
*/
        l2tcount = 0;
        l2tpos = 0;
        t2lstate = 0;
        t2lpos = 0;
        t2lcount = sizeof(IP_MESSAGE_HEADER);
        hSocket = accept(ListenSocket, (struct sockaddr *)&address, &addrlen);
        if(debugmode & 8) fprintf(stderr, "New Connection established.\n");
        SetMySockOptions(hSocket);

#if !OPEN_LON_ON_START
#if 01
        hLdv = open(DevName, O_RDWR);
        if(hLdv < 0)
            {
            fprintf(stderr, "Could not open %s", DevName);
            perror(": ");
            sleep(1);
            continue;
            //return -1;
            }
#else
		hLdv = socket(AF_PACKET, SOCK_RAW, htons(0x8950));
        if(hLdv < 0)
            {
            fprintf(stderr, "Could not create socket");
            perror(": ");
            sleep(1);
            continue;
            //return -1;
            }
		struct ifreq ifr = { 0 };
		size_t ifnamelen = strlen(DevName);
		memcpy(ifr.ifr_name, DevName, ifnamelen);
		if (ioctl(hLdv, SIOCGIFINDEX, &ifr) < 0)
		{
            perror("ioctl(SIOCGIFINDEX)");
            close(hLdv);
            sleep(1);
            continue;
			//return errno;
		}

		struct sockaddr_ll addr = { 0 };
		addr.sll_family = AF_PACKET;
		addr.sll_ifindex = ifr.ifr_ifindex;

		if (bind(hLdv, (struct sockaddr*) &addr, sizeof(struct sockaddr_ll)) < 0)
		{
            perror("bind()");
            close(hLdv);
            sleep(1);
            continue;
			//return errno;
		}
#endif
/*
            {
            char msg[] = { 0x50, 0};
            write(hLdv, &msg, 2);
            }
*/
#endif
        int     ReadNoAliveCount    = 0;
        int     WriteAliveCount     = 0;
        time_t  LastAliveTime       = time(NULL);

		if (Alive_Timeout != ALIVE_TIMEOUT)
		{  // -> send ID_ALIVE_TIMEOUT packet
			IP_ALIVE_TIMEOUT IpAliveTimeout;
			IpAliveTimeout.h.Id = cpu_to_le16(ID_ALIVE_TIMEOUT);
			IpAliveTimeout.h.Len = cpu_to_le16(sizeof(DWORD));
			IpAliveTimeout.AliveTimeout = cpu_to_le32(Alive_Timeout * 1000);
			if (debugmode & 2)
				fprintf(stderr, "Sending ID_ALIVE_TIMEOUT-Datagram.\n");

			result = write(hSocket, ((char *) &IpAliveTimeout), sizeof(IpAliveTimeout));
			if (result <= 0)
			{
				if ((result < 0) && (debugmode & 8))
				{
					fprintf(stderr, "Sending ID_ALIVE_TIMEOUT-Datagram to hSocket failed.\n");
					perror("write");
				}
	            close(hLdv);
	            sleep(1);
				continue;
			}
		}

        while( 1 )
            {
            time_t  DiffTime = time(NULL) - LastAliveTime;
            if (DiffTime > Alive_Timeout)
                {
                LastAliveTime = time(NULL);
                ReadNoAliveCount++;     WriteAliveCount++;
                if (debugmode & 8)
                    fprintf(stderr, "ReadNoAliveCount=%d WriteAliveCount=%d DiffTime=%ld\n",
                        ReadNoAliveCount, WriteAliveCount, DiffTime);
                if (ReadNoAliveCount >= MAX_READ_NOALIVE_COUNT)
                    {
                    break;
                    }
                if (WriteAliveCount >= MAX_WRITE_ALIVE_COUNT)
                    {  // -> send ID_ALIVE packet
                    l2tbuf.h.Id  = cpu_to_le16(ID_ALIVE);
                    l2tbuf.h.Len  = cpu_to_le16(0);
                    if(debugmode & 2) fprintf(stderr, "Sending ID_ALIVE-Datagram.\n");

                    result = write(hSocket, ((char *)&l2tbuf), 4);
                    if(result <= 0)
                        {
                        if((result < 0)&&(debugmode & 8))
                            {
                            fprintf(stderr, "Sending ID_ALIVE-Datagram to hSocket failed.\n");
                            perror("write");
                            }
                        break;
                        }
                    WriteAliveCount = 0;
                    }
                }

            fds[0].fd = hSocket;
            fds[0].revents = 0;
            fds[1].fd = hLdv;
            fds[1].revents = 0;
            fds[0].events = POLLIN;
            fds[1].events = 0;
            if( l2tcount )
                fds[0].events |= POLLOUT;
            else
                fds[1].events = POLLIN;

            result = poll(fds, 2, Alive_Timeout*500);   // poll timeout (ms) half of Alive_Timeout (s)
            if (debugmode & 8) fprintf(stderr, "poll()=%d: ", result);

            if(result < 0)  ///// ERROR /////
                {
                if(debugmode & 8)
                    {
                    fprintf(stderr, "poll() returned errorcode.\n");
                    perror("poll");
                    }
                break;
                }
            else if(result == 0) //// TIMEOUT ////
                {
                }
            else    // Event on socket
                {
                if( fds[0].revents & POLLIN )
                    {
                    ReadNoAliveCount = 0;
                    if( !t2lstate )     // msg starts
                        {
                        result = read(hSocket, ((char *)&t2lbuf.h.Id) + t2lpos, t2lcount);
                        if(result <= 0)
                            {
                            if((result < 0)&&(debugmode & 8))
                                {
                                fprintf(stderr, "Reading datagram from hSocket failed. (t2lstate==0)\n");
                                perror("read");
                                }
                            break;
                            }

                        t2lpos += result;
                        t2lcount -= result;

                        if( !t2lcount )     // msg header complete
						  {
							switch (le16_to_cpu(t2lbuf.h.Id))
							{
							case ID_LON:
								//if (debugmode & 1)  fprintf(stderr, "<  ");
								break;
							case ID_ALIVE:
								if(debugmode & 2) fprintf(stderr, "ID_ALIVE-Datagram received.\n");
								break;
							case ID_ALIVE_TIMEOUT:
								if(debugmode & 2) fprintf(stderr, "ID_ALIVE_TIMEOUT-Datagram received.\n");
								break;
							default:  // unknown ID
								if(debugmode & 2) fprintf(stderr, "Datagram with ID=%i received.\n", le16_to_cpu(t2lbuf.h.Id));
								break;
							}
							t2lpos = 0;
							t2lcount = le16_to_cpu(t2lbuf.h.Len);
							if (t2lcount > sizeof(t2lbuf.Msg))
									break;  // prevent buffer overflow -> close connection
							if (t2lcount)
								t2lstate = 1;
							else
								t2lcount = sizeof(IP_MESSAGE_HEADER);
						  }
                        }
                    else    // msg continues
                        {
                        result = read(hSocket, ((char *)&t2lbuf.Msg) + t2lpos, t2lcount);
                        if(result <= 0)
                            {
                            if((result < 0)&&(debugmode & 8))
                                {
                                fprintf(stderr, "Reading datagram from hSocket failed. (t2lstate==1)\n");
                                perror("read");
                                }
                            break;
                            }

                        t2lpos += result;
                        t2lcount -= result;
                        if(debugmode & 1)
                            {
                            char dump[(t2lpos+4)*3+1];
                            hexdump(dump, ((char *)&t2lbuf), t2lpos + 4, ' ', -4);
                            fprintf(stderr, ">> %s\n", dump);
                            }

                        if( !t2lcount )     // msg complete
						  {
							switch (le16_to_cpu(t2lbuf.h.Id))
							{
							case ID_LON:
								if (debugmode & 1)
								{
									char dump[t2lpos * 3 + 1];
									//fprintf(stderr, "<  ");
									hexdump(dump, ((char *) &t2lbuf.Msg), t2lpos, ' ', 0);
									fprintf(stderr, ">  %s\n", dump);
								}

								if (debugmode & 8)
									fprintf(stderr, "Sending %d bytes to LON-Interface.\n", t2lpos);
//                            int Loop = 0;
//SendAgain:
								errno = 0;
								/*ssize_t Anz =*/ write(hLdv, &t2lbuf.Msg, t2lpos);
#if 0
								if (Anz != t2lpos)
								{
									fprintf(stderr, "Loop=%2d: write(%d)=%d errno=%d\n", Loop, t2lpos, Anz, errno);
									if (errno == ENOBUFS && ++Loop <= 1)
									{
										usleep(100000);
										goto SendAgain;
									}
								}
#endif

								break;
							case ID_ALIVE_TIMEOUT:
								{
								PIP_ALIVE_TIMEOUT AliveTimeout = (PIP_ALIVE_TIMEOUT) &t2lbuf;
								if (debugmode & 2)
									fprintf(stderr, "ID_ALIVE_TIMEOUT = %d\n", le32_to_cpu(AliveTimeout->AliveTimeout));
								Alive_Timeout = le32_to_cpu(AliveTimeout->AliveTimeout) / 1000;
								break;
								}
							}
							t2lstate = 0;
							t2lpos = 0;
							t2lcount = sizeof(IP_MESSAGE_HEADER);
						  }
                        }
                    }


                if( fds[0].revents & POLLOUT )
                    {
                    if(debugmode & 1)
                        {
                        char dump[l2tcount*3+1];
                        hexdump(dump, ((char *)&l2tbuf) + l2tpos, l2tcount, ' ', -4);
                        fprintf(stderr, "<< %s\n", dump);
                        }

                    result = write(hSocket, ((char *)&l2tbuf) + l2tpos, l2tcount);
                    if(result <= 0)
                        {
                        if((result < 0)&&(debugmode & 8))
                            {
                            fprintf(stderr, "Sending ID_LON-Datagram to hSocket failed.\n");
                            perror("write");
                            }
                        break;
                        }
                    WriteAliveCount = 0;
                    l2tpos += result;
                    l2tcount -= result;
                    }


                if( fds[1].revents & POLLHUP)
                    {
                    if(debugmode & 8) fprintf(stderr, "Received POLLHUP from LON-Interface:");
//                    LonClosed = TRUE;
                    break;
                    }

                if( fds[1].revents & POLLIN )
                    {
                    if(debugmode & 8) fprintf(stderr, "Received data from LON-Interface:");
                    l2tcount = read(hLdv, &l2tbuf.Msg, 256);
                    if(debugmode & 8)  fprintf(stderr, "=%d\n", l2tcount);
                    if(debugmode & 1)
                        {
                        char dump[l2tcount*3+1];
                        hexdump(dump, ((char *)&l2tbuf.Msg), l2tcount, ' ', 0);
                        fprintf(stderr, "<  %s\n", dump);
                        }
                    if (l2tcount > 0)
                        {
                        l2tbuf.h.Len = cpu_to_le16(l2tcount);
                        l2tcount += 4;
                        l2tbuf.h.Id  = cpu_to_le16(ID_LON);
                        l2tpos = 0;
                        }
                    else
                        l2tcount = 0;
                    }
                }
            }  // while( 1 )
#if !OPEN_LON_ON_START
            close(hLdv);
#endif
        }
    return 0;
    }
