//***************************************************************************
//
// Module Name:	upload.cpp
// Abstract:	Upload and other LON routines
//				(based on upload from R.Feiland)
//
//***************************************************************************

#include <unistd.h>		// fuer usleep(), fork(), setsid() und chdir()
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/poll.h>

#include "windows_compat.h"
#include "ethernet.h"
#include "nm_codes.h"
#include "ni_privat.h"
#include "ni_host.h"
#include "xDbg.h"
#include "Utils.h"
#include "yaList.h"
#include "fifo.h"
#include "DebugMsg.h"

extern	CNet 		NETWORK;

#define LDV_TIMEOUT			180000				// 180 seconds
//#define LDV_TIMEOUT				20000
#define DEFAULT_RPTRETRY	0x43				// default repeat / retry
#define DEFAULT_TXTIMER		0x05				// transmit timer

//***************************************************************************
// following definitions are for "auto_configure" the host adapter
// it is used, if the host adapter neuron is not CONFIG_ONLINE
// Default domain (Index 0) address
#define	DSUBNET		142				// 1..255
#define	DNODE		123				// 1..127
#define	DDLEN		0				// 0, 1, 3 or 6
// Work domain (Index 1) address
#define WSUBNET		142				// 1..255
#define	WNODE		124				// 1..127
#define	WDLEN		1				// 0, 1, 3 or 6
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// node upload flags ...
#define	NFLAG_POLL	 	0x0001      // do node poll
#define VALID_ROD		0x0010
#define VALID_LOC		0x0020
#define VALID_DOM		0x0040
#define VALID_ADR		0x0080
#define VALID_CFG		0x0100
#define VALID_SNV		0x0200
#define VALID_FIX		0x0400
#define	NFLAG_BIND		0x2000      // bind this node
#define	NFLAG_UPLD		0x4000      // error: node fully uploaded
#define	NFLAG_UNAV		0x8000      // error: node unavailable

// NV upload flags ...
#define NVFLAG_SET			1		// value is set (valid)
#define NVFLAG_SETS			2		// value is saved
#define NVFLAG_TOPC			4		// send value to PC
#define NVFLAG_CFG		   16		// modify nv config

#define RM_SNVT			 0xFF

//***************************************************************************
// Globals ...
		CRITICAL_SECTION gLdvLock;				// protect access

		HANDLE		gLP4		= INVALID_HANDLE_VALUE;
        BOOL        gRemote     = FALSE;        // use Rni Client
		//HANDLE		gInEvt		= NULL;			// incoming event
		HANDLE		gInEvt		= 0;			// incoming event
		CFifo		*gInMsgs	= NULL;			// msgs not handled here
		NIDSTRUCT	*gNIDs		= NULL;			// start of NID list
		short		gNIDcount	= 0;			// # of known neurons

		BYTE		gRPTRETRY	= DEFAULT_RPTRETRY;
		BYTE		gTXTIMER	= DEFAULT_TXTIMER;

// read thread synchronisation ...
static	short		gRun		= 2;			// thread should run
//static	HANDLE		gRespEvt	= NULL;			// response event
static	HANDLE		gRespEvt	= 0;			// response event
static	BYTE		gCommWait	= 0;			// waiting for neuron answer
static	short		gGotData	= 0;			// got data answer
static	ComplType	gCplCode;					// completion code of answer
static	BYTE		gNi_reset[]	= { 0x50, 0 };

static	BYTE		*gBuf		= NULL;			// scratch buffer
static	ExpAppBuf	gMsgOut, gMsgRsp;

//---------------------------------------------------------------------------
// some local prototypes ...
static	void		Ni_Reset		(void);

static	DWORD		ReadThread		(DWORD arg);
static	void		ProcessMsg		(ExpAppBuf *pMsg, DWORD len);
static	void		ProcessIncoming	(ExpAppBuf *pMsg, DWORD len);

static	ComplType	send_msg		(short len);
static	ComplType	send_local		(short len);
static	ComplType	r2q_off			(BYTE *nid, BYTE didx);
static	void		r2q_off_known	(void);
static	void		nid2header		(BYTE *nid, BYTE svc);
static	BYTE		*getMem			(long size);
static	bool		add_nid			(BYTE didx, BYTE *nidpid);
static	int			readmem			(NODE *node,
									 WORD adr, short size, BYTE mode, BYTE *buf);

#define NODENV(x) ((NVSTRUCT *)(node->panvs + (x * sizeof(NVSTRUCT))))

//***************************************************************************

bool L_ReOpen(HANDLE devhdl, BOOL bRemote)
{
	CLockCrit lock(&gLdvLock);				// protect access

	gLP4 = devhdl;
    gRemote = bRemote;
	return (true);
}

bool L_Open(HANDLE devhdl, BOOL bRemote)
{
	InitializeCriticalSection(&gLdvLock);
	CLockCrit lock(&gLdvLock);				// protect access

	gLP4 = devhdl;
    gRemote = bRemote;
	Ni_Reset();								// reset local node

	gInMsgs	 = new CFifo(8000);
	gInEvt   = CreateEvent(NULL, false, false, NULL);
	gRespEvt = CreateEvent(NULL, false, false, NULL);
	CreateThread((LPVOID) &ReadThread, 0);			// permanently reads from neuron

	return (true);
}

//---------------------------------------------------------------------------

void L_Close(void)
{
	gRun = 1;								// evtl. stops read thread
	L_CloseLoad();							// removes upload resources
	gLP4 = INVALID_HANDLE_VALUE;
	DeleteCriticalSection(&gLdvLock);
	if (gInMsgs) delete gInMsgs;
	//if (gRespEvt) CloseHandle(gRespEvt);
	//if (gInEvt)   CloseHandle(gInEvt);
}

//---------------------------------------------------------------------------
ssize_t ldv_write(int fd, const void *buf, size_t count)
    {
#if 0
    char dbuf[500];
    sprintf(dbuf, "ldv_write(%i) (", count);
    hexdump(&dbuf[strlen(dbuf)], (BYTE *)buf, count, ' ');
    sprintf(&dbuf[strlen(dbuf)], ")\n");
    fprintf(stderr, "%s", dbuf);
#endif
    if (gRemote)
        {
        IP_MESSAGE IPBuffer;
        count = min(count, sizeof(IPBuffer.Msg));
        IPBuffer.h.Id = ID_LON;
        IPBuffer.h.Len = count;
        memcpy(IPBuffer.Msg, buf, count);
        return send(fd, &IPBuffer, IPBuffer.h.Len + IP_MESSAGE_HEADER_SIZE, 0);
        }
    else
        {
        int ret = write(fd, buf, count);
        //fprintf(stderr, "write(%d)=%d\n", count, ret);
        if (ret < 0)
            return ret;
        return count;
/*
        DWORD rLen;
        if (!WriteFile(gLP4, pout, olen, &rLen, NULL)) {
            DBGERR("LdvWrite failed");
            return (MSG_FAILS);
            }
*/
        }
    }
//---------------------------------------------------------------------------

ssize_t ldv_read(int fd, void *buf, size_t count)
    {
    if (gRemote)
        {
        static CHAR     ReadSocketMsg[sizeof(IP_MESSAGE)];
        static DWORD    RemainDataSize = 0;
        PUCHAR cBuffer = PUCHAR(buf);
        size_t dwBufferLen = count;
        ssize_t dwReadCount = 0;
        DWORD	Anz;
        int iInCount = 0;
        struct timeval  tSelectTime = {0,0};
        fd_set          SocketWatch;

        if (cBuffer == NULL && dwBufferLen == 0)
            {
            RemainDataSize = 0;  // Buffer leeren
            return 0; //MY_NOERROR;
            }

        //Pruefen ob Buffer min. LON-Paket Laenge
//        if (dwBufferLen < MAX_FRAME_LEN)
//            return 0; //BUFFER_TO_SHORT;

//        *dwReadCount = 0;

        //Pruefen ob noch Daten im Buffer sind
        if (RemainDataSize >= IP_MESSAGE_HEADER_SIZE)
            {
            PIP_MESSAGE IpMsg = (PIP_MESSAGE) ReadSocketMsg;
            Anz = IpMsg->h.Len + IP_MESSAGE_HEADER_SIZE;
            //Pruefen ob noch ein ganzes Paket im Buffer ist.
            if (Anz > RemainDataSize)
                goto DoRecv;

            //Es liegt mindestens ein Buffer im LocalBuffer
            switch (IpMsg->h.Id)
                {
                case ID_LON:
                    //Buffer zur Uebergabe kopieren.
                    memcpy( cBuffer, &IpMsg->Msg, Anz-IP_MESSAGE_HEADER_SIZE);
                    dwReadCount = Anz-IP_MESSAGE_HEADER_SIZE;

                    //Level im Buffer setzen.
                    RemainDataSize -= Anz;
                    if (RemainDataSize > 0)
                        memmove(&ReadSocketMsg[0], &ReadSocketMsg[Anz], RemainDataSize);

                    return dwReadCount; //PACKET_RECEIVED;

                case ID_ALIVE:
					if(debugmode & 2) fprintf(stderr, "ID_ALIVE-Datagram received.\n");
                	break;
				case ID_ALIVE_TIMEOUT:
					if(debugmode & 2) fprintf(stderr, "ID_ALIVE_TIMEOUT-Datagram received.\n");
                	break;
                }
            RemainDataSize -= Anz;
            if (RemainDataSize > 0)
                memmove(&ReadSocketMsg[0], &ReadSocketMsg[Anz], RemainDataSize);

            return dwReadCount; //PACKET_RECEIVED;
            }

DoRecv:
            if (fd == INVALID_SOCKET)  return -1;
            //Pruefen, ob Daten vorhanden
            FD_ZERO(&SocketWatch);
            FD_SET (fd, &SocketWatch);
            // printf("select(%d):\n", fd);
            int ev = select(fd+1, &SocketWatch, NULL, NULL, &tSelectTime);
            // printf("ldv_read(): select(%d)=%d\n", fd, ev);

            // Daten hinter den Rest schreiben.
            //Daten von TCP holen
            if (FD_ISSET (fd, &SocketWatch))
                iInCount = recv(fd, &ReadSocketMsg[RemainDataSize],
                (sizeof(ReadSocketMsg) - RemainDataSize), 0);
                /*    printf("recv()=%d\n", iInCount);
                if (iInCount > 0)
                {
                int i;
                for (i=0; i<iInCount; i++)
                printf("%02X ", ReadSocketMsg[RemainDataSize+i]);
                printf("\n");
                }
            */
            /* Daten liegen jetzt im Buffer */

            //Pruefen ob das lesen erfolgreich war.
            if( iInCount == INVALID_SOCKET )
                {
                dwReadCount = -1;
                return -1; //ERROR_TCP_READ;
                }
            else if( iInCount == 0)
                return 0; //ERROR_TCP_READ;

            RemainDataSize += iInCount;

            return 0; //NEED_MORE_DATA;
        }
    else
        {
        return read(fd, buf, count);
/*
        DWORD rLen;
        if (!WriteFile(gLP4, pout, olen, &rLen, NULL)) {
            DBGERR("LdvWrite failed");
            return (MSG_FAILS);
            }
*/
        }
    }
//---------------------------------------------------------------------------

static DWORD ReadThread(DWORD arg)					// ldv_read
{
	ssize_t     rLen = 0;
	ExpAppBuf	msg;
    struct timeval  tSelectTime = {0,0};
    fd_set          SocketWatch;

    IP_MESSAGE  l2tbuf;
    int         ReadNoAliveCount    = 0;
    int         WriteAliveCount     = 0;
    time_t      LastAliveTime       = time(NULL);
    int         result;
	int 		Alive_Timeout 		= Alive_Timeout_Init;

	if (gRemote && Alive_Timeout != ALIVE_TIMEOUT)
	{  // -> send ID_ALIVE_TIMEOUT packet
		IP_ALIVE_TIMEOUT IpAliveTimeout;
		IpAliveTimeout.h.Id = cpu_to_le16(ID_ALIVE_TIMEOUT);
		IpAliveTimeout.h.Len = cpu_to_le16(sizeof(DWORD));
		IpAliveTimeout.AliveTimeout = cpu_to_le32(Alive_Timeout * 1000);
		if (debugmode & 2)
			fprintf(stderr, "Sending ID_ALIVE_TIMEOUT-Datagram.\n");

		result = write(gLP4, ((char *) &IpAliveTimeout), sizeof(IpAliveTimeout));
		if (result <= 0)
		{
			if ((result < 0) && (debugmode & 8))
			{
				fprintf(stderr, "Sending ID_ALIVE_TIMEOUT-Datagram to hSocket failed.\n");
				perror("write");
			}
			//break;
		}
	}

	while (gRun == 2)
	{
        if (gRemote)
            {
            time_t  DiffTime = time(NULL) - LastAliveTime;
            if (DiffTime > Alive_Timeout)
                {
                LastAliveTime = time(NULL);
                ReadNoAliveCount++;     WriteAliveCount++;
                if (debugmode & 8)
                    fprintf(stderr, "ReadNoAliveCount=%d WriteAliveCount=%d DiffTime=%d\n",
                        ReadNoAliveCount, WriteAliveCount, DiffTime);
                if (ReadNoAliveCount >= MAX_READ_NOALIVE_COUNT)
                    {
//                    break;
                    }
                if (WriteAliveCount >= MAX_WRITE_ALIVE_COUNT)
                    {  // -> send ID_ALIVE packet
                    l2tbuf.h.Id  = ID_ALIVE;
                    l2tbuf.h.Len  = 0;
                    if(debugmode & 2) fprintf(stderr, "Sending ID_ALIVE-Datagram.\n");

                    result = write(gLP4, ((char *)&l2tbuf), 4);
                    if(result <= 0)
                        {
                        if((result < 0)&&(debugmode & 8))
                            {
                            fprintf(stderr, "Sending ID_ALIVE-Datagram to gLP4 failed.\n");
                            perror("write");
                            }
                        break;
                        }
                    WriteAliveCount = 0;
                    }
                }
            }
        errno = 0;
        FD_ZERO(&SocketWatch);
        FD_SET (gLP4, &SocketWatch);

        tSelectTime.tv_sec  = 1;
        tSelectTime.tv_usec = 0;
        // printf("select(%d):\n", fd);
        int ev = select(gLP4+1, &SocketWatch, NULL, NULL, &tSelectTime);
        // printf("select(%d)=%d\n", gLP4, ev);

		rLen = -1;
		if (ev < 0) continue;
        if (FD_ISSET(gLP4, &SocketWatch))
            {
TryNextBuf:
            rLen = ldv_read(gLP4, &msg, sizeof(msg));
            if (gRemote && rLen == 0)
                rLen = ldv_read(gLP4, &msg, sizeof(msg));

            /// fprintf(stdout, "read()=%i  gRun=%d\n", rLen, gRun);
            }

		if (rLen == 0) continue;
		else if(rLen == -1)
		{
			if (errno)  perror("Error while reading from /dev/lon");
            if (gRemote && (gRun == 2)
                && (errno == ECONNRESET || errno == ENOTCONN) )
                {
                NETWORK.Close();
                NETWORK.ReOpenRemoteNet();
//                _exit(1);
                }
		}
		if ((rLen >0) && (gRun == 2))
        {
//----------------------------------------------------------
			char dbuf[500];
			sprintf(dbuf, "ldv_read  [%i] (Header:", rLen);
			hexdump(&dbuf[strlen(dbuf)], (BYTE *)&msg, 5, 0);
			sprintf(&dbuf[strlen(dbuf)], "), (Address:");
			hexdump(&dbuf[strlen(dbuf)], &((BYTE *)&msg)[5], 11, 0);
			sprintf(&dbuf[strlen(dbuf)], "), (Code:");
			hexdump(&dbuf[strlen(dbuf)], &((BYTE *)&msg)[16], 1, 0);
			sprintf(&dbuf[strlen(dbuf)], ",NM_Code:0x");
			unsigned char code = msg.data.code;
			code = code & 0x1F;
			hexdump(&dbuf[strlen(dbuf)], &code, 1, 0);
			if(rLen > 17)
			{
				sprintf(&dbuf[strlen(dbuf)], "), (Data:");
				hexdump(&dbuf[strlen(dbuf)], &((BYTE *)&msg)[17], rLen-17, 0);
			}
			sprintf(&dbuf[strlen(dbuf)], ")\n");
			if (debugmode & 0x10)  fprintf(stderr, "%s", dbuf);

//-----------------------------------------------------------------

//			RETAILMSG(1,(TEXT("**********vor ProcessMsg\r\n")));
			ProcessMsg(&msg, rLen);
// 			Sleep(20);	// ????????????????????? ohne delay haengts hier ????
//			RETAILMSG(1,(TEXT("************nach ProcessMsg + Sleep(20)\r\n")));
            if (gRemote)  goto TryNextBuf;
		}
		else {
			// DBGERR("ldv_read rLen: 0");
		}
	}
	return (0);
}

//---------------------------------------------------------------------------
/*
// Literals for the 'cmd.q.queue' nibble of NI_Hdr.
typedef enum {
    niTQ		=  2, 			// Transaction queue
    niTQ_P		=  3, 			// Priority transaction queue
    niNTQ		=  4, 			// Non-transaction queue
    niNTQ_P		=  5, 			// Priority non-transaction queue
    niRESPONSE	=  6, 			// Response msg & completion event queue
    niINCOMING	=  8		    // Received message queue
} NI_Queue;

// Literals for the 'cmd.q.q_cmd' nibble of NI_Hdr.
typedef enum {
    niCOMM	     = 1,			// Data transfer to/from network
    niNETMGMT	 = 2			// Local network management/diagnostics
} NI_QueueCmd;

// Literals for the 'cmd.noq' byte of NI_Hdr.
typedef enum {
    niNULL			 = 0x00,
    niRESET			 = 0x50,
    niFLUSH_COMPLETE = 0x60,	// Uplink
    niFLUSH_CANCEL   = 0x60,	// Downlink
    niONLINE	     = 0x70,
    niOFFLINE	     = 0x80,
    niFLUSH			 = 0x90,
    niFLUSH_IGN      = 0xA0,
    niACK			 = 0xC0,
    niPUPXOFF	     = 0xE1,
    niPUPXON	     = 0xE2
} NI_NoQueueCmd;
*/

static void ProcessMsg(ExpAppBuf *pMsg, DWORD len)
{
	if (DBG(DBG_NET)) {
		char *dp, dbuf[300];
		sprintf(dbuf, "ProcessMsg cmd flags: ");
		dp = hexdump(&dbuf[strlen(dbuf)], (BYTE *)&pMsg->ni_hdr.noq.cmd, 1, 0);
		*dp++ = ' ';
		dp = hexdump(dp, (BYTE *)&pMsg->header.flags, 1, 0);
		DBGERR(dbuf);
	}

//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (pMsg->ni_hdr.noq.cmd == niRESET) {				// local reset
		DBGERR("local reset");
		if (gCommWait) {
			DBGERR("local reset someone waits");
			gCplCode  = MSG_FAILS;
			gCommWait = 0;
			SetEvent(gRespEvt);							// someone is waiting
		}
		return;
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (pMsg->ni_hdr.noq.cmd == (BYTE)((niCOMM << 4) + niINCOMING)) {
		ProcessIncoming(pMsg, len);
		return;
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (gCommWait == niCOMM) 	// answer messages from the network
	{
		if (pMsg->ni_hdr.noq.cmd == ((BYTE)((niCOMM << 4) + niRESPONSE)))
		{
			DBGERR("niCOMM/niRESPONSE !");
			BYTE cpl = pMsg->header.flags & 0x30;
			if (cpl)
			{
				gCplCode  = (ComplType)(cpl >> 4);		// completion event
				gCommWait = 0;
				SetEvent(gRespEvt);						// someone is waiting
			    return;
			}
			else
			{
				if ((pMsg->header.svc_tag & 0x60) == SVC_request)
				{
					DBGERR("niCOMM/SVC_request !");
					gGotData = (short)len;
					memcpy(&gMsgRsp, pMsg, len);
				}
			}
		}
		return;
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	else if (gCommWait == niNETMGMT) {	// answers from the local neuron
		if (pMsg->ni_hdr.noq.cmd == ((BYTE)((niCOMM << 4) + niRESPONSE))) {
			if ((pMsg->header.svc_tag & 0x60) == SVC_request) {
				DBGERR("niNETMGMT !");
				gGotData = (short)len;
				memcpy(&gMsgRsp, pMsg, len);
				gCplCode  = MSG_SUCCEEDS;
				gCommWait = 0;
				SetEvent(gRespEvt);						// someone is waiting
				return;
			}
			DBGERR("niNETMGMT ???");
		}
		return;
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	DBGERR("ProcessMsg no wait");
}

//---------------------------------------------------------------------------

static void	ProcessIncoming(ExpAppBuf *pMsg, DWORD len)
{
	//DBGERR("ProcessIncoming()");

	if (!gInMsgs) return;	// fifo not created

	if (!gInMsgs->Put((BYTE *)pMsg, (WORD)len)) {		// fifo full
		DBGERR("gInMsgs full");
		return;
	}

	SetEvent(gInEvt);	// inform handler (in LHA)
}

//---------------------------------------------------------------------------

ComplType L_LdvSndRcv(ExpAppBuf *pout, short olen,
					  ExpAppBuf *pin,  short *ilen)
// send a message and wait for an answer ...
{
	if (debugmode & 0x10)  fprintf(stderr, "\nL_LdvSndRcv()\n");
	CLockCrit	lock(&gLdvLock);						// protect access
	DWORD		dwWaitResult;
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
/*	if (DBG(DBG_NET)) {
		char dbuf[500];
		sprintf(dbuf, "ldv_write [%i] ", olen);
		hexdump(&dbuf[strlen(dbuf)], (BYTE *)pout, olen, 0);
		DBGERR(dbuf);
	}
*/
	char dbuf[500];
	sprintf(dbuf, "ldv_write [%i] (Header:", olen);
	hexdump(&dbuf[strlen(dbuf)], (BYTE *)pout, 5, 0);
	sprintf(&dbuf[strlen(dbuf)], "), (Address:");
	hexdump(&dbuf[strlen(dbuf)], &((BYTE *)pout)[5], 11, 0);
	sprintf(&dbuf[strlen(dbuf)], "), (Code:");
	hexdump(&dbuf[strlen(dbuf)], &((BYTE *)pout)[16], 1, 0);
	sprintf(&dbuf[strlen(dbuf)], ",NM_Code:0x");
	unsigned char code = pout->data.code;
	code = code & 0x1F;
	hexdump(&dbuf[strlen(dbuf)], &code, 1, 0);
	if(olen > 17)
	{
		sprintf(&dbuf[strlen(dbuf)], "), (Data:");
		hexdump(&dbuf[strlen(dbuf)], &((BYTE *)pout)[17], olen-17, 0);
	}
	sprintf(&dbuf[strlen(dbuf)], ")\n");
	if (debugmode & 0x10)  fprintf(stderr, "%s", dbuf);

//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	gCommWait = pout->ni_hdr.q.q_cmd;	// critical ???
	gGotData  = 0;						// no data received
	DWORD rLen;
//	RETAILMSG(1,(TEXT("**************WriteFile to LP4\r\n")));
/*	if (!WriteFile(gLP4, pout, olen, &rLen, NULL)) {
		DBGERR("ldv_write failed");
//		RETAILMSG(1,(TEXT("ldv_write failed\r\n")));
		return (MSG_FAILS);
	}
*/
	rLen = ldv_write(gLP4, pout, olen);
	if (rLen <= 0) {
		DBGERR("ldv_write failed");
//		RETAILMSG(1,(TEXT("ldv_write failed\r\n")));
		return (MSG_FAILS);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//	RETAILMSG(1,(TEXT("g2LonD +****WaitForSingleObject\r\n")));
	dwWaitResult = WaitForSingleObject(gRespEvt, LDV_TIMEOUT);
	if (dwWaitResult != WAIT_OBJECT_0) {
		DBGERR("WaitForSingleObject Net: timeout");
		return (MSG_FAILS);
	}

//	RETAILMSG(1,(TEXT("g2LonD -****WaitForSingleObject\r\n")));
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (ilen) {
		*ilen = gGotData;
		if (pin) memcpy(pin, &gMsgRsp, gGotData);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	return (gCplCode);
}

//---------------------------------------------------------------------------

ComplType L_LdvWrite(ExpAppBuf *pout, short olen)
{
// just send a message ...
	CLockCrit	lock(&gLdvLock);						// protect access
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (DBG(DBG_NET)) {
		char dbuf[500];
		sprintf(dbuf, "LdvWrite  ");
		hexdump(&dbuf[strlen(dbuf)], (BYTE *)pout, olen, 0);
		DBGERR(dbuf);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	DWORD rLen;
	rLen = ldv_write(gLP4, pout, olen);
	if (rLen <= 0) {
		DBGERR("LdvWrite failed");
		return (MSG_FAILS);
	}
	return (MSG_SUCCEEDS);
}

//---------------------------------------------------------------------------

static ComplType send_local(short len)
{
	ExpAppBuf *pout = &gMsgOut;

	pout->ni_hdr.q.q_cmd	= niNETMGMT;
	pout->ni_hdr.q.queue	= niTQ;
	pout->header.svc_tag	= SVC_request;
	pout->header.flags		= 8;					// explicit
    pout->ni_hdr.q.length	= len + 15;
    pout->header.data_len	= len + 1;

	return (L_LdvSndRcv(pout, len +17, NULL, NULL));
}

//---------------------------------------------------------------------------

static ComplType send_msg(short len)
{
	ExpAppBuf *pout = &gMsgOut;
//	0x12 = niCOMM + ...
//  niTQ	=  2, 			// Transaction queue
//  niTQ_P	=  3, 			// Priority transaction queue
    pout->ni_hdr.noq.cmd = (pout->header.flags & 0x80) ? 0x13 : 0x12;	// ???
	if ((pout->header.svc_tag & 0x60) == SVC_unackd) pout->ni_hdr.noq.cmd += 2;
    pout->ni_hdr.q.length = len + 15;
    pout->header.data_len = len + 1;

	return (L_LdvSndRcv(pout, len +17, NULL, NULL));
}

//---------------------------------------------------------------------------

bool L_CheckLocal(void)
// if the neuron is not CONFIG_ONLINE then it configures it to default values
//	(hopefully these values are not already in the network)
{
	ExpAppBuf *pout = &gMsgOut;

	pout->data.code = NM_query_status;
	if (debugmode & 0x10)  fprintf(stderr, "*****In checkLocal vor send_Local(0)\r\n");

	if (send_local(0) != MSG_SUCCEEDS)
	{
		fprintf(stderr, "******send_local(0) != MSG_SUCCEEDS\r\n");
		return (false);
	}
	if (debugmode & 0x10)  fprintf(stderr, "*****In checkLocal nach send_Local(0)\r\n");

	if (gMsgRsp.data.data[11] == 4)		   return (true);	// CONFIG_ONLINE
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	DBGERR("Local node not config_online");
//	RETAILMSG(1,(TEXT("Default set s/n %d %d \r\n"), DSUBNET, DNODE));	// ???
	BYTE ddomain[6] = {    0, 0,0,0,0,0 };	// Default Domain_ID
	BYTE wdomain[6] = { 0x01, 0,0,0,0,0 };	// Work Domain_ID

	if (!L_UpdDomain(NULL, 0, DDLEN, ddomain, DSUBNET, DNODE)) return (false);
	if (!L_UpdDomain(NULL, 1, WDLEN, wdomain, WSUBNET, WNODE)) return (false);
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	pout->data.code = NM_set_node_mode;
	pout->data.data[0] = CHANGE_STATE;						// Change state
	pout->data.data[1] = 4;									// Config online
	if (send_local(2) != MSG_SUCCEEDS) return (false);
	return (true);
}

//---------------------------------------------------------------------------

bool L_UploadLocal(NODE *node)
{
	ExpAppBuf *pout = &gMsgOut;

	node->num_domains = 2;
	node->num_adtbl   = 15;
	node->bigbuf      = TRUE;

	pout->data.code    = NM_read_memory;
	pout->data.data[0] = RM_RDONLY;
	pout->data.data[1] = 0;
	pout->data.data[2] = 0;
	pout->data.data[3] = 36;
	if (send_local(4) != MSG_SUCCEEDS) return (false);
	memcpy( node->nid, &gMsgRsp.data.data[0],  6 );			// Neuron ID
	memcpy( node->pid, &gMsgRsp.data.data[13], 8 );			// Program ID

	pout->data.data[0] = RM_CFGREL;
	pout->data.data[1] = 0;
	pout->data.data[2] = 2;
	pout->data.data[3] = 6;
	if (send_local(4) != MSG_SUCCEEDS) return (false);
	memcpy( node->location, &gMsgRsp.data.data[0], 6 );		// Location

	pout->data.data[1] = 0;
	pout->data.data[2] = 25;
	pout->data.data[3] = 30;
	if (send_local(4) != MSG_SUCCEEDS) return (false);
	memcpy( node->domain[0].id, &gMsgRsp.data.data[0], 30 );	// Domain tables
	node->domain[0].node &= 0x7F;
	node->domain[1].node &= 0x7F;
	return (true);
}

//---------------------------------------------------------------------------
// initialize header with neuron id ...
static void nid2header(BYTE *nid, BYTE svc)
{
	ExpAppBuf *pout = &gMsgOut;

	pout->header.flags				 = 8;						// Explicit
	pout->address.adr.send.rpt_retry = gRPTRETRY;
	pout->address.adr.send.tx_timer  = gTXTIMER;
    pout->address.adr.send.dnet_grp  = 0;						// Subnet 0
	pout->address.format			 = 2;						// Neuron ID
	pout->address.adr.send.dom_node  = 0;
	memcpy( pout->address.adr.send.nid, nid, 6 );
	pout->header.svc_tag = svc;
}


// initialize header with subnet / node ...
static void snd2header(BYTE idx, BYTE subnet, BYTE node, BYTE svc)
{
	ExpAppBuf *pout = &gMsgOut;

	pout->header.flags				 = 8;						// Explicit
	pout->address.adr.send.rpt_retry = gRPTRETRY;
	pout->address.adr.send.tx_timer  = gTXTIMER;
    pout->address.adr.send.dnet_grp  = subnet;					// Subnet
	pout->address.format			 = 1;						// Subnet/Node
	node &= 0x7F;
	if(idx) node |= 0x80;
	pout->address.adr.send.dom_node  = node;					// Domain idx & Node
	pout->header.svc_tag = svc;
}

//---------------------------------------------------------------------------

static ComplType r2q_off(BYTE *nid, BYTE didx)
{
	ExpAppBuf	*pout = &gMsgOut;
	ComplType	result;
	BYTE		buf[6];

	nid2header(nid, SVC_unackd);
	pout->data.code    = NM_respond_to_query;			// Resp2query
	pout->data.data[0] = 0;								// Mode = off

	if (didx) pout->address.adr.send.dom_node = 0x80;		// ?

	result = send_msg(1);							// deselect

	L_ReadMem(nid, 2, 6, RM_CFGREL, buf);			// dummy read location ?
	return (result);
}

static void r2q_off_known(void)
{
	NIDSTRUCT *nid = gNIDs;
	while (nid) {
		r2q_off(nid->nidpid, nid->dom_idx);
		nid = nid->next;
	}
}

//---------------------------------------------------------------------------

bool L_DoScan(BYTE didx)
{
	ExpAppBuf	*pout = &gMsgOut;
	bool		found = false;
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// broadcast: Resp2query Mode = on ...
	pout->header.flags = 8;									// Explicit
	pout->address.adr.send.rpt_retry = gRPTRETRY;
	pout->address.adr.send.tx_timer  = gTXTIMER;
    pout->address.adr.send.dnet_grp  = 0;						// Subnet 0

	pout->address.format  = 3;									// Broadcast
    pout->address.adr.send.dom_node = didx * 0x80 + 15;			// Backlog 15
    pout->header.svc_tag = SVC_unackd_rpt;
	pout->data.code    = NM_respond_to_query;				// Resp2query
	pout->data.data[0] = 1;									// Mode = on
	if (send_msg(1) == MSG_FAILS) return (found);
	r2q_off_known();
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// send QueryID and deselect the incoming "nids" ...
	while (true) {
		pout->address.format  = 3;								// Broadcast
	    pout->address.adr.send.dom_node = didx * 0x80 +15;		// Backlog 15
	    pout->header.svc_tag = SVC_request;
		pout->data.code    = NM_query_id;					// Query ID
		//pout->data.data[0] = 1;								// selected
		pout->data.data[0] = 2;								// selected+uncfg
		if (send_msg(1) != MSG_SUCCEEDS) return (found);
		found = true;
		add_nid(didx, gMsgRsp.data.data);					// add to nid list
		if (r2q_off(gMsgRsp.data.data, didx) == MSG_FAILS) return (found);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	return (found);
}

//---------------------------------------------------------------------------

static void set_nv_doc(NODE* node, int i, int as, char* nv_name, char* nv_doc )
{
	NVSTRUCT *pnv = NODENV(i);
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (nv_name) {
		my_strncpy(pnv->nm_name, nv_name, sizeof(pnv->nm_name));
	}
	else {
		sprintf(pnv->nm_name, "<%u>", i );		// <nv_index>
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (nv_doc) {
		my_strncpy(pnv->sd_name, nv_doc, sizeof(pnv->sd_name));
	}
	else {
		pnv->sd_name[0] = 0;
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (!as) return;							// it is not an array

// copy name and selfdoc to all members of array ...
	short n;
	char *nm = pnv->nm_name;
	char *sd = pnv->sd_name;
	for (n = 1; n < as; n++) {
		pnv++;							// next nv
		strcpy(pnv->nm_name, nm);
		strcpy(pnv->sd_name, sd);
	}
	pnv = NODENV(i);					// back to actual NV

// add index to nv name ...
	char sind[16];
	for (n = 0; n < as; n++) {
		sprintf(sind, "[%u]", n);			// [array_index]

		if ((strlen(pnv->nm_name) + strlen(sind)) >= sizeof(pnv->nm_name))
			pnv->nm_name[sizeof(pnv->nm_name) - strlen(sind) -1] = 0;

		strcat(pnv->nm_name, sind);
		pnv++;		// next nv
	}
}

//***************************************************************************

static BOOL ldByQueryConfig(NODE *node)
{
	ExpAppBuf	*pout = &gMsgOut;
	BYTE		*b;
	short		i;
	BOOL		brk;

	pout->data.code = NM_query_nv_config;
	brk = FALSE;
	i = 0;
	b = gBuf;
	while( !brk && (i < 4096) ) {
		if( i<255 ) {
			pout->data.data[0] = i & 0xFF;
			if( send_msg(1) != MSG_SUCCEEDS ) brk = TRUE;
		} else {
			pout->data.data[0] = 0xFF;
			pout->data.data[1] = i >> 8;
			pout->data.data[2] = i & 0xFF;
			if( send_msg(3) != MSG_SUCCEEDS ) brk = TRUE;
		}
		if( gMsgRsp.data.code != NM_query_nv_config_succ ) brk = TRUE;
		if( !brk ) {
			memcpy(b, &gMsgRsp.data.data[0], 3);
			b += 3;
			i++;
		}
	}
	node->num_nvs = i;
	return (TRUE);
}

//---------------------------------------------------------------------------

bool L_UploadNode(NODE *node)
{
	BYTE snvt_len[128] = {
		 0, 2, 2, 2,  2, 2, 2, 0,  2, 2, 4, 1,  3, 2, 2, 2,	// 0
		 2, 2, 2, 2,  2, 1, 1, 2,  2, 2, 2, 2,  2, 2, 2, 2,	// 16
		 2, 2, 2, 2, 31, 0, 0, 2,  0, 2, 2, 2,  2, 2, 2, 2,	// 32
		 4, 0, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,	// 48
		 4, 4, 4, 4,  4, 1, 0, 2,  4,12,27, 4,  2, 2, 2, 2,	// 64
		 0, 0, 0, 2,  7, 0, 0, 7, 29, 0, 0, 0,  3, 6, 0, 2,	// 80
		 0, 0, 0, 0,  0, 0, 0, 0,  0, 2, 0, 2,  0, 1, 0, 0,	// 96
		 0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,	// 112
	};
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (!gBuf) gBuf = new BYTE[32768];		// scratch buffer to read in
//	if (node->nodeflgs & NFLAG_UPLD) return (true);
	if (DBG(DBG_NET)) {
		char hnid[32], hpid[32];
		hexdump(hnid, node->nid, 6, 0);
		readable(hpid, (char *)&(node->pid), 8);
		//RETAILMSG(1,(L"Loading node: %a %a\r\n", hnid, hpid));
		fprintf(stderr, "Loading node: %s %s\r\n", hnid, hpid);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	ExpAppBuf	*pout = &gMsgOut;
	int		i, n;
	int		as, ai;
	BYTE	brk, snvt_mode, snvt_type, desc,extrec;
	BYTE	*b;
	char	*nv_name, *nv_doc;
	WORD	nv_fixed, p_snvt, adr, size, nv_count, sel_hi, num_netvars;
	char	expadr,absize,nbsize;
	NVSTRUCT *nv;
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Read only structure
	b = gBuf;
	if( !readmem( node, 8, 28, RM_RDONLY, b ) ) return (false);
	nv_fixed = *b++;
	nv_fixed = (nv_fixed << 8) + *b++;
	nv_count = *b++ & 0x3F;
	node->num_nvs = nv_count;
	p_snvt = *b++;
	p_snvt = (p_snvt << 8) + *b++;
	memcpy(node->pid, b, 8);						// copy progid
	b += 8;											// skip prog id
	if( *b & 0x40 ) node->num_domains = 2; else node->num_domains = 1;
	expadr = *b++ & 0x20;
	node->num_adtbl = *b++ >> 4;
	b++ ;						// skip receive_trans_count
	absize = *b++ & 0x0F;		// app inbuf size
	nbsize = *b >> 4;			// net outbuf size
	if( !nbsize || (nbsize>=11) ){
		if(expadr) {
			if( !absize || (absize>=10) ) node->bigbuf = TRUE;	// >=50 bytes
		} else {
			if( !absize || (absize>=9) ) node->bigbuf = TRUE;	// >=42 bytes
		}
	}
	node->nodeflgs |= VALID_ROD;
//DBG(DBG_NET, "VALID_ROD");
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Query Status
	pout->data.code = NM_query_status;
	if( send_msg(0) != MSG_SUCCEEDS ) return (false);
	if( gMsgRsp.data.code != NM_query_status_succ ) return (false);
	if (gMsgRsp.data.data[11] == 3 ) return (false);
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Location
	if( !readmem(node,2,6,RM_CFGREL,node->location) ) return(false);
	node->nodeflgs |= VALID_LOC;
//DBG(DBG_NET, "VALID_LOC");
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Domain tables
	size = node->num_domains * 15;
	if( !readmem(node,25,size,RM_CFGREL,node->domain[0].id) ) return (false);
	node->domain[0].node &= 0x7F; node->domain[1].node &= 0x7F;
	node->nodeflgs |= VALID_DOM;
//DBG(DBG_NET, "VALID_DOM");
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Address tables
	adr  = size + 25;
	size = node->num_adtbl * 5;
	if( !readmem(node,adr,size,RM_CFGREL,&(node->adtbl[0].fmt_size)) )
		return(FALSE);
	node->nodeflgs |= VALID_ADR;
//DBG(DBG_NET, "VALID_ADR");
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// NV Config
	if( p_snvt == 0xFFFF ) {
//DBG(DBG_NET, "ldByQueryConfig");
		brk = ldByQueryConfig(node);
		nv_count = node->num_nvs;
	}
	else if( nv_count ) {
		adr += size;
		if( !readmem(node,adr,nv_count*3,RM_CFGREL, gBuf) ) return (false);
	}
	if( nv_count ) {
//RETAILMSG(1, (TEXT("...nv_count %d\r\n"), nv_count));
		if (!node->panvs) {
//DBG(DBG_NET, "...get nv mem");
			node->panvs = getMem( nv_count * sizeof(NVSTRUCT) );
		}
		b = gBuf;
		for( i=0; i<nv_count; i++ ) {
//RETAILMSG(1, (TEXT("... %d\r\n"), i));
			nv = NODENV(i);
			nv->index = i;

			nv->pri_len = *b & 0x80;
			sel_hi		= *b++ & 0x7F;
			nv->sel_dir = (sel_hi << 8) + *b++;
			nv->cfg		= *b++;
		}
	}
	node->nodeflgs |= VALID_CFG;
//DBG(DBG_NET, "VALID_CFG");
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// SNVT Structure
	snvt_mode = RM_ABS;
	if( p_snvt ) {
		if( p_snvt == 0xFFFF ) { p_snvt = 0; snvt_mode = RM_SNVT; }
		if( readmem(node,p_snvt,6,snvt_mode, gBuf) ) {
			b = gBuf;
			size = *b++;
			size = (size << 8) + *b++;
			adr = 5;
			num_netvars = *b++;
			if( *b++ ) { num_netvars += ((WORD)(*b) << 8); adr = 6; }
			size -= adr;
			if( !readmem(node,adr+p_snvt,size,snvt_mode, gBuf) ) return (false);
			b = gBuf + num_netvars*2;
			while( *b++ );						// Skip node selfdocu string
			i=0; n=0;
			while( (i<nv_count) && (n<num_netvars) ) {
				desc = gBuf[n*2];
				snvt_type = gBuf[n*2+1];
				as = 0;
				nv_name = nv_doc = NULL;
				if( desc & 0x80 ) {
					extrec = *b++;
					if( extrec & 0x80 ) b++;			// skip mre
					if( extrec & 0x40 ) b++;			// skip re
					if( extrec & 0x20 ) {
						nv_name = (char*)b;
						while( *b++ );					// skip nv name
					}
					if( extrec & 0x10 ) {
						nv_doc = (char*)b;
						while( *b++ );					// skip nv selfdoku
					}
					if( extrec & 0x08 ) {				// array size
						as = *b++;
						as = (as << 8) + *b++;
					}
				}
				if( as > nv_count-i ) as = nv_count-i;
				set_nv_doc( node, i, as, nv_name, nv_doc );
				if( as ) {								// nv array
					for( ai=0; ai < as; ai++ ) {
						NODENV(i)->snvt = snvt_type;
						i++;
					}
				} else {								// simple nv
					NODENV(i)->snvt = snvt_type;
					i++;
				}
				n++;
			}
		}
	}
	node->nodeflgs |= VALID_SNV;

	if( !nv_count ) {
		node->nodeflgs |= VALID_FIX;
		node->nodeflgs |= NFLAG_UPLD;
		return (true);
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// NV length
	if( snvt_mode != RM_SNVT ) {
		if( !readmem(node,nv_fixed,nv_count*3,RM_ABS, gBuf) ) return (false);
		for( i=0; i<nv_count; i++ ) NODENV(i)->pri_len |= (gBuf[i*3] & 0x1F);
	} else {
		for( i=0; i<nv_count; i++ ) {
			n = NODENV(i)->snvt;
			if( n<128 ) n = snvt_len[n]; else n = 0;
			if( !n ) {
				pout->data.code = NM_nv_fetch;
				if( i<255 ) {
					pout->data.data[0] = i;
					if( send_msg(1) != MSG_SUCCEEDS ) return (false);
				} else {
					pout->data.data[0] = 255;
					pout->data.data[1] = i >> 8;
					pout->data.data[2] = i & 0xFF;
					if( send_msg(3) != MSG_SUCCEEDS ) return (false);
				}
				if( gMsgRsp.data.code != NM_nv_fetch_succ ) return (false);
				n = gMsgRsp.header.data_len - 2;
			}
			NODENV(i)->pri_len |= (BYTE)n;
		}
	}

	node->nodeflgs |= VALID_FIX;
//DBG(DBG_NET, "VALID_FIX");
	node->nodeflgs |= NFLAG_UPLD;
	return (true);
}

//---------------------------------------------------------------------------

void Ni_Reset(void)
{
	BYTE	bData[256];

	DWORD	dwCnt = ldv_write(gLP4, gNi_reset, sizeof(gNi_reset));
    //if (debugmode & 0x10)  fprintf(stderr, "ldv_write(%d)=%d\n", sizeof(gNi_reset), dwCnt);

	dwCnt=0;
	//Sleep(1000);
	usleep(1000*1000);
	//ReadFile(gLP4, bData, 256, &dwCnt, NULL);
	dwCnt = read(gLP4, &bData, 256);
}

//***************************************************************************

bool L_SetNodeMode(BYTE *nid, BYTE mode, BYTE state)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access

	gMsgOut.data.code	= NM_set_node_mode;
	gMsgOut.data.data[0] = mode;
	gMsgOut.data.data[1] = state;

	if (!nid) return (send_local(mode == CHANGE_STATE ? 2 : 1) == MSG_SUCCEEDS);

	nid2header(nid, SVC_request);
	return (send_msg(mode == CHANGE_STATE ? 2 : 1) == MSG_SUCCEEDS);
}

bool L_SetNodeMode(BYTE didx, BYTE subnet, BYTE nodenr, BYTE mode, BYTE state)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	snd2header(didx, subnet, nodenr, SVC_request);
	gMsgOut.data.code	= NM_set_node_mode;
	gMsgOut.data.data[0] = mode;
	gMsgOut.data.data[1] = state;
	return (send_msg(mode == CHANGE_STATE ? 2 : 1) == MSG_SUCCEEDS);
}

//---------------------------------------------------------------------------

bool L_Wink(BYTE *nid)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	if (!nid) return (false);
	nid2header(nid, SVC_unackd_rpt);
	gMsgOut.data.code = NM_wink;
	return (send_msg(0) == MSG_SUCCEEDS);
}

bool L_Wink(BYTE didx, BYTE subnet, BYTE nodenr)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	snd2header(didx, subnet, nodenr, SVC_request);
	gMsgOut.data.code = NM_wink;
	return (send_msg(0) == MSG_SUCCEEDS);
}

//---------------------------------------------------------------------------

bool L_QueryStatus(BYTE *nid, BYTE *status)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	ComplType err;
	gMsgOut.data.code = NM_query_status;
	if (!nid) {
		err = send_local(0);
	}
	else {
		nid2header(nid, SVC_request);
		err = send_msg(0);
	}
	if (err != MSG_SUCCEEDS) return (false);

	if (status) memcpy(status, gMsgRsp.data.data, 12);	// ??? 12
	return (gMsgRsp.data.data[11] == 4);
}

bool L_QueryStatus(BYTE didx, BYTE subnet, BYTE nodenr, BYTE *status)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	short	 err;
	gMsgOut.data.code = NM_query_status;
	snd2header(didx, subnet, nodenr, SVC_request);

	err = send_msg(0);
	if (err != MSG_SUCCEEDS) return (false);
	if (status) memcpy(status, gMsgRsp.data.data, 12);	// ??? 12
	return (gMsgRsp.data.data[11] == 4);
}

//---------------------------------------------------------------------------

bool L_ClearStatus(BYTE *nid)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	if (!nid) return (false);
	nid2header(nid, SVC_request);
	gMsgOut.data.code = NV_clear_status;
	return (send_msg(0) == MSG_SUCCEEDS);
}

//---------------------------------------------------------------------------

short nvfetchStub(WORD index, BYTE* buf)
{
	short len;
	gMsgOut.data.code = NM_nv_fetch;
	if (index < 255) {
		gMsgOut.data.data[0] = (BYTE)index;
		if (send_msg(1) != MSG_SUCCEEDS) return (0);
	}
	else {
		gMsgOut.data.data[0] = 255;
		gMsgOut.data.data[1] = index >> 8;
		gMsgOut.data.data[2] = index & 0xFF;
		if (send_msg(3) != MSG_SUCCEEDS) return (0);
	}

	if (gMsgRsp.data.code != NM_nv_fetch_succ ) return (0);
	if ((len = gMsgRsp.header.data_len - 2) < 1 ) return (0);
	memcpy(buf, &gMsgRsp.data.data[1], len);
	return (len);
}

short L_NVFetch(BYTE *nid, WORD index, BYTE* buf)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	if (!nid) return (0);
	nid2header(nid, SVC_request);
	return (nvfetchStub(index, buf));
}

short L_NVFetch(BYTE didx, BYTE subnet, BYTE nodenr, WORD index, BYTE* buf)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	snd2header(didx, subnet, nodenr, SVC_request);
	return (nvfetchStub(index, buf));
}

//---------------------------------------------------------------------------

static bool nvsetStub(WORD sel_dir, WORD len, BYTE *data)
{
	gMsgOut.data.code    = (BYTE)(((sel_dir >> 8) & 0x3F) + 0x80);
	gMsgOut.data.data[0] = (BYTE)sel_dir;
	memcpy(&gMsgOut.data.data[1], data, len);
	if (send_msg(len +1) != MSG_SUCCEEDS) return (false);
	return (true);
}

bool L_NVSet(BYTE *nid, WORD sel_dir, WORD len, BYTE *data)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	if (!nid) return (false);
	nid2header(nid, SVC_ackd);
	return (nvsetStub(sel_dir, len, data));
}

bool L_NVSet(BYTE didx, BYTE subnet, BYTE nodenr, WORD sel_dir, WORD len, BYTE *data)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	snd2header(didx, subnet, nodenr, SVC_request);
	return (nvsetStub(sel_dir, len, data));
}

//---------------------------------------------------------------------------

int readmem(NODE *node, WORD adr, short size, BYTE mode, BYTE *buf)
{
	ExpAppBuf	*pout = &gMsgOut;
	int			blocksize;

	if( node->bigbuf ) blocksize = 32; else blocksize = 16;
	nid2header( node->nid, SVC_request );

	if( mode != RM_SNVT ) {
		pout->data.code    = NM_read_memory;
		pout->data.data[0] = mode;							// abs/rdonly/cfg
		while( size ){
			if( blocksize > size ) blocksize = size;
			pout->data.data[1] = adr >> 8;
			pout->data.data[2] = adr & 0xFF;
			pout->data.data[3] = blocksize;
			if( send_msg(4) != MSG_SUCCEEDS ) return( FALSE );
			if( gMsgRsp.data.code != NM_read_memory_succ ) return( FALSE );
			memcpy( buf, gMsgRsp.data.data, blocksize );
			buf  += blocksize;
			adr  += blocksize;
			size -= blocksize;
		}
	}
	else {
		pout->data.code = NM_query_snvt;
		while( size ){
			if( blocksize > size ) blocksize = size;
			pout->data.data[0] = adr >> 8;
			pout->data.data[1] = adr & 0xFF;
			pout->data.data[2] = blocksize;
			if( send_msg(3) != MSG_SUCCEEDS ) return( FALSE );
			if( gMsgRsp.data.code != NM_query_snvt_succ ) return( FALSE );
			memcpy( buf, gMsgRsp.data.data, blocksize );
			buf += blocksize;
			adr += blocksize;
			size -= blocksize;
		}
	}
	return( TRUE );
}

//---------------------------------------------------------------------------

static int readmemStub(bool local, WORD adr, short size, BYTE *buf)
{
	ExpAppBuf	*pout = &gMsgOut;
	int			blocksize = 16;

	while (size) {
		if (blocksize > size) blocksize = size;
		pout->data.data[1] = adr >> 8;
		pout->data.data[2] = adr & 0xFF;
		pout->data.data[3] = blocksize;
		if (local) {
			if (send_local(4) != MSG_SUCCEEDS) return (FALSE);
		}
		else {
//			RETAILMSG(1,(TEXT("***SendMSG\r\n")));
			if (send_msg(4) != MSG_SUCCEEDS) return (FALSE);
		}
		if (gMsgRsp.data.code != NM_read_memory_succ) return (FALSE);
		memcpy(buf, gMsgRsp.data.data, blocksize);
		buf  += blocksize;
		adr  += blocksize;
		size -= blocksize;
	}
	return (TRUE);
}

int L_ReadMem(BYTE *nid, WORD adr, short size, BYTE mode, BYTE *buf )
{
	//CLockCrit lock(&gLdvLock);	// protect access
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access

	if (nid) nid2header(nid, SVC_request);
	gMsgOut.data.code    = NM_read_memory;
	gMsgOut.data.data[0] = mode;							// abs/rdonly/cfg
	return (readmemStub(nid == NULL, adr, size, buf));
}

//---------------------------------------------------------------------------

short L_ReadMem(BYTE didx, BYTE subnet, BYTE nodenr,
			    WORD adr, short size, BYTE mode, BYTE *buf)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	snd2header(didx, subnet, nodenr, SVC_request);
	gMsgOut.data.code    = NM_read_memory;
	gMsgOut.data.data[0] = mode;	// abs/rdonly/cfg
//	RETAILMSG(1,(TEXT("****************READMEM1\r\n")));
//  by IDel changed for local read
//	return (readmemStub(false, adr, size, buf));
	return (readmemStub((didx == 0 && subnet == 0 && nodenr == 0), adr, size, buf));
}

//---------------------------------------------------------------------------

static bool writememStub(bool local, WORD adr, short size, BYTE mode, BYTE *buf)
{
	short		ind, anz;
	ExpAppBuf	*pout = &gMsgOut;

	ind = 0;
	while (size > 0) {
		anz = size;
		if (anz > 11) anz = 11;
		size -= anz;

		pout->data.code    = NM_write_memory;
		pout->data.data[0] = mode;
		pout->data.data[1] = (BYTE)(adr >> 8);
		pout->data.data[2] = (BYTE)(adr & 0xFF);
		pout->data.data[3] = (BYTE)(anz);
		pout->data.data[4] = size > 0 ? 0 : 1;		// recalculate checksum
		memcpy(&pout->data.data[5], &buf[ind], anz);
		if (local) {
			if (MSG_SUCCEEDS != send_local(anz +5)) return (false);
		}
		else {
			if (MSG_SUCCEEDS != send_msg(anz +5)) return (false);
		}
		ind += anz;
	}
	return (true);
}

bool L_WriteMem(BYTE *nid, WORD adr, short size, BYTE mode, BYTE *buf)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	if (nid) nid2header(nid, SVC_ackd);
	return (writememStub(nid == NULL, adr, size, mode, buf));
}

bool L_WriteMem(BYTE didx, BYTE subnet, BYTE nodenr,
			    WORD adr, short size, BYTE mode, BYTE *buf)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	snd2header(didx, subnet, nodenr, SVC_ackd);
	return (writememStub(false, adr, size, mode, buf));
}

//---------------------------------------------------------------------------

bool L_SendMsg(BYTE didx, BYTE subnet, BYTE nodenr,
			   BYTE svr, BYTE *data, BYTE len)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	snd2header(didx, subnet, nodenr, svr);
	gMsgOut.data.code = data[0];
	len--;
	if (len) memcpy(&gMsgOut.data, &data[1], len);
	if (send_msg(len) != MSG_SUCCEEDS) return (false);
	return (true);
}

//---------------------------------------------------------------------------

bool L_UpdDomain(BYTE *nid, BYTE didx, BYTE domlen, BYTE *id,
						    BYTE subnet, BYTE nodenr)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access

	if ((didx != 0) && (didx != 1)) return (false);
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	ExpAppBuf *pout = &gMsgOut;
	ComplType	  err;

	pout->data.code    = NM_update_domain;
	pout->data.data[0] = didx;							// Domain index
	memset(&pout->data.data[1], 0, 6);
	if (domlen)	memcpy(&pout->data.data[1], id, domlen);	// Domain ID

	pout->data.data[7] = subnet;							// Subnet
	pout->data.data[8] = (nodenr | 0x80);				// Node
	pout->data.data[9] = domlen;							// Domain length
	memset(&pout->data.data[10], 0xFF, 6);				// Domain key

	if (!nid) {
		err = send_local(16);						// Update domain
	}
	else {
		nid2header(nid, SVC_request);
		err = send_msg(16);
	}
	if (err != MSG_SUCCEEDS) return (false);
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	pout->data.code = NM_set_node_mode;
	pout->data.data[0] = CHANGE_STATE;					// Change state
	pout->data.data[1] = 4;								// Config online
	if (!nid) {
		send_local(2);
	}
	else {
		nid2header(nid, SVC_request);
		if (MSG_SUCCEEDS != send_msg(16)) return (false);
	}
	return (true);
}

//---------------------------------------------------------------------------

bool L_SendSvp(BYTE *nid, BYTE *pid)
{
	ExpAppBuf	*pout = &gMsgOut;

	pout->header.flags = 8;									// Explicit
	pout->address.adr.send.rpt_retry = gRPTRETRY;
	pout->address.adr.send.tx_timer  = gTXTIMER;

    pout->address.adr.send.dnet_grp  = 0;						// Subnet 0
    pout->address.adr.send.dom_node  = 0;

	pout->address.format  = 3;									// Broadcast
    pout->header.svc_tag = SVC_unackd;

	pout->data.code    = 0x7F;								// service pin
	memcpy(&pout->data.data[0], nid, 6);
	memcpy(&pout->data.data[6], pid, 8);

	if (send_msg(14) == MSG_FAILS) return (false);

	return (true);
}

//***************************************************************************

static bool add_nid(BYTE didx, BYTE *nidpid)
{
	NIDSTRUCT *nid = gNIDs;
	while (nid) {
		if (memcmp(nid->nidpid, nidpid, 6) == 0) {
		//DBG(DBG_NET, "..already there");
			return (false);							// already there
		}
		nid = nid->next;
	}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	gNIDcount++;
	nid = new NIDSTRUCT;
	memset(nid, 0, sizeof(NIDSTRUCT));

	nid->dom_idx = didx;
	memcpy(nid->nidpid, nidpid, sizeof(nid->nidpid));

//char h[32];
//hexdump(h, nid->nidpid, 14, 0);
//RETAILMSG(1, (TEXT("add_nid %a\r\n"), h));

	nid->next = gNIDs;
	gNIDs	  = nid;
	return (true);
}

//---------------------------------------------------------------------------

void L_CloseLoad(void)
{
	NIDSTRUCT	*nid, *next;
	next = gNIDs;
	while ((nid = next))
	{
		next = nid->next;
		delete nid;
	}
	gNIDcount = 0;
	if (gBuf) delete gBuf;
	gBuf = NULL;
}

//---------------------------------------------------------------------------

static BYTE *getMem(long size)
{
	BYTE *mem = new BYTE[size];
	if (!mem) {
		DBGERR("*** g2LonC: memory not allocated");
		return (NULL);
	}

	memset(mem, 0, size);
	return (mem);
}


//***************************************************************************
// am 21.5.2003 ergaenzt von Ingo Grothues

bool L_UpdateAddress(BYTE *nid, BYTE addr_idx, ADTBL *entry)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access

	gMsgOut.data.code	= NM_update_adr;

	gMsgOut.data.data[0] = addr_idx;
	gMsgOut.data.data[1] = entry->fmt_size;
	gMsgOut.data.data[2] = entry->dom_node;
	gMsgOut.data.data[3] = gRPTRETRY;
	gMsgOut.data.data[4] = gTXTIMER;
	gMsgOut.data.data[5] = entry->net_grp;

	if (!nid) return (send_local(6) == MSG_SUCCEEDS);

	nid2header(nid, SVC_request);
	return (send_msg(6) == MSG_SUCCEEDS);
}


bool L_UpdateNVconfig(BYTE *nid, CNV *nv)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access

	int msg_len;

	WORD idx     = nv->Index();
	WORD sel_dir = nv->GetSel_dir();
	BYTE cfg     = nv->GetCfg();

	gMsgOut.data.code	= NM_update_nv_config;

	if(idx < 255)
	{
		gMsgOut.data.data[0] = idx & 0xFF;
		gMsgOut.data.data[1] = sel_dir >> 8;
		gMsgOut.data.data[2] = sel_dir & 0xFF;
		gMsgOut.data.data[3] = cfg;
		msg_len = 4;
	}
	else
	{
		gMsgOut.data.data[0] = 0xFF;
		gMsgOut.data.data[1] = idx >> 8;
		gMsgOut.data.data[2] = idx & 0xFF;
		gMsgOut.data.data[3] = sel_dir >> 8;
		gMsgOut.data.data[4] = sel_dir & 0xFF;
		gMsgOut.data.data[5] = cfg;
		msg_len = 6;
	}

	if (!nid) return (send_local(msg_len) == MSG_SUCCEEDS);

	nid2header(nid, SVC_request);
	return (send_msg(msg_len) == MSG_SUCCEEDS);
}

//------------------------------------------------------------------------------


// Diese Funktionen liefern den vollen Statuscode (15 Bytes) zurueck

bool L_QueryStatus2(BYTE *nid, BYTE *status)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	ComplType err;
	gMsgOut.data.code = NM_query_status;
	if (!nid) {
		err = send_local(0);
	}
	else {
		nid2header(nid, SVC_request);
		err = send_msg(0);
	}
	if (err != MSG_SUCCEEDS) return (false);
	if (status) memcpy(status, gMsgRsp.data.data, 15);
	return true;
}

bool L_QueryStatus2(BYTE didx, BYTE subnet, BYTE nodenr, BYTE *status)
{
	CLockCrit lock(&gLdvLock, ALLOW_2ND_LOCK);	// protect access
	//CLockCrit lock(&gLdvLock);						// protect access
	short	 err;
	gMsgOut.data.code = NM_query_status;
	snd2header(didx, subnet, nodenr, SVC_request);

	err = send_msg(0);
	if (err != MSG_SUCCEEDS) return (false);
	if (status) memcpy(status, gMsgRsp.data.data, 15);
	return true;
}

//***************************************************************************
