//////////////////////////////////////////////////////////////////////
//
// watcher.cpp:  implementation of basic watcher routines
//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
// Copyright (c) 1999  Gesytec GmbH
//
// Date:    10. 1.99
//
// Author: mb, mboehlke@gesytec.de
//
//
//////////////////////////////////////////////////////////////////////

#include "windows_compat.h"
#include "ni_privat.h"
#include "g2Lon.h"
#include "watcher.h"
#include "Utils.h"

#include <sys/ioctl.h>
#include <stdio.h>
#include <string.h>

// some globals ...
extern HANDLE	gLP4;

//***************************************************************************

#define CMD_INIT	0
#define CMD_PACKET	1

//---------------------------------------------------------------------------

static short dec2flat(FLAT *flat, BYTE *paket, long len)
// this is the very recent version of the LonTalk to FLAT structure decoder
{
	BYTE i;

	if (len < 8) {
	//	RETAILMSG(1, (TEXT("<8 ")));			// packet too short
		return (-1);
	}
	if (paket[6] & CRC_ERR) {					// CRC check
	//	RETAILMSG(1, (TEXT("crc? ")));
		return (-1);
	}
//  paket[6] & 1 -> verlust						// ???
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	BYTE *s  = paket;
	//BYTE *dp = (BYTE *)&flat->itime;
	flat->itime = 0;
	memcpy(&flat->itime, s, 6);
	s += 6;

	flat->flags    = *s++;
	flat->mac      = *s++;

	flat->version  = (*s >> 6) & 3;
	flat->pfmt	   = (*s >> 4) & 3;
	flat->afmt	   = (*s >> 2) & 3;
	flat->domlen   = *s++	   & 3;
	if (flat->domlen > 1) {
		if		(flat->domlen == 2) flat->domlen = 3;
		else if (flat->domlen == 3) flat->domlen = 6;
	}

	flat->ssubnet  = *s++;
	flat->snode    = *s++;
	flat->dnet_grp = *s++;

	if (flat->snode & 0x80) i = 6 >> ((3 - flat->afmt) << 1);
	else					i = 3;
	if (i) memcpy(&flat->dnode, s, i);
	s += i;

	if (flat->domlen) {
		memcpy(&flat->domain, s, flat->domlen);
		s += flat->domlen;
	}

	if (flat->pfmt != 3) {
		flat->auth = (*s >> 7) & 1;
		flat->ptyp = (*s >> 4) & 7;
		flat->tanr = *s++  & 0x0F;

		if ((flat->ptyp & 6) == 4) {
			flat->mlistlen = *s++;
			if (flat->mlistlen) {
				memcpy(flat->mlist, s, flat->mlistlen);
				s += flat->mlistlen;
			}
		}
	}

	flat->dlen = len - (s - paket);		// 2 CRC bytes following ????

	// hoeh ?? - BYTE ist unsigned 8byte Datentyp, wie soll der <0 oder >255 werden ?

/*	if (flat->dlen < 0) {
		return (-1);
	}
	if (flat->dlen > 255) {
		return (-1);
	}
*/

	// WORKAROUND
	if((len - (s - paket) < 0) || (len - (s - paket) >255)) return (-1);

	memcpy(&flat->data, s, flat->dlen);
	return (flat->dlen);
}

//---------------------------------------------------------------------------

static short do_watcher(BYTE *buffer, long bufsize, BYTE cmd, int count)
{
	DWORD BytesReturned;

	buffer[0] = count >> 8;
	buffer[1] = count & 0xFF;
	buffer[2] = cmd;

	BytesReturned = ioctl(gLP4, IOCTL_WATCHER, &buffer[0]);
	if (BytesReturned < 0)
	{
		perror("DeviceIoControl failed !\r\n");
		return (-1);
	}
/*
	if (!DeviceIoControl(gLP4, IOCTL_WATCHER,
		&buffer[0], bufsize,
		&buffer[0], bufsize,
        &BytesReturned, NULL )) {

		RETAILMSG( 1, (L"DeviceIoControl failed !\r\n") );
		return(-1);
	}
*/	if (BytesReturned < 0) return (-1);
	if (BytesReturned < 2) return (0);		// was ist mit 1 ?

	count = buffer[0];
	count = (count << 8) + buffer[1];
	return (count);							// length of following data
}

//---------------------------------------------------------------------------

bool W_Init(void)
{
	BYTE buffer[64];
	buffer[3] = 5;	// FTT10				//		0 : 1,2 MBit ???
	short len = do_watcher(buffer, sizeof(buffer), CMD_INIT, 2);
	return (len >= 0);
}

//---------------------------------------------------------------------------

short W_GetPacket(BYTE *packet, DWORD len)
{
// check for and get data from watcher ...
	return (do_watcher(packet, len, CMD_PACKET, 1));
}

//---------------------------------------------------------------------------

short W_Dec2Flat(FLAT *flat, BYTE *packet, short len)
{
	short dlen;									// data length
	if (len <= 0) return (-1);
	dlen = dec2flat(flat, packet, len);
	return (dlen);
}

//***************************************************************************

bool W_IsSrv_Sent(FLAT *flat)					//
{
	if (flat->dlen != 15)	return (false);		// length must be 15
	if (flat->ssubnet)		return (false);		// source subnet == 0
	if (flat->snode & 0x7F)	return (false);		// source node   == 0
	if (flat->data[0] != 0x7F) return (false);	// msg code == 0x7F
	return (true);
}

//---------------------------------------------------------------------------

bool W_IsNV_Sent(FLAT *flat)
{
	if (flat->dlen <= 2)		 return (false);	// selector + some data
	if (!(flat->data[0] & 0x80)) return (false);	// NV selector

	if (flat->pfmt == 0) {							// TPDU services
		if (flat->ptyp == 0) return (true);			//  ACKD: 0
		if (flat->ptyp == 1) return (true);			//  UnACKD_Rpt: 1
		if (flat->ptyp == 5) return (true);			//  REM/MSG: 5
		return (false);
	}

	return (false);
}

//---------------------------------------------------------------------------

bool W_IsMsg_Sent(FLAT *flat)
{
	if (flat->dlen < 1)		  return (false);		// msg code must be there
	if (flat->data[0] & 0x80) return (false);		// msg code 00 - 7F

	if (flat->pfmt == 0) {							// TPDU services
		if (flat->ptyp == 0) return (true);			//  ACKD: 0
		if (flat->ptyp == 1) return (true);			//  UnACKD_Rpt: 1
		if (flat->ptyp == 5) return (true);			//  REM/MSG: 5
		return (false);
	}

	if (flat->pfmt == 1) {							// SPDU services
		if (flat->ptyp == 0) return (true);			//  REQUEST: 0
		if (flat->ptyp == 5) return (true);			//  REM/MSG: 5
		return (false);
	}

	return (false);
}

//---------------------------------------------------------------------------
// returns the CNV object by selector given in packet ...
CNV *W_GetSentNV(FLAT *flat, CNeuron *neuron)
{
    if (flat->dlen <= 2) return (NULL);				// must have selector

// get selector ...
	WORD sel;
 	sel  = (WORD)(flat->data[0] & 0x3F) << 8;
	sel += (WORD)flat->data[1];

// find sending neuron's output NV ...
    return (W_GetNVbySel(neuron, sel + 0x4000));	// find output NV
}

//---------------------------------------------------------------------------
// searches the NV by selector ...
CNV *W_GetNVbySel(CNeuron *neuron, WORD sel)
{
	CNV  *pnv;
	WORD h;
	for (short npos = 0; npos < neuron->NVCount(); npos++) {
		pnv = neuron->NVbyInd(npos);
		h = pnv->Selector();
		if (pnv->isOutput()) h += 0x4000;
		if (h == sel) return (pnv);
	}
	return (NULL);
}

//***************************************************************************

