#ifndef _WINDOWS_COMPAT_H_
#define _WINDOWS_COMPAT_H_

#define LONC_API	/*dummy*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>

#include <pthread.h>
#include <linux/types.h>

#include "globals.h"


#define _int64	__s64

#define	HANDLE					signed
#define	INVALID_HANDLE_VALUE	0xFFFFFFFF

//! SOCKET type, needed for gesytec/echelon/windows code
typedef int        SOCKET;

// This is used instead of -1, since the
// SOCKET type is unsigned.
#define INVALID_SOCKET  (SOCKET)(~0)
#define SOCKET_ERROR            (-1)

#ifndef cpu_to_le16
  #define cpu_to_le16(x) htole16(x)
#endif
#ifndef le16_to_cpu
  #define le16_to_cpu(x) le16toh(x)
#endif
#ifndef cpu_to_le32
  #define cpu_to_le32(x) htole32(x)
#endif
#ifndef le32_to_cpu
  #define le32_to_cpu(x) le32toh(x)
#endif

typedef u_int8_t  BYTE, UCHAR;
typedef u_int16_t WORD;
typedef u_int32_t DWORD;
typedef int32_t   LONG;

#define BOOL	bool
#ifndef FALSE
#define FALSE	false
#endif
#ifndef TRUE
#define TRUE	true
#endif


typedef unsigned char   UCHAR;
typedef signed char     CHAR;

//! unsigned character pointer, needed for gesytec/echelon/windows code
typedef UCHAR*      PUCHAR;
typedef CHAR*       PCHAR, LPSTR, PSTR, LPTSTR, PTSTR;
typedef const CHAR* PCCHAR, LPCSTR, PCSTR, LPCTSTR, PCTSTR;

//typedef USHORT*     PUSHORT;
//typedef SHORT*      PSHORT;

//! character, needed for gesytec/echelon/windows code
typedef char       TCHAR;


#define LPVOID		void*
#define LPVARIANT	void*

//#define VARTYPE		void
//#define FILETIME	char*

#define CRITICAL_SECTION 				pthread_mutex_t
#define InitializeCriticalSection(x)
#define DeleteCriticalSection(x)
#define EnterCriticalSection(x) 		pthread_mutex_lock(x);
#define TryCriticalSection(x) 			pthread_mutex_trylock(x);
#define LeaveCriticalSection(x)			pthread_mutex_unlock(x);

typedef void*(*LPTHREAD_START_ROUTINE)(void*);


#define	WAIT_OBJECT_0		0
#define	WAIT_TIMEOUT		1
#define	WAIT_ABANDONED		2
#define INFINITE			0xFFFFFFFF

#ifndef min
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#endif

#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

//long 	max(long a, long b);

HANDLE	CreateEvent(void* attr, bool reset, bool init_val, char* name);
void	SetEvent(HANDLE &evt);
DWORD	WaitForSingleObject(HANDLE &evt, DWORD timeout);

#endif
