
#include "windows_compat.h"
#include "yaList.h"

#include <stdio.h>
#include <string.h>

//***************************************************************************
// yet another list (a very simple one) 
// there is some similarity to CObList 
// !! these routines are not thread save !!

#define LISTINC	64

CyaList::CyaList()
{
	m_len   = 1;		// LISTINC; // often there are only list constructors
	m_list  = new CyaObj *[m_len];				// create initial space
	m_count = 0;								// the list is empty
	m_inc   = LISTINC;
}

CyaList::CyaList(long inc)
{
	m_len   = inc;
	m_list  = new CyaObj *[m_len];				// create initial space
	m_count = 0;								// the list is empty
	m_inc   = inc;
}

CyaList::~CyaList()
{
	if (m_list) delete m_list;					// release memory space
}

//---------------------------------------------------------------------------

bool CyaList::increase(long inc)
{
	m_len += inc;								// increment for new length
	CyaObj **nl = new CyaObj *[m_len];			// get new memory
	if (!nl) {
		perror("CyaList::increase: no memory !\r\n");
		//RETAILMSG(1, (TEXT("CyaList::increase: no memory !\r\n")));
		return (false);
	}

	memcpy(nl, m_list, m_count * sizeof(CyaObj *));

	delete m_list;								// delete old list
	m_list = nl;								//  and set to new list
	return (true);
}

//---------------------------------------------------------------------------

void CyaList::AddTail(CyaObj *p)
{
	if (m_count >= m_len) {						// uups: no more space
		if (!increase(m_inc)) return;
	}											//  now we must have space

	m_list[m_count++] = p;						// just store and inc count
}

//---------------------------------------------------------------------------

void CyaList::InsertAt(long pos, CyaObj *p)
{
	if (pos < 0) return;
	if (pos >= m_count) { AddTail(p); return; }
	if (m_count == 0)   { AddTail(p); return; }

	if (m_count >= m_len) {						// uups: no more space
		if (!increase(m_inc)) return;
	}											//  now we must have space

	memmove(&m_list[pos +1], &m_list[pos], (m_count - pos) * sizeof(CyaObj *));

	m_list[pos] = p;							// just store 
	m_count++;									//	and inc count
}

//---------------------------------------------------------------------------

void CyaList::SetAt(long pos, CyaObj *p)
{
	if (pos < 0)		return;
	if (pos >= m_count) return;
	m_list[pos] = p;
}

//---------------------------------------------------------------------------

CyaObj *CyaList::GetAt(long pos)
{
	if (pos <		 0) return (NULL);
	if (pos >= m_count) return (NULL);
	return (m_list[pos]);						// just return the pointer
}

//---------------------------------------------------------------------------

long CyaList::GetCount(void)
{
	return (m_count);
}

//---------------------------------------------------------------------------

long CyaList::Find(CyaObj *p)
{
	for (long pos = 0; pos < m_count; pos++) {
		if (p == m_list[pos]) return (pos);
	}
	return (-1);
}

//---------------------------------------------------------------------------

bool CyaList::Remove(CyaObj *p)
{
	long pos = Find(p);
	if (pos < 0) return (false);
	RemoveAt(pos);
	return (true);
}

//---------------------------------------------------------------------------

void CyaList::RemoveAt(long pos)
{
	if (pos <		 0) return;
	if (pos >= m_count) return;

	m_count--;
	long n = m_count - pos;
	if (n) memmove(&m_list[pos], &m_list[pos +1], n * sizeof(CyaObj *));
}

//---------------------------------------------------------------------------

void CyaList::RemoveAll()
// we just reset the counter, we dont release the memory
{
	m_count = 0;
}

//---------------------------------------------------------------------------

void CyaList::DeleteAll()
{
	for (long pos = 0; pos < GetCount(); pos++) {
		delete GetAt(pos);
	}
	RemoveAll();
}

//***************************************************************************

CyaList* CyaList::Clone(bool dup)
{
	CyaList* list ;
	list = new CyaList( max( m_count * ( dup ? 2 : 1 ), 1 ) ) ;
	memcpy(list->m_list, m_list, m_count * sizeof(CyaObj *));
	list->m_count = m_count ;
	return( list ) ;
}

//---------------------------------------------------------------------------

void CyaList::Sort( bool (*comp)( void* a, void* b ) )
{
	if( m_count <= 1 ) return ;
	if( m_len < 2 * m_count ) increase( 2 * m_count - m_len ) ;
	m_count *= 2 ; // knstliche Verdoppelung der Elementanzahl
	srt( 0, m_count / 2 , comp, m_count / 2 ) ;
	m_count /= 2 ; // Verdoppelung rckgngig
}

//---------------------------------------------------------------------------

void CyaList::srt( long p, long n, bool (*comp)( void* a, void* b ), long f )
{
	if( n > 1 ) {
		long n1 = n/2 ;
		long n2 = (n+1)/2 ;
		srt( p, n1, comp, f ) ;
		srt( p+n1, n2, comp, f+n1 ) ;
		mrg( comp, f, n1, f+n1, n2, p ) ;
		memcpy( &m_list[f], &m_list[p], n * sizeof(CyaObj *));
	}
	else m_list[f] = m_list[p] ;
}

//---------------------------------------------------------------------------

void CyaList::mrg( bool (*comp)( void* a, void* b ), 
	long p1, long n1, long p2, long n2, long p )
{
	while( n1 && n2 ) {
		if( comp( m_list[p1], m_list[p2] ) ) {
			m_list[p++]=m_list[p1++] ;
			n1--;
		}
		else {
			m_list[p++]=m_list[p2++] ;
			n2--;
		}
	}
	while( n1-- ) m_list[p++]=m_list[p1++] ;
	while( n2-- ) m_list[p++]=m_list[p2++] ;
}

//***************************************************************************
