#ifndef _YASTRING_H
#define _YASTRING_H

#include "yaList.h"

//---------------------------------------------------------------------------

class CyaString : public CyaObj
{
public:
// Constructors
	CyaString	();
 	CyaString	(const CyaString& stringSrc);
	CyaString	(LPCSTR lpsz);
	~CyaString	();

	void	Lock	(void);
	void	UnLock	(void);
// Attributes & Operations
//	as an array of characters
	int  GetLength() const;
	BOOL IsEmpty() const;
	void Empty();                       // free up the data

	char GetAt(int nIndex) const;		// 0 based
//	char operator[](int nIndex) const;	// same as GetAt
	void SetAt(int nIndex, char ch);
//	operator LPCSTR() const;			// as a C string
	char *cstring(void) const;			// as a C string

// overloaded assignment
	const CyaString& operator=(const CyaString& stringSrc);
	const CyaString& operator=(char ch);
	const CyaString& operator=(LPCSTR lpsz);

	// string concatenation
	const CyaString& operator+=(const CyaString& string);
	const CyaString& operator+=(char ch);
	const CyaString& operator+=(LPCSTR lpsz);

	friend CyaString operator+(const CyaString& string1,
			const CyaString& string2);
	friend CyaString operator+(const CyaString& string, char ch);
	friend CyaString operator+(char ch, const CyaString& string);
	friend CyaString operator+(const CyaString& string, LPCSTR lpsz);
	friend CyaString operator+(LPCSTR lpsz, const CyaString& string);

	// string comparison
	int Compare(LPCSTR lpsz) const;				// straight character

	// searching (return starting index, or -1 if not found)
	// look for a single character match
	int Find(char ch) const;					// like "C" strchr

// Implementation
protected:
	CRITICAL_SECTION mLock;	
	int		nAllocLength;
	char	*mData;

// implementation helpers
	void Init			(void);
	void AllocBuffer	(int nLen);
	void ExpandBuffer	(int nLen);
	void Set			(char *str);
	void ConcatCopy		(char *str1, char *str2);
	void ConcatInPlace	(char *str);
	void Release		(void);
};
/*
	int ReverseFind(char ch) const;
	int FindOneOf(LPCSTR lpszCharSet) const;

	// look for a specific sub-string
	int Find(LPCSTR lpszSub) const;        // like "C" strstr

  // simple sub-string extraction
	CyaString Mid(int nFirst, int nCount) const;
	CyaString Mid(int nFirst) const;
	CyaString Left(int nCount) const;
	CyaString Right(int nCount) const;

	CyaString SpanIncluding(LPCSTR lpszCharSet) const;
	CyaString SpanExcluding(LPCSTR lpszCharSet) const;

	// upper/lower/reverse conversion
	void MakeUpper();
	void MakeLower();
	void MakeReverse();

	// trimming whitespace (either side)
	void TrimRight();
	void TrimLeft();

//	int CompareNoCase(LPCSTR lpsz) const;		// ignore case
//	int Collate(LPCSTR lpsz) const;				// NLS aware

  // Compare helpers
bool operator==(const CyaString& s1, const CyaString& s2);
bool operator==(const CyaString& s1, LPCSTR s2);
bool operator==(LPCSTR s1, const CyaString& s2);
bool operator!=(const CyaString& s1, const CyaString& s2);
bool operator!=(const CyaString& s1, LPCSTR s2);
bool operator!=(LPCSTR s1, const CyaString& s2);
bool operator<(const CyaString& s1, const CyaString& s2);
bool operator<(const CyaString& s1, LPCSTR s2);
bool operator<(LPCSTR s1, const CyaString& s2);
bool operator>(const CyaString& s1, const CyaString& s2);
bool operator>(const CyaString& s1, LPCSTR s2);
bool operator>(LPCSTR s1, const CyaString& s2);
bool operator<=(const CyaString& s1, const CyaString& s2);
bool operator<=(const CyaString& s1, LPCSTR s2);
bool operator<=(LPCSTR s1, const CyaString& s2);
bool operator>=(const CyaString& s1, const CyaString& s2);
bool operator>=(const CyaString& s1, LPCSTR s2);
bool operator>=(LPCSTR s1, const CyaString& s2);
*/
/////////////////////////////////////////////////////////////////////////////

#endif
